/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */


/* bases on programming exercise, "here's the left write the right".
 wasn't so easy than it sounds */


#include "scache.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>
#include <syslog.h>
#include <obstack.h>

#define obstack_chunk_alloc    malloc
#define obstack_chunk_free     free

#define OBSTACK_CHUNK_SIZE     (1024 * 1024)

struct sc_tree_node_t {

    enum {
	NODE_RED,
	NODE_BLACK
    } color;

    struct sc_tree_node_t *left;
    struct sc_tree_node_t *right;
    struct sc_tree_node_t *parent;

    char *key;
    void *data;

#ifndef NDEBUG
    /* 0xb00b when out, 0xbeeb when in */
    void *b00b;
#endif
};

struct sc_tree_t {

    struct sc_tree_node_t *root;
    struct sc_tree_node_t *nil;
    struct sc_tree_node_t nil_node;

#ifndef NDEBUG
    /* 0xb00b when out, 0xbeeb when in */
    void *b00b;
#endif
};



/* memory allocation variables */
struct obstack obs;

#ifdef STAT_ME_HARDER
typedef struct sc_memqueue_t sc_memqueue_type;

struct sc_memqueue_t memq16     = { .token = "16" };
struct sc_memqueue_t memq32     = { .token = "32" };
struct sc_memqueue_t memq64     = { .token = "64" };
struct sc_memqueue_t memq128    = { .token = "128" };
struct sc_memqueue_t memq256    = { .token = "256" };
struct sc_memqueue_t memq512    = { .token = "512" };
struct sc_memqueue_t memq1024   = { .token = "1024" };

#else

struct sc_memqueue_t {
    void *data;
    struct sc_memqueue_t *next;
};
typedef struct sc_list_t *sc_memqueue_type;

struct sc_list_t *memq16;
struct sc_list_t *memq32;
struct sc_list_t *memq64;
struct sc_list_t *memq128;
struct sc_list_t *memq256;
struct sc_list_t *memq512;
struct sc_list_t *memq1024;
#endif


/* var */
static struct sc_list_t *node_q = NULL;
static struct sc_list_t *tree_q = NULL;

static struct sc_stats_t stats_struct;

/* extern */
time_t             sc_lazytime   = 0;
struct sc_stats_t *sc_stats      = &stats_struct;


void *
sc_obs_alloc(int size) {

    sc_stats->obstack_size += size;
    return obstack_alloc(&obs, size);
}

void *
sc_obs_calloc(int size) {

    void *addr = sc_obs_alloc(size);
    memset(addr, 0, size);
    return addr;
}


void *sc_alloc_s(
#ifdef STAT_ME_HARDER
		 struct sc_memqueue_t *queue,
#else
		 struct sc_list_t **queue,
#endif 
		 int qsize) {

    
    struct sc_list_t *item;
    struct sc_memb_t *memb;

#ifdef STAT_ME_HARDER
    item = queue->queue;
#else
    item = *queue;
#endif

    if (item) {

	memb = (struct sc_memb_t *)item;
	memb--;

	assert((memb->refcount == 0));
	assert((memb->b00f == (void *)0xbeef));
	assert((memb->b00f = (void *)0xb00f));
	memb->refcount = 1;

#ifdef STAT_ME_HARDER
	assert((memb->queue == queue));
	queue->queue = item->next;
	queue->out++;
	queue->total++;
	return item;
#else
	assert((memb->queue == queue));
	return (*queue = item->next), (void *)item;
#endif
    }

    /* else */
    memb = (struct sc_memb_t *)
	sc_obs_alloc(qsize + sizeof(struct sc_memb_t));
    memb->refcount = 1;
    memb->queue = queue;

#ifdef STAT_ME_HARDER
    queue->allocs++;
    queue->total++;
    queue->out++;
#endif

    assert((memb->b00f = (void *)0xb00f));
    return (void *)(++memb);
}



void
sc_membinfo(void *addr) {

    struct sc_memb_t *memb = (struct sc_memb_t *)addr;

    memb--;

    assert((addr));
#ifndef NDEBUG
    fprintf(stderr, 
	    "memb: %p (%p)\n"
	    "  refcount : %i\n"
	    "  b00f     : %p\n",
	    memb, addr, memb->refcount, memb->b00f);
#endif
}



void
sc_free(void *addr) {

    struct sc_list_t *item = (struct sc_list_t *)addr;
    struct sc_memb_t *memb = (struct sc_memb_t *)addr;

    memb--;

    assert((addr));
    assert((memb->refcount == 0));
    assert((memb->b00f == (void *)0xb00f));

    if (memb->queue) {

	assert(((memb->queue == &memq16) ||
		(memb->queue == &memq32) ||
		(memb->queue == &memq64) ||
		(memb->queue == &memq128) ||
		(memb->queue == &memq256) ||
		(memb->queue == &memq512) ||
		(memb->queue == &memq1024)));

	assert((memb->b00f = (void *)0xbeef));

	/* put back to queue */
#ifdef STAT_ME_HARDER
	memb->queue->out--;
	assert((memb->queue->out >= 0));

	item->next = memb->queue->queue;
	memb->queue->queue = item;
#else
	item->next = *memb->queue;
	*memb->queue = item;
#endif

    } else {

	/* was large and so malloced, free it */
	sc_stats->mallocs_freed++;
	free(memb);
    }	
}



static struct sc_tree_node_t *
tree_node_alloc() {

    struct sc_tree_node_t *node;

    if (node_q) { /* try to get from queue */
	node = (struct sc_tree_node_t *)node_q;
	node_q = node_q->next;
    } else {
	node = (struct sc_tree_node_t *)
	    sc_obs_alloc(sizeof(struct sc_tree_node_t));

#ifndef NDEBUG
	node->b00b = (void *)0xbeeb;
#endif
    }


    assert((node->b00b == (void *)0xbeeb));
    assert((node->b00b = (void *)0xb00b));

    STAT_HARD(sc_stats->tree_nodes_created++);
    STAT_HARD(sc_stats->tree_nodes_out++);
    return node;
}

static void
tree_node_free(struct sc_tree_node_t *node) {

    assert((node));

    ((struct sc_list_t *)node)->next = node_q;
    node_q = (struct sc_list_t *)node;

#ifdef STAT_ME_HARDER
    STAT_HARD(sc_stats->tree_nodes_out--);
    STAT_HARD(sc_stats->tree_nodes_freed++);
    assert((sc_stats->tree_nodes_out >= 0));
#endif

    assert((node->b00b == (void *)0xb00b));
    assert((node->b00b = (void *)0xbeeb));
}


static struct sc_tree_t *
tree_entry_alloc() {

    struct sc_tree_t *tree;

    if (tree_q) {
	tree = (struct sc_tree_t *)tree_q;
	tree_q = tree_q->next;
    } else {

	tree = (struct sc_tree_t *)
	    sc_obs_alloc(sizeof(struct sc_tree_t));
#ifndef NDEBUG
	tree->b00b = (void *)0xbeeb;
#endif
    }

    assert((tree->b00b == (void *)0xbeeb));
    assert((tree->b00b = (void *)0xb00b));

    STAT_HARD(sc_stats->trees_out++);
    STAT_HARD(sc_stats->trees_created++);
    return tree;
}

static void
tree_entry_free(struct sc_tree_t *tree) {

    assert((tree));
    assert((tree->nil == &(tree->nil_node)));

    ((struct sc_list_t *)tree)->next = tree_q;
    tree_q = (struct sc_list_t *)tree;

#ifdef STAT_ME_HARDER
    STAT_HARD(sc_stats->trees_out--);
    STAT_HARD(sc_stats->trees_freed++);
    assert((sc_stats->trees_out >= 0));
#endif

    assert((tree->b00b == (void *)0xb00b));
    assert((tree->b00b = (void *)0xbeeb));
}



static void
tree_rotate_left(struct sc_tree_t *tree,
		 struct sc_tree_node_t *x) {

    struct sc_tree_node_t *t;

    t = x->right;
    x->right = t->left;

    if (t->left != tree->nil) t->left->parent = x;

    if (!(t->parent = x->parent)) {
        tree->root = t;
    } else {
        if (x->parent->left == x) x->parent->left = t;
        else x->parent->right = t;
    }
    t->left = x;
    x->parent = t;
}

static void
tree_rotate_right(struct sc_tree_t *tree,
		  struct sc_tree_node_t *x) {

    struct sc_tree_node_t *t;

    t = x->left;
    x->left = t->right;

    if (t->right != tree->nil) t->right->parent = x;

    if (!(t->parent = x->parent)) {
        tree->root = t;
    } else {
        if (x->parent->right == x) x->parent->right = t;
        else x->parent->left = t;
    }
    t->right = x;
    x->parent = t;
}


/* tree_insert_int
 *
 * op == 0, dont replace, when exists
 * op == 1, set or replace always
 * op == 2, replace only when exists
 *
 */

static void *
tree_insert_int(struct sc_tree_t *tree, 
		char *key, void *data,
		int op) {

    int cmp_val = 0;
    struct sc_tree_node_t *x = NULL;
    struct sc_tree_node_t *y = NULL;
    struct sc_tree_node_t *z = NULL;

    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));
    assert((tree));
    assert((key));

    if (!tree || !key) {
	syslog(LOG_ERR, "%s: internal error: null tree (%p) or key (%p)",
	       __FUNCTION__, tree, key);
	abort();
    }

    /* precheck */
    if (tree->root &&
	tree->root != tree->nil) {
	
	for(y = tree->root; y != tree->nil;) {

	    z = y;
	    cmp_val = strcmp(y->key, key);
	    
	    if (cmp_val > 0) y = y->left;
	    else if (cmp_val < 0) y = y->right;
	    else break;
	}
	
	if (y == tree->nil) {

	    if (op > 1) {
		/* set requested, but dont exist */
		return NULL;
	    }

	    /* insert */
	    x = tree_node_alloc();
	    x->left = x->right = tree->nil;
	    x->parent = z;

	    x->key = key;
	    x->data = data;

	    if (z && z != tree->nil) {
		if (cmp_val > 0) z->left = x;
		else z->right = x;
	    }

	} else {

	    if (op > 0) {

		/* replace */
		void *pd = y->data;

		sc_unref(y->key);
		sc_unref(y->data);
		y->key = key;
		y->data = data;
		return pd;

	    } else {

		/* add requested, but already exists */
		return y->data;
	    }
	}

    } else {

	/* first */
	if (op < 2) {
	    x = tree_node_alloc();
	    x->left = x->right = tree->nil;
	    
	    x->parent = NULL;
	    x->key = key;
	    x->data = data;
	    
	    tree->root = x;
	    tree->root->color = NODE_BLACK;
	}

	return NULL;
    }
    
    /* balance */
    x->color = NODE_RED;
    while(x != tree->root && 
	  x->parent->color == NODE_RED) {

	if (x->parent == x->parent->parent->left) {

	    y = x->parent->parent->right;
	    if (y->color == NODE_RED) {

		x->parent->color = NODE_BLACK;
		y->color = NODE_BLACK;
		x->parent->parent->color = NODE_RED;
		x = x->parent->parent;

	    } else {

		if (x == x->parent->right) {
		    x = x->parent;
		    tree_rotate_left(tree, x);
		}

		x->parent->color = NODE_BLACK;
		x->parent->parent->color = NODE_RED;
		tree_rotate_right(tree, x->parent->parent);
	    }

	} else {

	    y = x->parent->parent->left;
	    if (y->color == NODE_RED) {

		x->parent->color = NODE_BLACK;
		y->color = NODE_BLACK;
		x->parent->parent->color = NODE_RED;
		x = x->parent->parent;

	    } else {

		if (x == x->parent->left) {
		    x = x->parent;
		    tree_rotate_right(tree, x);
		}

		x->parent->color = NODE_BLACK;
		x->parent->parent->color = NODE_RED;
		tree_rotate_left(tree, x->parent->parent);
	    }
	}
    }
    
    tree->root->color = NODE_BLACK;
    return NULL;
}


/* INSERT ROUTINES
 *
 *
 * */

#define TREE_OPL(__FNAME, __K_OP, __D_OP, __REPL)	\
void *tree_ ## __FNAME(struct sc_tree_t *tree, \
		       char *key, void *data, int length) {	\
    return tree_insert_int(tree, __K_OP, __D_OP, __REPL); }

#define TREE_OPN(__FNAME, __K_OP, __D_OP, __REPL)	\
void *tree_ ## __FNAME(struct sc_tree_t *tree, \
		       char *key, void *data) {	\
    return tree_insert_int(tree, __K_OP, __D_OP, __REPL); }

TREE_OPL(add, sc_strdup(key), sc_memdup(data, length), 0);
TREE_OPL(set, sc_strdup(key), sc_memdup(data, length), 1);
TREE_OPL(repl, sc_strdup(key), sc_memdup(data, length), 2);

TREE_OPN(add_referable_d, sc_strdup(key), sc_ref(data), 0);
TREE_OPN(set_referable_d, sc_strdup(key), sc_ref(data), 1);
TREE_OPN(repl_referable_d, sc_strdup(key), sc_ref(data), 2);

TREE_OPL(add_referable_k, sc_ref(key), sc_memdup(data, length), 0);
TREE_OPL(set_referable_k, sc_ref(key), sc_memdup(data, length), 1);
TREE_OPL(repl_referable_k, sc_ref(key), sc_memdup(data, length), 2);

TREE_OPN(add_referable_kd, sc_ref(key), sc_ref(data), 0);
TREE_OPN(set_referable_kd, sc_ref(key), sc_ref(data), 1);
TREE_OPN(repl_referable_kd, sc_ref(key), sc_ref(data), 2);


void *
tree_remove(struct sc_tree_t *tree, const char *key) {

    void *rv = NULL;
    void *kv = NULL;

    struct sc_tree_node_t *w;
    struct sc_tree_node_t *x;
    struct sc_tree_node_t *y;
    struct sc_tree_node_t *z;
    
    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));

    /* precheck */
    if (!tree || !key) {
	syslog(LOG_ERR, "%s: internal error: null tree (%p) or key (%p)",
	       __FUNCTION__, tree, key);
	abort();
    }
    
    for(z = tree->root; z != tree->nil;) {
	int cmp_val = strcmp(z->key, key);
	if (cmp_val > 0) z = z->left;
	else if (cmp_val < 0) z = z->right;
	else break;
    }
    

    if (z == tree->nil) {
	return NULL; /* not found */
    } else {
	rv = (void *)z->data;
	kv = (void *)z->key;
    }
 
    if (z->left == tree->nil ||	z->right == tree->nil) {
	y = z;
    } else {
	y = z->right;
	while(y->left != tree->nil) y = y->left;
    }
    
    if (y->left != tree->nil) x = y->left; 
    else x = y->right;
    
    if ((x->parent = y->parent)) {

	if (y == y->parent->left) y->parent->left = x;
	else y->parent->right = x;

    } else {

	tree->root = x;
	tree->root->color = NODE_BLACK;
	
	if (tree->root == tree->nil) tree->root = NULL;

	sc_unref(kv);
	sc_unref(rv);
	tree_node_free(y);
	return rv;
    }

    if (y != z) {
	z->key = y->key;
	z->data = y->data;
    }
    
    /* balance */
    if (y->color == NODE_BLACK) {
	while (x != tree->root && x->color != NODE_RED) {

	    if (x->parent->left == x) {

		w = x->parent->right;
		if (w->color == NODE_RED) {
		    w->color = NODE_BLACK;
		    x->parent->color = NODE_RED;
		    tree_rotate_left(tree, x->parent);
		    w = x->parent->right;
		}

		if (w->left->color == NODE_BLACK && 

		    w->right->color == NODE_BLACK) {
		    w->color = NODE_RED;
		    x = x->parent;

		} else {

		    if (w->right->color == NODE_BLACK) {
			w->left->color = NODE_BLACK;
			w->color = NODE_RED;
			tree_rotate_right(tree, w);
			w = x->parent->right;
		    }

		    w->color = w->parent->color;
		    x->parent->color = NODE_BLACK;
		    
		    w->right->color = NODE_BLACK;
		    tree_rotate_left(tree, x->parent);
		    x = tree->root;
		}
	    } else {

		w = x->parent->left;
		if (w->color == NODE_RED) {
		    w->color = NODE_BLACK;
		    x->parent->color = NODE_RED;
		    tree_rotate_right(tree, x->parent);
		    w = x->parent->left;
		}

		if (w->right->color == NODE_BLACK && 

		    w->left->color == NODE_BLACK) {
		    w->color = NODE_RED;
		    x = x->parent;

		} else {

		    if (w->left->color == NODE_BLACK) {
			w->right->color = NODE_BLACK;
			w->color = NODE_RED;
			tree_rotate_left(tree, w);
			w = x->parent->left;
		    }

		    w->color = x->parent->color;
		    x->parent->color = NODE_BLACK;
		    
		    w->left->color = NODE_BLACK;
		    tree_rotate_right(tree, x->parent);
		    x = tree->root;
		}
	    }
	}
	x->color = NODE_BLACK;
    }
    
    /* palautus */
    sc_unref(kv);
    sc_unref(rv);
    tree_node_free(y);
    return rv;
}


void *
tree_fetch(struct sc_tree_t *tree, const char *key) {
    
    struct sc_tree_node_t *z;
    
    assert((tree));
    assert((key));
    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));

    /* precheck */
    if (!tree || !key) {
	syslog(LOG_ERR, "%s: internal error: null tree (%p) or key (%p)",
	       __FUNCTION__, tree, key);
	abort();
    }

    for(z = tree->root; z && z != tree->nil;) {
	int cmp_val = strcmp(z->key, key);
	
	if (cmp_val > 0) z = z->left;
	else if (cmp_val < 0) z = z->right;
	else break;
    }
    
    if (z && z != tree->nil) {
	return (void *)z->data;
    }

    /* else */
    return NULL;
}

void *
tree_fetch_both(struct sc_tree_t *tree, const char *key,
		char **keyslot, void *dataslot) {
    
    struct sc_tree_node_t *z;
    
    assert((tree));
    assert((key));
    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));

    /* precheck */
    if (!tree || !key) {
	syslog(LOG_ERR, "%s: internal error: null tree (%p) or key (%p)",
	       __FUNCTION__, tree, key);
	abort();
    }

    for(z = tree->root; z && z != tree->nil;) {
	int cmp_val = strcmp(z->key, key);
	
	if (cmp_val > 0) z = z->left;
	else if (cmp_val < 0) z = z->right;
	else break;
    }
    
    if (z && z != tree->nil) {

	if (keyslot) *keyslot = z->key;
	if (dataslot) *(void **)dataslot = z->data;
	return (void *)z->data;
    }

    /* else */
    return NULL;
}



struct sc_tree_t *
tree_create() {
    
    struct sc_tree_t *new_tree;
    
    /* fits in sc_tree_node_t... */
    new_tree = tree_entry_alloc();

    /* nil-node */
    new_tree->nil = &(new_tree->nil_node);
    new_tree->nil_node.color = NODE_BLACK;
    new_tree->root =
	new_tree->nil_node.right =
	new_tree->nil_node.left =
	new_tree->nil_node.parent = NULL;

    return new_tree;
}

static void
tree_enum_asc_int(struct sc_tree_t *tree,
		  struct sc_tree_node_t *node,
		  void (*cb)(const char *, void *, void *),
		  void *userdata) {

    if (node->left != tree->nil) 
	tree_enum_asc_int(tree, node->left, cb, userdata);

    (*cb)(node->key, node->data, userdata);

    if (node->right != tree->nil)
	tree_enum_asc_int(tree, node->right, cb, userdata);
}

static void
tree_enum_desc_int(struct sc_tree_t *tree,
		   struct sc_tree_node_t *node,
		   void (*cb)(const char *, void *, void *),
		   void *userdata) {

    if (node->right != tree->nil) 
	tree_enum_asc_int(tree, node->right, cb, userdata);

    (*cb)(node->key, node->data, userdata);

    if (node->left != tree->nil)
	tree_enum_asc_int(tree, node->left, cb, userdata);
}

void
tree_enum_asc(struct sc_tree_t *tree,
	      void (*cb)(const char *, void *, void *),
	      void *userdata) {

    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));
    tree_enum_asc_int(tree, tree->root, cb, userdata);
}

void
tree_enum_desc(struct sc_tree_t *tree,
	       void (*cb)(const char *, void *, void *),
	       void *userdata) {

    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));
    tree_enum_desc_int(tree, tree->root, cb, userdata);
}


static void
tree_destroy_int(struct sc_tree_t *tree,
		 struct sc_tree_node_t *node) {


    if (node->left != tree->nil) 
	tree_destroy_int(tree, node->left);
    if (node->right != tree->nil) 
	tree_destroy_int(tree, node->right);

    sc_unref(node->key);
    sc_unref(node->data);
    if (node != tree->nil) {
	tree_node_free(node);
    }
}

void
tree_destroy(struct sc_tree_t *tree, 
	     void (*cb)(const char *, void *, void *),
	     void *userdata) {

    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));

    if (tree->root &&
	tree->root != tree->nil) {

	if (cb) tree_enum_asc_int(tree, tree->root, cb, userdata);
	tree_destroy_int(tree, tree->root);
    }

    tree_entry_free(tree);
}

static inline void
tree_destroy_unref_r(struct sc_tree_t *tree,
		     struct sc_tree_node_t *node) {

    if (node->left != tree->nil) tree_destroy_unref_r(tree, node->left);
    if (node->right != tree->nil) tree_destroy_unref_r(tree, node->right);

    sc_unref(node->key);
    sc_unref(node->data);
    if (node != tree->nil) tree_node_free(node);
}

void
tree_destroy_unref(struct sc_tree_t *tree) {

    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));

    if (tree->root && tree->root != tree->nil) 
	tree_destroy_unref_r(tree, tree->root);
    
    tree_entry_free(tree);
}

int
tree_isempty(struct sc_tree_t *tree) {
    assert((tree));
    assert((((struct sc_tree_t *)tree)->b00b == (void *)0xb00b));

    return (tree->root == NULL);
}

int
common_init(enum sc_bootstages_t stage) {

    /* obstack */
    if (stage == SC_STAGE_PREPARE) {
	memset(&obs, 0, sizeof(struct obstack));
	obstack_init(&obs);
	obstack_chunk_size(&obs) = OBSTACK_CHUNK_SIZE;
    }

    return 0;
}


/*
  Local variables: ***
  mode: c ***
  c-basic-offset: 4 ***
  End: ***
*/

