/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */


#include "scache.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <syslog.h>
#include <pwd.h>
#include <grp.h>

#include <regex.h>

#include "ext/php-scache-proto.h"


#define SC_TOKENBUF_LEN        256
#define SC_UNLIMITED_LIMIT     ((int64_t)1 << 40) /* 1TiB */
//#define SC_DEFAULT_CACHE_LIMIT (1024 * 1024)

#define SC_MIN_OBSTACK_CHUNK   8191
#define SC_MIN_READ_BUFFERLEN  2048
#define SC_MIN_SOCK_SNDBUF     4096
#define SC_MIN_SOCK_RCVBUF     4096
#define SC_MIN_LISTEN_BACKLOG  0
#define SC_MIN_LISTEN_PORT     1
#define SC_MIN_LOGLEVEL        0

#define SC_MAX_OBSTACK_CHUNK   0
#define SC_MAX_READ_BUFFERLEN  0
#define SC_MAX_SOCK_SNDBUF     0
#define SC_MAX_SOCK_RCVBUF     0
#define SC_MAX_LISTEN_BACKLOG  0
#define SC_MAX_LISTEN_PORT     65535
#define SC_MAX_LOGLEVEL        255


/* define */
struct sc_limitgroup_t {

    char *name;

    int allow_ring;
    int allow_cache;
    int allow_lvals;
    int allow_stats;
    int allow_shared;
    int allow_session;
    int allow_collect;
    int allow_shcollect;

#ifdef SESSION_MEMORY_ACCOUNTING
    int64_t memory_limit;
#endif
#ifdef SESSION_NODES_ACCOUNTING
    int64_t nodes_limit;
#endif
    int64_t cache_limit;

    int64_t ring_nodes_limit;
    int64_t ring_memory_limit;

    struct sc_partition_t *partition;
};

struct sc_assign_t {

    regex_t re;
    char *string;

    struct sc_limitgroup_t *group;
    struct sc_partition_t *partition;
    struct sc_assign_t *next;
};




/* static */
static int                     current_line = 1;
static struct sc_tree_t       *groups;
static struct sc_assign_t     *host_assigns;
static struct sc_assign_t     *auth_assigns;
static struct sc_config_t      config = {
    
    .progname            = SC_CONF_PROGNAME,
    .config_path         = SC_CONF_CONFIG_PATH,
    .pidfile_path        = SC_CONF_PIDFILE_PATH,
    .usock_path          = SC_CONF_USOCK_PATH,

    .debug               = SC_CONF_DEBUG,
    .nodetach            = SC_CONF_NODETACH,
    .loglevel            = SC_CONF_LOGLEVEL,
    .syslog_fac          = SC_CONF_SYSLOG_FAC,
    .syslog_tag          = SC_CONF_SYSLOG_TAG,


    .obstack_chunk       = SC_CONF_OBSTACK_CHUNK,
    .read_bufferlen      = SC_CONF_READ_BUFFERLEN,
    .expire_delay        = SC_CONF_EXPIRE_DELAY,

    .sock_rcvbuf         = SC_CONF_SOCK_RCVBUF,
    .sock_sndbuf         = SC_CONF_SOCK_SNDBUF,

    .max_reqsize         = SC_CONF_MAX_REQSIZE,

    .usock_mask          = SC_CONF_USOCK_MASK,
    .listen_backlog      = SC_CONF_LISTEN_BACKLOG,
    .listen_port         = SCACHE_PTO_LISTEN_PORT,

#ifdef SESSION_MEMORY_ACCOUNTING
    .defpart_mlimit      = SC_CONF_DEFPART_MLIMIT,
#endif
#ifdef SESSION_NODES_ACCOUNTING
    .defpart_nlimit      = SC_CONF_DEFPART_NLIMIT,
#endif
    .defpart_ring        = SC_CONF_DEFPART_ALLOW_RING,
    .defpart_lvals       = SC_CONF_DEFPART_ALLOW_LVALS,
    .defpart_stats       = SC_CONF_DEFPART_ALLOW_STATS,
    .defpart_shared      = SC_CONF_DEFPART_ALLOW_SHARED,
    .defpart_session     = SC_CONF_DEFPART_ALLOW_SESSION,
    .defpart_collect     = SC_CONF_DEFPART_ALLOW_COLLECT,
    .defpart_shcollect   = SC_CONF_DEFPART_ALLOW_SHCOLLECT,
    .defpart_cache       = SC_CONF_DEFPART_ALLOW_CACHE,
    .defpart_cachemax    = SC_CONF_DEFPART_CACHEMAX,
    .defpart_ringnodeslimit    = SC_CONF_DEFPART_RINGNODESLIMIT,
    .defpart_ringmemorylimit   = SC_CONF_DEFPART_RINGMEMORYLIMIT,

    .enable_auth_partitions = SC_CONF_ENABLE_AUTHPARTITION,
    .enable_host_partitions = SC_CONF_ENABLE_HOSTPARTITION,
};

/* config-temp */
static const char *temp_runas_user       = SC_CONF_RUNAS_USER;
static const char *temp_runas_group      = SC_CONF_RUNAS_GROUP;
static const char *temp_usock_owner      = SC_CONF_USOCK_OWNER;
static const char *temp_usock_group      = SC_CONF_USOCK_GROUP;

/* extern */
const struct sc_config_t   *sc_config            = &config;



char *config_dump_stats(struct sc_session_t *session) {

    int i,alen;
    char *cp, buffer[16384]; /* reserve enought for sprintf */
    struct {
	const char *name;
	int64_t *value;
    } stat_64s[] = {
	{ "sessiontable_max", &(sc_stats->sessiontable_max) },
	{ "mallocs_allocated", &(sc_stats->mallocs_allocated) },
	{ "mallocs_freed", &(sc_stats->mallocs_freed) },
	{ "mallocs_total", &(sc_stats->mallocs_total) },
	{ "req_count_success", &(sc_stats->req_count_success) },
	{ "req_count_invalid", &(sc_stats->req_count_invalid) },
	{ "req_count_unknown", &(sc_stats->req_count_unknown) },
	{ "session_creates", &(sc_stats->session_creates) },
	{ "session_reuses", &(sc_stats->session_reuses) },
	{ "session_destroys", &(sc_stats->session_destroys) },
	{ "session_expires", &(sc_stats->session_expires) },
	{ "session_clears", &(sc_stats->session_clears) },
	{ "session_hit_by_idx", &(sc_stats->session_hit_by_idx) },
	{ "session_miss_by_idx", &(sc_stats->session_miss_by_idx) },
	{ "session_miss_by_search", &(sc_stats->session_miss_by_search) },
	{ "session_secret_mismatches", &(sc_stats->session_secret_mismatches) },
	{ "all_cache_hits", &(sc_stats->all_cache_hits) },
	{ "all_cache_misses", &(sc_stats->all_cache_misses) },
	{ "all_cache_adds", &(sc_stats->all_cache_adds) },
	{ "all_cache_dels", &(sc_stats->all_cache_dels) },
	{ "all_cache_expires", &(sc_stats->all_cache_expires) },
	{ "io_requests", &(sc_stats->io_requests) },
	{ "io_bytes_read", &(sc_stats->io_bytes_read) },
	{ "io_bytes_written", &(sc_stats->io_bytes_written) },
	{ "io_socks_accepted", &(sc_stats->io_socks_accepted) },
	{ "io_socks_accept_fails", &(sc_stats->io_socks_accept_fails) },
	{ "io_socks_closed", &(sc_stats->io_socks_closed) }

#ifdef STAT_ME_HARDER
	,
	{ "io_polls", &(sc_stats->io_polls) },
	{ "req_t_getopts", &(sc_stats->req_t_getopts) },
	{ "req_t_setopts", &(sc_stats->req_t_setopts) },
	{ "req_t_srvstats", &(sc_stats->req_t_srvstats) },
	{ "req_t_adds", &(sc_stats->req_t_adds) },
	{ "req_t_sets", &(sc_stats->req_t_sets) },
	{ "req_t_replaces", &(sc_stats->req_t_replaces) },
	{ "req_t_unsets", &(sc_stats->req_t_unsets) },
	{ "req_t_stats", &(sc_stats->req_t_stats) },
	{ "req_t_colls", &(sc_stats->req_t_colls) },
	{ "req_t_gets", &(sc_stats->req_t_gets) },
	{ "req_t_shadds", &(sc_stats->req_t_shadds) },
	{ "req_t_shsets", &(sc_stats->req_t_shsets) },
	{ "req_t_shreplaces", &(sc_stats->req_t_shreplaces) },
	{ "req_t_shunsets", &(sc_stats->req_t_shunsets) },
	{ "req_t_shstats", &(sc_stats->req_t_shstats) },
	{ "req_t_shcolls", &(sc_stats->req_t_shcolls) },
	{ "req_t_shgets", &(sc_stats->req_t_shgets) },
	{ "req_t_chgets", &(sc_stats->req_t_chgets) },
	{ "req_t_chputs", &(sc_stats->req_t_chputs) },
	{ "req_t_chclears", &(sc_stats->req_t_chclears) },
	{ "req_t_vadds", &(sc_stats->req_t_vadds) },
	{ "req_t_vsubs", &(sc_stats->req_t_vsubs) },
	{ "req_t_vgets", &(sc_stats->req_t_vgets) },
	{ "req_t_vsets", &(sc_stats->req_t_vsets) },
	{ "req_t_vunsets", &(sc_stats->req_t_vunsets) },
	{ "req_t_vbrks", &(sc_stats->req_t_vbrks) },
	{ "req_r_rngets", &(sc_stats->req_t_rngets) },
	{ "req_r_rnsizes", &(sc_stats->req_t_rnsizes) },
	{ "req_r_rnclears", &(sc_stats->req_t_rnclears) },
	{ "req_r_rnpushes", &(sc_stats->req_t_rnpushes) },
	{ "req_r_rnpops", &(sc_stats->req_t_rnpops) },
	{ "req_r_rnshifts", &(sc_stats->req_t_rnshifts) },
	{ "req_r_rnunshifts", &(sc_stats->req_t_rnunshifts) },
	{ "req_r_rnrotfs", &(sc_stats->req_t_rnrotfs) },
	{ "req_r_rnrotbs", &(sc_stats->req_t_rnrotbs) },
	{ "req_t_shdlrgets", &(sc_stats->req_t_shdlrgets) },
	{ "req_t_shdlrputs", &(sc_stats->req_t_shdlrputs) },
	{ "req_t_shdlrdrops", &(sc_stats->req_t_shdlrdrops) },
	{ "req_t_creates", &(sc_stats->req_t_creates) },
	{ "req_t_renames", &(sc_stats->req_t_renames) },
	{ "req_t_drops", &(sc_stats->req_t_drops) },
	{ "req_t_lasterrs", &(sc_stats->req_t_lasterrs) },
	{ "req_t_nones", &(sc_stats->req_t_nones) },
	{ "req_t_unknowns", &(sc_stats->req_t_unknowns) },
	{ "wblock_allocs", &(sc_stats->wblock_allocs) },
	{ "trees_created", &(sc_stats->trees_created) },
	{ "trees_freed", &(sc_stats->trees_freed) },
	{ "trees_out", &(sc_stats->trees_out) },
	{ "tree_nodes_created", &(sc_stats->tree_nodes_created) },
	{ "tree_nodes_freed", &(sc_stats->tree_nodes_freed) },
	{ "tree_nodes_out", &(sc_stats->tree_nodes_out) },
#endif
    };

    struct {
	const char *name;
	int64_t *value;
    } part_64s[] = {
	{ "cache_hits", &(session->partition->cache_hits) },
	{ "cache_misses", &(session->partition->cache_misses) },
	{ "cache_adds", &(session->partition->cache_adds) },
	{ "cache_dels", &(session->partition->cache_dels) },
	{ "cache_expires", &(session->partition->cache_expires) },
	{ "cache_max", &(session->partition->cache_max) }

	/* TODO, cache and ring might be unused and uninitialized
	   since causing crash. partition should be initialized as 
	   complete 
	,
	{ "cache_memory", &(session->partition->cache->cur_size) },
	{ "ring_nodes", &(session->partition->ring->cur_nodes) },
	{ "ring_memory", &(sesson->partition->ring->cur_memory) }
	*/
#ifdef SESSION_MEMORY_ACCOUNTING
	,
	{ "memory_allocated", &(session->partition->memory_allocated) },
	{ "memory_limit", &(session->partition->memory_limit) }
#endif
#ifdef SESSION_NODES_ACCOUNTING
	,
	{ "nodes_allocated", &(session->partition->nodes_allocated) },
	{ "nodes_limit", &(session->partition->nodes_limit) }
#endif
    };


#ifdef STAT_ME_HARDER
    struct sc_memqueue_t *mqs[] = {
	&memq16,
	&memq32,
	&memq64,
	&memq128,
	&memq256,
	&memq512,
	&memq1024
    };
#endif


    /* fake array */
    alen = sc_asize(stat_64s);
#ifdef STAT_ME_HARDER
    alen += sc_asize(mqs) /* memqueues */ * 3;
#endif

    //cp = buffer + sprintf(buffer, "a:%u:{", alen);
    cp = buffer + sprintf(buffer, "a:3:{s:6:\"SERVER\";a:%u:{", 
			  alen);

    /* SERVER */
    for(i = 0; i < sc_asize(stat_64s); i++) {
	cp += sprintf(cp, 
		      "s:%u:\"%s\";d:%llu;",

		      (int)strlen(stat_64s[i].name),
		      stat_64s[i].name, 
		      (long long unsigned int)*stat_64s[i].value);
    }

#ifdef STAT_ME_HARDER
    /* memqueues */
    for(i = 0; i < sc_asize(mqs); i++) {
	cp += sprintf(cp, 
		      "s:%u:\"memq%s_out\";d:%lli;"
		      "s:%u:\"memq%s_total\";d:%lli;"
		      "s:%u:\"memq%s_allocs\";d:%lli;",

		      (int)strlen(mqs[i]->token) + 8,
		      mqs[i]->token,
		      (long long int)mqs[i]->out,

		      (int)strlen(mqs[i]->token) + 10,
		      mqs[i]->token,
		      (long long int)mqs[i]->total,

		      (int)strlen(mqs[i]->token) + 11,
		      mqs[i]->token,
		      (long long int)mqs[i]->total);
    }

#endif

    /* PART */
    alen = sc_asize(part_64s);
    cp += sprintf(cp,  
		  "}s:9:\"PARTITION\";a:%u:{"
		  "s:10:\"allow_ring\";b:%u;"
		  "s:11:\"allow_lvals\";b:%u;"
		  "s:11:\"allow_stats\";b:%u;"
		  "s:11:\"allow_cache\";b:%u;"
		  "s:12:\"allow_shared\";b:%u;"
		  "s:13:\"allow_session\";b:%u;"
		  "s:13:\"allow_collect\";b:%u;"
		  "s:15:\"allow_shcollect\";b:%u;"
		  ,
		  alen + 8 /* +these */,
		  session->partition->allow_ring ? 1 : 0,
		  session->partition->allow_lvals ? 1 : 0,
		  session->partition->allow_stats ? 1 : 0,
		  session->partition->allow_cache ? 1 : 0,
		  session->partition->allow_shared ? 1 : 0,
		  session->partition->allow_session ? 1 : 0,
		  session->partition->allow_collect ? 1 : 0,
		  session->partition->allow_shcollect ? 1 : 0);
    

    for(i = 0; i < sc_asize(part_64s); i++) {
	cp += sprintf(cp, 
		      "s:%u:\"%s\";d:%llu;",

		      (int)strlen(part_64s[i].name),
		      part_64s[i].name,
		      (long long unsigned int)*part_64s[i].value);
    }


    /* SESSION */
    alen = 3;
#ifdef SESSION_NODES_ACCOUNTING
    alen++;
#endif
#ifdef SESSION_MEMORY_ACCOUNTING
    alen++;
#endif
    cp += sprintf(cp,  
		  "}s:7:\"SESSION\";a:%u:{"
		  "s:5:\"index\";i:%u;"
		  "s:7:\"expires\";i:%u;"
		  "s:7:\"maxidle\";i:%u;"
#ifdef SESSION_NODES_ACCOUNTING
		  "s:15:\"nodes_allocated\";d:%llu;"
#endif
#ifdef SESSION_MEMORY_ACCOUNTING
		  "s:16:\"memory_allocated\";d:%llu;"
#endif
		  ,
		  alen /* arraysize */,
		  (unsigned int)session->idx,
		  (unsigned int)session->expires,
		  (unsigned int)session->maxidle
#ifdef SESSION_NODES_ACCOUNTING
		  ,
		  (long long unsigned int)session->nodes_allocated
#endif
#ifdef SESSION_MEMORY_ACCOUNTING
		  ,
		  (long long unsigned int)session->memory_allocated
#endif
		  );

    strcat(cp, "}}\n");
    return sc_strdup(buffer);
}



struct sc_partition_t *
config_create_partition(struct sc_limitgroup_t *limit) {

    struct sc_partition_t *partition;

    partition = sc_alloc(sizeof(struct sc_partition_t));
    memset(partition, 0, sizeof(struct sc_partition_t));

    partition->id = limit->name;
    partition->allow_ring = limit->allow_ring;
    partition->allow_lvals = limit->allow_lvals;
    partition->allow_cache = limit->allow_cache;
    partition->allow_stats = limit->allow_stats;
    partition->allow_shared = limit->allow_shared;
    partition->allow_session = limit->allow_session;
    partition->allow_collect = limit->allow_collect;
    partition->allow_shcollect = limit->allow_shcollect;
    partition->cache_max = limit->cache_limit;
    partition->ring_nodes_limit = limit->ring_nodes_limit;
    partition->ring_memory_limit = limit->ring_memory_limit;
	
	/* accounting */
#ifdef SESSION_MEMORY_ACCOUNTING
    partition->memory_limit = limit->memory_limit;
#endif
    
#ifdef SESSION_NODES_ACCOUNTING
    partition->nodes_limit = limit->nodes_limit;
#endif
    
#ifdef SEPARATE_SESSIONTREES
    partition->session_tree = tree_create();
#endif

    return partition;
}

struct sc_partition_t *
config_get_default_partition() {

    static struct sc_partition_t *partition = NULL;

    if (!partition) { /* once */

	struct sc_limitgroup_t limit = {
	    
	    .name = "_DEFAULT_PARTITION_",
	    .allow_ring = sc_config->defpart_ring,
	    .allow_lvals = sc_config->defpart_lvals,
	    .allow_stats = sc_config->defpart_stats,
	    .allow_cache = sc_config->defpart_cache,
	    .allow_shared = sc_config->defpart_shared,
	    .allow_session = sc_config->defpart_session,
	    .allow_collect = sc_config->defpart_collect,
	    .allow_shcollect = sc_config->defpart_shcollect,

#ifdef SESSION_MEMORY_ACCOUNTING
	    .memory_limit = sc_config->defpart_mlimit,
#endif
#ifdef SESSION_NODES_ACCOUNTING
	    .nodes_limit = sc_config->defpart_nlimit,
#endif
	    .cache_limit = sc_config->defpart_cachemax,
	    .ring_nodes_limit = sc_config->defpart_ringnodeslimit,
	    .ring_memory_limit = sc_config->defpart_ringmemorylimit
	};

	partition = config_create_partition(&limit);
    }

    return partition;
}

struct sc_partition_t *
config_get_unlimited_partition() {

    static struct sc_partition_t *partition = NULL;

    if (!partition) {

	partition = sc_alloc(sizeof(struct sc_partition_t));
	memset(partition, 0, sizeof(struct sc_partition_t));
	
	partition->id = sc_strdup(SCACHE_PTO_UNKNOWN_CLIENT);
	partition->allow_shcollect = sc_config->defpart_shcollect;
	
	/* accounting */
#ifdef SESSION_MEMORY_ACCOUNTING
	partition->memory_limit = SC_UNLIMITED_LIMIT;
#endif
#ifdef SESSION_NODES_ACCOUNTING
	partition->nodes_limit = SC_UNLIMITED_LIMIT;
#endif
    }

    return partition;
}


struct sc_partition_t *
config_get_hostinfo(const char *id) {

    struct sc_assign_t *queue = host_assigns;

    while(queue) {

	if (!regexec(&queue->re, id, 0, NULL, 0)) {

	    if (!queue->partition) {
		queue->partition = config_create_partition(queue->group);
	    }
	    return queue->partition;
	}

	queue = queue->next;
    }

    return NULL;
}

struct sc_partition_t *
config_get_authinfo(const char *id) {

    struct sc_assign_t *queue = auth_assigns;

    while(queue) {

	if (!regexec(&queue->re, id, 0, NULL, 0)) {
	    if (!queue->partition) {
		queue->partition = config_create_partition(queue->group);
	    }
	    return queue->partition;
	}

	queue = queue->next;
    }

    return NULL;
}



/* =================================================
 * config_parse_resolve_path()
 *
 * enumerate command line once to resolve config files location
 * */
static int
config_parse_resolve_path() {

    int i, argc;
    char **argv, *config_path = NULL;

    argc = sc_config->program_argc;
    argv = sc_config->program_argv;

    for(i = 1; i < argc; i++) {

        if (!strncmp(argv[i], "--config=", 9)) {

            config_path = argv[i] + 9;
            if (!*config_path) {
                fprintf(stderr, "missing argument for --config\n");
                return -1;
            }

        } else if (!strcmp(argv[i], "--config") ||
                   !strcmp(argv[i], "-C")) {

            if (++i < argc) {
                config_path = argv[i];
            } else {
                fprintf(stderr, "missing argument for --config\n");
                return -1;
            }

        } else if (!strcmp(argv[i], "--")) {
            break;
        }
    }

    if (config_path) {

        struct stat st;
        if (stat(config_path, &st) < 0) {
            fprintf(stderr, "can't stat config file: %s\n",
                    config_path);
            return -1;
        }

        config.config_path = strdup(config_path);
    }

    return 0;
}


static void
config_error(const char *format, ...) {

    va_list ap;

    va_start(ap, format);
    fprintf(stderr, "\nError in configuration file on line %u:\n", 
	    current_line);
    vfprintf(stderr, format, ap);
    fprintf(stderr, "\n");
    va_end(ap);
}

static const char *
config_extract_token(FILE *file, char *sbuffer) {

    int c, sbpos = 0;

    /* sbuffer is assumed to be always SC_TOKENBUF_LEN */
    sbuffer[0] = 0;
    while((c = fgetc(file)) && (c != EOF)) {

	if (c == (int)'\n') current_line++;
	if ((isspace(c) || iscntrl(c)) && sbuffer[0] != '"') {
	    if (sbpos) return sbuffer;
	    continue;
	}
	
	if (c == (int)'"') {
	    if (sbuffer[0] == '"' && /* TODO: \\" */
		sbuffer[sbpos - 1] != '\\') return sbuffer + 1;

	    /* else */
	    sbuffer[sbpos++] = '"';
	    sbuffer[sbpos] = 0;
	    continue;
	}

	if (c == (int)'#') {

	    do {
		c = fgetc(file);
	    } while(c != (int)'\n' &&
		    c != (int)'\r' &&
		    c != EOF);

	    continue;
	}

	sbuffer[sbpos++] = (char)c;
	sbuffer[sbpos] = 0;

	if (sbpos > (SC_TOKENBUF_LEN - 1)) {
	    config_error("buffer overflow when parsing config file");
	    return NULL;
	}
    }

    sbuffer[sbpos] = 0;
    return sbuffer;
}

static const char *
config_require_value(FILE *file, char *buffer, const char *param) {

    const char *token = config_extract_token(file, buffer);
    if (!token || !*token) {
	config_error("parameter '%s' requires value, none given",
		     param);
	return NULL;
    }

    return token;
}

static int64_t
config_get_int(FILE *file, const char *token,
	       int64_t min_val, int64_t max_val) {

    int64_t i, m;
    char *endp, buffer[SC_TOKENBUF_LEN];
    const char *ival;

    ival = config_require_value(file, buffer, token);
    if (!ival) return -1;

    /* whether kilo, mega, giga, block -suffix given */
    i = strlen(buffer);
    switch((buffer[i - 1]) | 0x20) {
    case 'b': m = 512; break;
    case 'k': m = 1024; break;
    case 'm': m = 1024 * 1024; break;
    case 'g': m = 1024 * 1024 * 1024; break;
    default: m = 1; break;
    }
    if (m > 1) buffer[i - 1] = 0;

    i = strtol(ival, &endp, 0);
    if (*endp != 0) { /* invalid value */
	config_error("invalid numeric value for parameter '%s', "
		     "got '%s'", token, ival);
	return -1;
    }

    i = i * m;
    if (min_val && (i < min_val)) {
	config_error("value %lu for parameter '%s' is below minimum, "
		     "using minimum %lu",
		     (unsigned long)i, token, (unsigned long)min_val);
	return min_val;
    }

    if (max_val && (i > max_val)) {
	config_error("value %lu for parameter '%s' is over maximum, "
		     "using maximum %lu",
		     (unsigned long)i, token, (unsigned long)min_val);
	return max_val;
    }

    return i;
}

static int
config_get_bool(FILE *file, const char *token) {

    char buffer[SC_TOKENBUF_LEN];
    const char *bval;

    bval = config_require_value(file, buffer, token);
    if (!bval) return -1;

    if (!strcasecmp(bval, "true") ||
	!strcasecmp(bval, "yes") ||
	!strcasecmp(bval, "1") ||
	!strcasecmp(bval, "on") ||
	!strcasecmp(bval, "enable") ||
	!strcasecmp(bval, "enabled")) {

	return 1;

    } else if (!strcasecmp(bval, "false") ||
	       !strcasecmp(bval, "no") ||
	       !strcasecmp(bval, "0") ||
	       !strcasecmp(bval, "off") ||
	       !strcasecmp(bval, "disable") ||
	       !strcasecmp(bval, "disabled")) {

	return 0;
	
    } 

    /* else */
    config_error("parameter '%s' expects boolean value, got '%s'", 
		 token, bval);
    return -1;
}


static int
config_collect_group(FILE *file, const char *name) {

    int i;
    char buffer[SC_TOKENBUF_LEN];
    const char *token;
    struct sc_limitgroup_t *gp, group = {
	.allow_ring = sc_config->defpart_ring,
	.allow_lvals = sc_config->defpart_lvals,
	.allow_cache = sc_config->defpart_cache,
	.allow_stats = sc_config->defpart_stats,
	.allow_shared = sc_config->defpart_shared,
	.allow_session = sc_config->defpart_session,
	.allow_collect = sc_config->defpart_collect,
	.allow_shcollect = sc_config->defpart_shcollect,

#ifdef SESSION_MEMORY_ACCOUNTING
	.memory_limit = sc_config->defpart_mlimit,
#endif
#ifdef SESSION_NODES_ACCOUNTING
	.nodes_limit = sc_config->defpart_nlimit,
#endif
	.cache_limit = sc_config->defpart_cachemax,
	.ring_nodes_limit = sc_config->defpart_ringnodeslimit,
	.ring_memory_limit = sc_config->defpart_ringmemorylimit
    };

    /* grouped */
    struct {
	const char *name;
	int64_t *value;
	int64_t min_val;
	int64_t max_val;
    } ints[] = {
#ifdef SESSION_MEMORY_ACCOUNTING
	{ "memory-limit", &(group.memory_limit),
	  0, SC_UNLIMITED_LIMIT },
#endif
#ifdef SESSION_NODES_ACCOUNTING
	{ "nodes-limit",  &(group.nodes_limit),
	  0, SC_UNLIMITED_LIMIT },
#endif
	{ "cache-limit",  &(group.cache_limit),
	  0, SC_UNLIMITED_LIMIT },
	{ "ring-nodes-limit", &(group.ring_nodes_limit),
	  0, SC_UNLIMITED_LIMIT },
	{ "ring-memory-limit", &(group.ring_memory_limit),
	  0, SC_UNLIMITED_LIMIT },
    };

    struct {
	const char *name;
	int *value;
    } bools[] = {
	{ "allow-ring", &(group.allow_ring) },
	{ "allow-cache", &(group.allow_cache) },
	{ "allow-lvals", &(group.allow_lvals) },
	{ "allow-stats", &(group.allow_stats) },
	{ "allow-shared", &(group.allow_shared) },
	{ "allow-session", &(group.allow_session) },
	{ "allow-collect", &(group.allow_collect) },
	{ "allow-shcollect", &(group.allow_shcollect) }
    };

    token = config_extract_token(file, buffer);
    if (!token || strcmp(token, "{")) {

	config_error("invalid configuration file:."
		     "  '{' expected in partition definition '%s'",
		     name);
	return -1;
    }

    while((token = config_extract_token(file, buffer)) && 
	  (token && *token)) {

	for(i = 0; i < sc_asize(ints); i++) {
	    if (!strcmp(token, ints[i].name)) {
		*ints[i].value =
		    config_get_int(file, ints[i].name,
				   ints[i].min_val, ints[i].max_val);
		goto next;
	    }
	}

	for(i = 0; i < sc_asize(bools); i++) {
	    if (!strcmp(token, bools[i].name)) {
		*bools[i].value =
		    config_get_bool(file, bools[i].name);

		if (*bools[i].value < 0) return -1;
		goto next;
	    }
	}

	if (!strcmp(token, "}")) {

	    break;
	} else {
	    config_error("unknown token '%s' in definition of "
			 "partition '%s'",
			 token, name);
	    return -1;
	}

    next:
	continue; /* how-to-do?? */

    }

    if (token && strcmp(token, "}")) {
	config_error("'}' expected in partition definition '%s'",
		     name);
	return -1;
    }

    /* else fill rest and add to tree */
    gp = sc_memdup(&group, sizeof(group));
    gp->name = sc_strdup(name);

    if (tree_add_referable_kd(groups, gp->name, gp)) {

	config_error("partition '%s' already defined", name);
	sc_unref(gp->name);
	sc_unref(gp);
	return -1;
    }

    return 0;
}

static int
config_collect_assignment(FILE *file, 
			  struct sc_assign_t **queue, 
			  const char *ident) {
    int rv;
    char buffer[SC_TOKENBUF_LEN];
    const char *part;
    struct sc_assign_t *assign;
    struct sc_limitgroup_t *group;

    part = config_extract_token(file, buffer);
    if (part && !strcmp(part, "to")) {

	/* skip */
	part = config_extract_token(file, buffer);
	if (part && !strcmp(part, "partition")) {

	    /* skip */
	    part = config_extract_token(file, buffer);
	}
    }

    if (!part || !*part) {
	config_error("assignment expects partition identifier to "
		     "assign regex '%s' to.", ident);
	return -1;
    }

    group = tree_fetch(groups, part);
    if (!group) {
	config_error("partition '%s' not found", part);
	return -1;
    }

    assign = sc_obs_alloc(sizeof(struct sc_assign_t));
    memset(assign, 0, sizeof(struct sc_assign_t));

    assign->string = sc_obs_alloc(strlen(ident) + 1);
    memcpy(assign->string, ident, strlen(ident) + 1);

    assign->group = group;
    sc_ref(group);
    
    if (queue == &host_assigns) { /* :b */
	rv = regcomp(&assign->re, ident, 
		     REG_EXTENDED | REG_ICASE | REG_NOSUB);
    } else {
	rv = regcomp(&assign->re, ident, 
		     REG_EXTENDED | REG_NOSUB);
    }

    if (rv) { /* compile fail */

	char errbuf[512];

	regerror(rv, &assign->re, errbuf, sizeof(errbuf));
	config_error("compiling regular expression '%s' failed: %s",
		     ident, errbuf);
	return -1;
    }
    
    /* to end of queue */
    while(*queue) queue = &((*queue)->next);
    *queue = assign;

    return 0;
}

static int
config_parse_config() {

    int i;
    char buffer[SC_TOKENBUF_LEN];
    FILE *file;
    const char *token, *param;

    struct {
	const char *name;
	int *value;
	int min_val;
	int max_val;
    } ints[] = {

	{ "loglevel", &(config.loglevel),
	  SC_MIN_LOGLEVEL, SC_MAX_LOGLEVEL },

	{ "obstack-chunk", &(config.obstack_chunk),
	  SC_MIN_OBSTACK_CHUNK, SC_MAX_OBSTACK_CHUNK },

	{ "read-bufferlen", &(config.read_bufferlen),
	  SC_MIN_READ_BUFFERLEN, SC_MAX_READ_BUFFERLEN },

	{ "expire-delay", &(config.expire_delay),
	  SCACHE_PTO_MIN_EXPIRE, SCACHE_PTO_MAX_EXPIRE },

	{ "sock-sndbuf", &(config.sock_sndbuf),
	  SC_MIN_SOCK_SNDBUF, SC_MAX_SOCK_SNDBUF },

	{ "sock-rcvbuf", &(config.sock_rcvbuf),
	  SC_MIN_SOCK_RCVBUF, SC_MAX_SOCK_RCVBUF },

	{ "max-reqsize", &(config.max_reqsize),
	  0, SCACHE_PTO_MAX_REQLEN },

	{ "listen-backlog", &(config.listen_backlog),
	  SC_MIN_LISTEN_BACKLOG, SC_MAX_LISTEN_BACKLOG },
	
	{ "listen-port", &(config.listen_port),
	  SC_MIN_LISTEN_PORT, SC_MAX_LISTEN_PORT },

	{ "unix-sock-mask", &(config.usock_mask),
	  0, 0 }
    };

    struct {
	const char *name;
	const char **value;
    } strings[] = {
	
	{ "pidfile-path", &(config.pidfile_path) },
	{ "runas-user", &temp_runas_user },
	{ "runas-group", &temp_runas_group },
	{ "unix-sock-owner", &temp_usock_owner },
	{ "unix-sock-group", &temp_usock_group },
	{ "unix-sock-path", &(config.usock_path) }
    };

    struct {
	const char *name;
	int *value;
    } bools[] = {

	{ "debug", &(config.debug) },
	{ "nodetach", &(config.nodetach) },
	{ "foreground", &(config.nodetach) },
	{ "enable-auth-partitions", &(config.enable_auth_partitions) },
	{ "enable-host-partitions", &(config.enable_host_partitions) },
	{ "disable-unix", &(config.disable_unix) },
	{ "disable-ipv4", &(config.disable_ipv4) },
	{ "disable-ipv6", &(config.disable_ipv6) },
	{ "defpart-allow-ring", &(config.defpart_ring) },
	{ "defpart-allow-lvals", &(config.defpart_lvals) },
	{ "defpart-allow-cache", &(config.defpart_cache) },
	{ "defpart-allow-stats", &(config.defpart_stats) },
	{ "defpart-allow-shared", &(config.defpart_shared) },
	{ "defpart-allow-session", &(config.defpart_session) },
	{ "defpart-allow-collect", &(config.defpart_collect) },
	{ "defpart-allow-shcollect", &(config.defpart_shcollect) }
    };

    file = fopen(sc_config->config_path, "r");
    if (!file) {

	/* return silently, no config file. existence
	   of user specified was already verified and
	   errored */
	return 0;
    }

    token = config_extract_token(file, buffer);
    while(token && *token) {

	/* integer parameters */
	for(i = 0; i < sc_asize(ints); i++) {
	    if (!strcmp(token, ints[i].name)) {
		*ints[i].value = 
		    config_get_int(file, ints[i].name,
				   ints[i].min_val, ints[i].max_val);

		if (*ints[i].value < 0) return -1;
		goto next;
	    }
	}

	/* string parameters */
	for(i = 0; i < sc_asize(strings); i++) {
	    if (!strcmp(token, strings[i].name)) {

		param = config_require_value(file, buffer,
					     strings[i].name);
		if (!param) return -1;
		*strings[i].value = strdup(param);
		goto next;
	    }
	}

	/* boolean parameters */
	for(i = 0; i < sc_asize(bools); i++) {
	    if (!strcmp(token, bools[i].name)) {
		*bools[i].value = config_get_bool(file, token);
		if (*bools[i].value < 0) return -1;

		goto next;
	    }
	}

	/* misc */
	if (!strcmp(token, "listen-addr")) {

	    struct sc_list_t *list = malloc(sizeof(*list));

	    /* queue of parameters */
	    list->data = 
		(void *)config_require_value(file, buffer, 
					     "listen-addr");
	    if (!list->data) return -1;

	    list->data = strdup(list->data);
	    list->next = config.listen_addrs_ipv4;
	    config.listen_addrs_ipv4 = list;

#ifdef ENABLE_IPV6
	} else if (!strcmp(token, "listen-addr-ipv6")) {

	    struct sc_list_t *list = malloc(sizeof(*list));

	    /* queue of parameters */
	    list->data = 
		(void *)config_require_value(file, buffer, 
					     "lister-addr-ipv6");
	    if (!list->data) return -1;

	    list->data = strdup(list->data);
	    list->next = config.listen_addrs_ipv6;
	    config.listen_addrs_ipv6 = list;
#endif

	} else if (!strcmp(token, "partition")) {

	    param = config_extract_token(file, buffer);
	    if (!param || !*param) {
		config_error("group definition expects identifier, "
			     "none found.");
		return -1;
	    }

	    if (config_collect_group(file, param) < 0) return -1;

	} else if (!strcmp(token, "assign")) {

	    struct sc_assign_t **queue = NULL;

	    param = config_extract_token(file, buffer);
	    if (param && !strcmp(param, "host")) {
		queue = &host_assigns;
	    } else if (param && !strcmp(param, "auth")) {
		queue = &auth_assigns;
	    } else {
		config_error("assign expects either 'host' or 'auth'");
		return -1;
	    }

	    param = config_extract_token(file, buffer);
	    if (!param || !*param) {
		config_error("identifier expected for assignment");
		return -1;
	    }

	    if (config_collect_assignment(file, queue, param) < 0) {
		return -1;
	    }

	} else {

	    config_error("unknown token '%s' in config file, exiting",
			 token);
	    return -1;
	}

    next:
	token = config_extract_token(file, buffer);
    }

    /* close */
    fclose(file);
    return 0;
}

static int
config_parse_argv() {

    return 0;
}

static uid_t
config_getuser(const char *tag, const char *user) {

    char *cp;
    uid_t id = 0;
    struct passwd *pwd;

    if (user) { /* NULL is for root */
	
	id = strtol(user, &cp, 0);
	if (*cp && (pwd = getpwnam(user))) {
	    id = pwd->pw_uid;
	} else if (*cp) {
	    fprintf(stderr, "resolving '%s' failed: invalid user '%s'\n",
		    tag, user);
	    return -1;
	}
    }
    
    return id;
}

static gid_t 
config_getgroup(const char *tag, const char *group) {

    char *cp;
    gid_t id = 0;
    struct group *grp;

    if (group) { /* NULL is for root */
	
	id = strtol(group, &cp, 0);
	if (*cp && (grp = getgrnam(group))) {
	    id = grp->gr_gid;
	} else if (*cp) {
	    fprintf(stderr, "resolving '%s' failed: invalid group '%s'\n",
		    tag, group);
	    return -1;
	}
    }
    
    return id;
}

static int
config_postprocess() {

    /* name-mapping */
    if ((config.runas_uid = config_getuser("runas-uid", 
					   temp_runas_user)) < 0 ||
	(config.runas_gid =  config_getgroup("runas-gid", 
					     temp_runas_group)) < 0 ||
	(config.usock_owner = config_getuser("unix-sock-owner", 
					     temp_usock_owner)) < 0 ||
	(config.usock_group = config_getgroup("unix-sock-group",
					      temp_usock_group)) < 0) {

	return -1;
    }


    return 0;
}

static void
config_dump_config() {

    fprintf(stdout,
	    "   config-path      : %s\n" /* 3 */
	    "   pidfile-path     : %s\n" /* 4 */
	    "\n"
	    "   loglevel         : %i\n" /* 5 */
	    "   obstack-chunk    : %i\n" /* 6 */
	    "   read-bufferlen   : %i\n" /* 7 */
	    "   expire-delay     : %i\n" /* 8 */
	    "   sock-sndbuf      : %i\n" /* 9 */
	    "   sock-rcvbuf      : %i\n" /* 10 */
	    "   max-reqsize      : %i KiB\n" /* 11 */
	    "   listen-backlog   : %i\n" /* 12 */
	    "   listen-port      : %i\n" /* 13 */
	    "   unix-sock-mask   : 0%o (oct)\n" /* 14 */
	    "   runas-user       : %s\n" /* 15 */
	    "   runas-group      : %s\n" /* 16 */
	    "   unix-sock-owner  : %s\n" /* 17 */
	    "   unix-sock-group  : %s\n" /* 18 */
	    "   unix-sock-path   : %s\n" /* 19 */
	    "   disable-unix     : %s\n" /* 20 */
	    "   disable-ipv4     : %s\n" /* 21 */
	    "   disable-ipv6     : %s\n" /* 22 */
	    "   defpart-allow-ring       : %s\n" /* 23 */
	    "   defpart-allow-lvals      : %s\n" /* 24 */
	    "   defpart-allow-cache      : %s\n" /* 25 */
	    "   defpart-allow-stats      : %s\n" /* 26 */
	    "   defpart-allow-shared     : %s\n" /* 27 */
	    "   defpart-allow-session    : %s\n" /* 28 */
	    "   defpart-allow-collect    : %s\n" /* 29 */
	    "   defpart-allow-shcollect  : %s\n" /* 30 */
	    "   enable-host-partitions   : %s\n" /* 31 */
	    "   enable-auth-partitions   : %s\n" /* 32 */
	    "\n" 
	    "   compile-options were :\n"
#ifdef ENABLE_DEVEL
	    "        ENABLE_DEVEL\n"
#endif
#ifdef DISABLE_NODES_ACCOUNTING
	    "        DISABLE_NODES_ACCOUNTING\n"
#else
	    "        ENABLE_NODES_ACCOUNTING\n"
#endif
#ifdef DISABLE_MEMORY_ACCOUNTING
	    "        DISABLE_MEMORY_ACCOUNTING\n"
#else
	    "        ENABLE_MEMORY_ACCOUNTING\n"
#endif
#ifdef ENABLE_FULL_PARTITIONING
	    "        ENABLE_FULL_PARTITIONING\n"
#else
	    "        DISABLE_FULL_PARTITIONING\n"
#endif
#ifdef ENABLE_EXTENDED_STATS
	    "        ENABLE_EXTENDED_STATS\n"
#else
	    "        DISABLE_EXTENDED_STATS\n"
#endif
	    "\n",

	    config.config_path,
	    config.pidfile_path,
	    config.loglevel,
	    config.obstack_chunk,
	    config.read_bufferlen,
	    config.expire_delay,
	    config.sock_rcvbuf,
	    config.sock_sndbuf,
	    (config.max_reqsize / 1024),
	    config.listen_backlog,
	    config.listen_port, /* 13 */
	    config.usock_mask,
	    temp_runas_user ? temp_runas_user : "(root)",
	    temp_runas_group ? temp_runas_group : "(root)",
	    temp_usock_owner ? temp_usock_owner : "(root)",
	    temp_usock_group ? temp_usock_group : "(root)",
	    config.usock_path, /* 19 */
	    config.disable_unix ? "true" : "false",
	    config.disable_ipv4 ? "true" : "false",
	    config.disable_ipv6 ? "true" : "false",
	    config.defpart_ring ? "true" : "false",
	    config.defpart_lvals ? "true" : "false",
	    config.defpart_cache ? "true" : "false",
	    config.defpart_stats ? "true" : "false",
	    config.defpart_shared ? "true" : "false",
	    config.defpart_session ? "true" : "false",
	    config.defpart_collect ? "true" : "false",
	    config.defpart_shcollect ? "true" : "false",
	    config.enable_host_partitions ? "true" : "false",
	    config.enable_auth_partitions ? "true" : "false");

    
}

int
config_dump_default_config() {

    /* header after --help */
    fprintf(stdout,
	    "Current builtin configuration values :\n\n");

    /* common config */
    config_dump_config();

    /* default partition limits */
    fprintf(stdout,
	    "   Default partition\n"
	    "\n"
	    "    cache-limit       : %lli bytes\n"
	    "    ring-nodes-limit  : %lli nodes\n"
	    "    ring-memory-limit : %lli bytes\n"
#ifdef SESSION_MEMORY_ACCOUNTING
	    "    memory-limit      : %lli bytes\n"
#endif
#ifdef SESSION_NODES_ACCOUNTING
	    "    nodes-limit       : %lli bytes\n"
#endif
	    "\n",
	    (long long int)config.defpart_cachemax,
	    (long long int)config.defpart_ringnodeslimit,
	    (long long int)config.defpart_ringmemorylimit
#ifdef SESSION_MEMORY_ACCOUNTING
	    ,
	    (long long int)config.defpart_mlimit
#endif
#ifdef SESSION_NODES_ACCOUNTING
	    ,
	    (long long int)config.defpart_nlimit
#endif
	    );

    return 0;
}

int 
config_dump_compiled_config() {

    /* header */
    fprintf(stdout,
	    "Configuration values after reading %s :\n\n",
	    config.config_path);

    /* common config */
    config_dump_config();

    /* TODO partitions */
    /* TODO assignments */

    return 0;
}

int
config_init(enum sc_bootstages_t stage) {

    /* peer default */
    if (stage == SC_STAGE_PREPARE) {

	groups = tree_create();

        if (config_parse_resolve_path() < 0) return -1;
	if (config_parse_config() < 0) return -1;
	if (config_parse_argv() < 0) return -1;
	if (config_postprocess() < 0) return -1;

    }

    return 0;
}

/*
  Local variables: ***
  mode: c ***
  c-basic-offset: 4 ***
  End: ***
*/
