<?php

@ini_set('memory_limit', '256M');

  /* for correct performance measurement you should run
   couple simultaneous tests and combine theirs results.
   otherwise latencies keep scached idling without 
   nothing to do */

define('TEST_ID', uniqid());
define('TEST_PREFIX', 'scache-test-' . TEST_ID);

if (0) {
    define('TEST_HOST', false);
    //define('TEST_HOST', 'remote.host.name.that.listens');
    define('TEST_HOST_NAME', false);
    define('TEST_HOST_IPV4', false);
    define('TEST_HOST_IPV6', false);
 } else {
    define('TEST_HOST', false);
    //define('TEST_HOST', 'remote.host.name.that.listens');
    // localhost might map to disabled ipv6
    define('TEST_HOST_NAME', '127.0.0.1' /* or 'localhost' */);
    define('TEST_HOST_IPV4', '127.0.0.1');
    define('TEST_HOST_IPV6', '::1');
 }

/* common functions and logging */
function utime_float() {
    list($usec, $sec) = explode(" ", microtime());
    return ((float)$usec + (float)$sec);
}

function m($line, $indent = 1) {
    echo str_repeat('.', ($indent * 4)) . "$line\n";
    return true;
}

function f($line, $nodie = false) {
    m("failed: $line", 2);
    if (!$nodie) die();
}

/* dump all functions defined by scache-module */
function dump_functions() {
    m('enumerating defined functions');
    $functions = get_extension_funcs('scache');
    foreach($functions as $func) {
	m("$func", 2);
    }
}

/* connectivity with different sockets */
function connect_unix() {
    $conn = new SCache(TEST_PREFIX . '_unix', TEST_HOST);
    if (!$conn) f('connection failed');
    else m('ok');
}

function connect_ipv4() {
    $conn = new SCache(TEST_PREFIX . '_ipv4', TEST_HOST_IPV4);
    if (!$conn) f('connection failed');
    else m('ok');
}

function connect_ipv6() {
    $conn = new SCache(TEST_PREFIX . '_ipv6', TEST_HOST_IPV6);
    if (!$conn) f('connection failed');
    else m('ok');
}


/* getindex internals, same session must return same index slot.
 SCache->open need be used for testing, because SCache->connect
 does not query server and hence dont get hand to server returned
 current index slot */
function connect_persistent_index() {

    m("persistent connections and index logic");
    $pfx1 = TEST_PREFIX . '_idx1';
    $pfx2 = TEST_PREFIX . '_idx2';

    $conn = new SCache($pfx1, TEST_HOST);
    $conn->reset();

    $conn = new SCache($pfx2, TEST_HOST);
    $conn->reset();

    $conn1a = new SCache($pfx1, TEST_HOST);
    $conn1a->open();
    $conn1b = new SCache($pfx2, TEST_HOST);
    $conn1b->open();
    $conn2a = new SCache($pfx1, TEST_HOST, false, true);
    $conn2a->open();
    $conn2b = new SCache($pfx2, TEST_HOST, false, true);
    $conn2b->open();

    if (!(($conn1a->getindex() !=
	   $conn1b->getindex()) ||
	  ($conn2a->getindex() !=
	   $conn2b->getindex()))) {
	f("indexes mismatch");
    } else {
	m("ok, passed");
	$conn1a->drop();
	$conn1b->drop();
	$conn2a->drop();
	$conn2b->drop();
    }
}

/* example index hinting for backend, might have some performance 
 impact when using with very large session tables. */
function fast_connections() {

    m("fast connections");
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $index = $conn->getindex();
    $newconn = new SCache(TEST_PREFIX,TEST_HOST,
			  false, false, false, false, $index);
    if (!$newconn->status()) {
	f('should not fail');
    } else {
	m('ok, passed');
	$conn->drop();
	$newconn->drop();
    }
}


/* flood server with multiple connections. when opened with
 nonpersisten functions module opens always new socket-connection to
 backend..
 
 when using persistent versions module reuses existing connection if
 host (path) and port matches.

 this happens even if connection is originally initialized by another
 separate unrelated client. every request bunch to backend is
 self-contained by protocol
*/

function flood_connections() {
    
    $count = 512;
    $conns = Array();

    m("flooding $count connections");
    for($i = 0; $i < $count; $i++) {
	$conns[] = $conn = new SCache(TEST_PREFIX, TEST_HOST_NAME);
	$conn->open();

	if (!$conn) {
	    f('connection failed');
	    return;
	}
    }
    m('ok, passed');
    $conn->drop();
}

/* if secret is provided on first initialization, it must
 be matched on all subsequent accesses.

 this might offer some extra security when sharing same
 scached-backend with multiple different clients */
function open_with_secret_key() {

    $prefix = TEST_PREFIX . '_secret';

    m("SCache->open with secret key");
    $conn = new SCache($prefix, TEST_HOST, 
		       false, false, 'MySecret');
    $conn->reset();

    $fail1 = new SCache($prefix, TEST_HOST, 
			false, false, 'InvalidSecret');
    $fail2 = new SCache($prefix, TEST_HOST);

    if ((@$fail1->reset()) || (@$fail2->reset())) {
	f("got connection despite of secret mismatch");
    } else {
	m("ok, passed");
	$conn->drop();
    }
}

/* merely "ping" to check socket is still connected.
 leftover from debugging and no sensible use known */
function session_status_result() {

    m('connection round-trip with SCache->status()');
    $conn = new SCache(TEST_PREFIX . '_status_exists', TEST_HOST);
    $conn->reset();

    if (!$conn->status()) {
	f('should not fail');
    } else {
	m('ok, passed');
	$conn->drop();
    }
}

/* large number of parallel sessions. might result odd performance
 fluxuation, because sessions are not immediately destroyed and 
 freed from backends memory. */
function large_session_tree() {

    $count = 20000;
    $checkpoint = ($count / 10);

    m('populating large session tree');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i sessions created", 2);
	$conn = new SCache(TEST_PREFIX . "_large_session$i", 
			   TEST_HOST, false, true);
	$conn->reset();
	$conn->set('value', $i);
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m('verifying sessions');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i sessions verified", 2);
	$conn = new SCache(TEST_PREFIX . "_large_session$i", 
			   TEST_HOST, false, true);
	if ($conn->get('value') !== $i) {
	    f("key unexistent");
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m('dropping sessions');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i sessions dropped", 2);
	$conn = new SCache(TEST_PREFIX . "_large_session$i",
			   TEST_HOST, false, true);
	$conn->drop();
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");
}

/* connect-by-name */
function connect_byname_localhost() {

    $conn = new SCache(TEST_PREFIX . '_byname', TEST_HOST_NAME);
    if (!$conn) {
	f('connection failed');
    } else {
	m('ok');
	$conn->drop();
    }
}

/* session options */
function getset_options() {

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    for($i = 1000; $i < 2000; $i++) {
	if (!$conn->setopt(SCACHE_OPT_EXPIRES, $i)) {
	    f('set expires failed: ' . $conn->strerror());
	    return;
	}

	if (!($exp = $conn->getopt(SCACHE_OPT_EXPIRES))) {
	    f('set expires failed: ' . $conn->strerror());
	    return;
	}

	if ($exp != $i) {
	    f("set expires mismatch, was set as $i, returned $exp");
	    return;
	}
    }

    m("ok, get and set session expires");

    $conn->setopt(SCACHE_OPT_KEEPONELIMIT, true);
    if (($conn->getopt(SCACHE_OPT_KEEPONELIMIT) === true) &&
	$conn->setopt(SCACHE_OPT_KEEPONELIMIT, false) &&
	($conn->getopt(SCACHE_OPT_KEEPONELIMIT) === false) &&
	($conn->lasterr() == SCERR_SUCCESS)) {
	m("ok, get and set keeponelimit");
    } else {
	f("get/set keeponelimit failed: " . $conn->strerror());
    }

    $conn->drop();
}


/* minimum session timeout is currently 5 seconds, maximum
 is about year. values are rounded to limits if overflown */

function session_timeout() {

    $tag = TEST_PREFIX . '_session_timeout';

    m('reset and populate with short timeout');
    $conn1 = new SCache($tag, TEST_HOST, 5);
    $conn1->reset();
    $conn1->set('test', 'test');

    m('test immediate reconnection');
    $conn2 = new SCache($tag, TEST_HOST);
    $conn2->open();
    if ($conn2->get('test') == 'test') {
	m('check value ok');
    } else {
	f('check value fail');
    }

    m('sleep 6 seconds and expect to be expired');
    sleep(6);

    m('reconnect and assume expired');
    $conn3 = new SCache($tag, TEST_HOST);
    $conn3->open();
    if ($conn3->get('test') == 'test') {
	f('session did not expire', true);
    } else {
	m('expired, good');
    }

    $conn1->drop();
    $conn2->drop();
    $conn3->drop();
}

/* example of modifying session timeout afterwards */
function modifying_session_timeout() {

    m('reset with 60 second timeout');
    $conn = new SCache(TEST_PREFIX, TEST_HOST, 60);
    $conn->reset();
    $conn->set('test', 'test');

    m('sleep 6 seconds and verify not expired');
    sleep(6);
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->open();
    if ($conn->get('test') != 'test') {
	f('session expired prematurely');
	return;
    }

    m('session ok, modify timeout to 5 seconds');
    if (($ov = $conn->setopt(SCACHE_OPT_EXPIRES, 5)) != 60) {
	f("SCache->setopt() returned wrong old timeout: $ov");
    }

    m('sleep another 6 seconds and verify session expired');
    sleep(6);
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->open();
    if ($conn->exists('test')) {
	f('session expired prematurely');
	return;
    }
    
    m('ok, passed');
    $conn->drop();
}

function session_renaming() {

    $count = 100000;
    $checkpoint = ($count / 10);

    $prefix = TEST_PREFIX . posix_getpid();
    $conn = new SCache("$prefix-$count", TEST_HOST);
    $conn->drop();


    m("setup first session");
    $conn = new SCache("$prefix", TEST_HOST);
    $conn->reset();
    $conn->set("testvalue", "testvalue");

    m("start renaming");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i rounds", 2);

	$new = "$prefix-$i";
	if (!$conn->rename("$new")) {
	    f("failed: " . $conn->strerror());
	    return;
	}
    }

    $duration = utime_float() - $start;
    $keys = round((($count * 3) / $duration), 2);
    m("ok, $keys ops/sec");

    $conn->drop();
}

function session_dropping() {

    $tag = 'value';

    m("reset session and sanitycheck");
    $conn = new SCache(TEST_PREFIX, TEST_HOST, false, true);
    $conn->reset();
    $conn->set($tag, true);
    if ($conn->get($tag) !== true) {
	f("value mismatch: " . $conn->get($tag));
    }

    m("drop session");
    if (!$conn->drop()) {
	f("drop failed: " . $conn->strerror());
	return;
    }

    m("re-open session with same name and check values does not exist");
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->open();
    if (($val = $conn->get($tag))) {
	f("value exists: $val");
    }

    if ($conn->exists($tag)) {
	f("value exists");
    }

    m("ok, all passed");
    $conn->drop();
}

/* set and unset values */
function unsetting_values() {

    $count = 50000;

    m("unsetting values");
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    m("populating session", 2);
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	$conn->set("$i", "$i");
	$conn->set("path/$i", "$i");
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m("next, unsetting values", 2);
    $start = utime_float();
    $conn->clear("path");
    for($i = 0; $i < $count; $i++) {
	$conn->clear("$i");
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    m("next, verifying they are gone", 2);
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($conn->exists("$i") ||
	    $conn->exists("path/$i")) {
	    
	    f("value still exists");
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, passed: $keys ops/sec");
    
    $conn->drop();
}

/* error codes */
function error_codes() {

    $sesid1 = TEST_PREFIX;
    $sesid2 = TEST_PREFIX . '_notexist';
    $sesid3 = TEST_PREFIX . '_dummy';

    $conn = new SCache($sesid1, TEST_HOST);
    $conn->reset();

    $null = new SCache($sesid2, TEST_HOST);
    $null->reset();
    $null->drop();

    $null = new SCache($sesid2, TEST_HOST);
    $dummy = new SCache($sesid3, TEST_HOST);
    $dummy->reset();

    /* expect no-session */
    if (!$null->status($null) &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('session status');
    } else {
	f('session status: ' . $null->strerror());
    }

    /* expect no-session */
    if (!$null->rename("$sesid2-rename") &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('rename unexistent session');
    } else {
	f('rename unexistent session: ' . $null->strerror());
    }

    /* expect exists */
    if (!$conn->rename($sesid3) &&
	($conn->lasterr() == SCERR_EXIST)) {
	m('rename to existing');
    } else {
	f('rename to existing: ' . $conn->strerror());
    }

    /* expect no-session */
    if (!$null->drop() &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('drop unexistent session');
    } else {
	f('drop unexistent session: ' . $null->strerror());
    }

    /* expect no-session */
    if (!$null->set('value', 'null') &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('no session on set');
    } else {
	f('no session on set: ' . $null->strerror());
    }

    $conn->set('value', 'value');

    /* expect unexisting */
    if (!$conn->get('unexisting') &&
	($conn->lasterr() == SCERR_NOTEXIST)) {
	m('get unexisting');
    } else {
	f('get unexisting: ' . $conn->strerror());
    }

    /* expect exists */
    if (!$conn->add('value', 'add') &&
	($conn->lasterr() == SCERR_EXIST)) {
	m('add to existing');
    } else {
	f('add to existing: ' . $conn->strerror());
    }

    /* expect not-exists */
    if (!$conn->replace('unexisting', 'add') &&
	($conn->lasterr() == SCERR_NOTEXIST)) {
	m('replace unexistent');
    } else {
	f('replace unexistent: ' . $conn->strerror());
    }

    /* expect unexisting */
    if ($conn->clear('unexisting') &&
	($conn->lasterr() == SCERR_NOTEXIST)) {
	m('unsetting unexisting');
    } else {
	f('unsetting unexisting: ' . $conn->strerror());
    }

    /* expect unexisting */
    if ($conn->vclear('unexisting') &&
	($conn->lasterr() == SCERR_NOTEXIST)) {
	m('vunset unexisting');
    } else {
	f('vunset unexisting: ' . $conn->strerror());
    }

    /* expect no-session */
    if (!$null->chget('value') &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('cache get without session');
    } else {
	f('cache get session: ' . $null->strerror());
    }

    /* expect no-session */
    if (!$null->chput('value', 'value') &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('cache put without session');
    } else {
	f('cache put session: ' . $null->strerror());
    }

    /* expect no-session */
    if (!$null->chclear('value') &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('cache clear without session');
    } else {
	f('cache clear session: ' . $null->strerror());
    }

    /* expect unexisting */
    if (!$conn->chget('unexisting') &&
	($conn->lasterr() == SCERR_NOTEXIST)) {
	m('cache get unexisting');
    } else {
	f('cache get unexisting: ' . $conn->strerror());
    }

    /* expect no-session */
    if (!$null->getopt(SCACHE_OPT_EXPIRES) &&
	($null->lasterr() == SCERR_NO_SESSION)) {
	m('getopt without session');
    } else {
	f('getopt session: ' . $null->strerror());
    }

    $conn->drop();
    $null->drop();
    $dummy->drop();
}

/* performance testing wide flat keyspace */
function wide_keyspace() {
    
    $count = 100000;
    $checkpoint = ($count / 10);
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    /* store */
    m("store values ($count rounds)");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values stored", 2);
	$conn->set("$i", $i);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    /* check */
    m("next, read and check values");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values read", 2);
	if ($conn->get("$i") != $i) {
	    f("invalid result: $i == " . $conn->get("$i"));
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");

    $conn->drop();
}


/* performance testing deep multilevel sessiontable */
function deep_tree() {

    $count = 20000;
    $checkpoint = ($count / 10);
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    /* store */
    m("store values ($count rounds)");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	$key = "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i" .
	    "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i";
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values stored", 2);
	$conn->set($key, $i);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    /* check */
    m("read and check values");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	$key = "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i" .
	    "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i";
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values read", 2);
	if ($conn->get($key) != $i) {
	    f("invalid result: $i == " . $conn->get($key));
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    $conn->drop();
}


/* example code of path testing. note that return values
 SCNODE_NONEXISTENT and SCNODE_EMPTY mean effectively same thing
 and can be both returned when no path exists 

 (actually, I have no idea what that did mean...)
*/
function stat_paths() {

    m("testing SCache->stat -operation");
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $conn->set('path/to/node', 'value');
    if (!(($path = 'path') ||
	  ($conn->stat($path) != SCNODE_BRANCH) ||

	  ($path = 'path/to') ||
	  ($conn->stat($path) != SCNODE_BRANCH) ||

	  ($path = 'path/to/node') ||
	  ($conn->stat($path) != SCNODE_VALUE) ||

	  ($path = 'path/to/nowhere') ||
	  ($conn->stat($path) != SCNODE_NONEXISTENT))) {

	f("failed on $path");
    }

    m('ok, passed');
    $conn->drop();
}

/* performance testing with flat keyspace */
function random_access_flat() {

    $count = 100000;
    $space = 1000;
    $reqs = 0;
    $checkpoint = ($count / 10);

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $set_count = 0;
    $clr_count = 0;
    $array = Array();

    /* store */
    m("random storing and fetching on $space items wide keyspace");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i accesses", 2);
	
	$r = rand() % $space;
	if (@$array["$r"]) {
	    $reqs++;
	    if ($conn->get("$r") != true) {
		f("invalid result: $i != true");
		return;
	    }

	    $reqs++;
	    $conn->set("$r", false);
	    $clr_count++;
	    $array["$r"] = false;
	} else {

	    $reqs++;
	    $conn->set("$r", true);
	    $set_count++;
	    $array["$r"] = true;
	}
    }
    $duration = utime_float() - $start;
    $keys = round(($reqs / $duration), 2);
    m("ok, $keys ops/sec, $set_count sets, $clr_count clears");

    $conn->drop();
}


/* performance test */
function get_set_performance_r($conn) {

    $count = 100000;
    $rounds = 0;
    $checkpoint = ($count / 10);
    $conn->reset();

    $start = utime_float();
    for($i = ($count / $checkpoint); $i > 0; $i--) {
	for($n = $checkpoint; $n > 0; $n--) {
	    $conn->set('key', $i);
	    $conn->get('key');
	}
	$rounds += $checkpoint;
	m("$rounds rounds done", 2);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);

    m("ok, passed: $keys ops/sec");
    $conn->drop();
}

function get_set_performance_unix() {
    m("performance with single key over unix socket");
    get_set_performance_r(new SCache(TEST_PREFIX, TEST_HOST));
}

function get_set_performance_localhost() {
    m("performance with single key over connection to localhost ");
    get_set_performance_r(new SCache(TEST_PREFIX, TEST_HOST_NAME));
}

function get_set_high_performance_iov_r($keys) {

    $count = 500000;
    $rounds = 0;
    $checkpoint = ($count / 10);
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    m("populating session");
    for($i = 0; $i < $keys; $i++) {
	$conn->set("$i", $i);
    }

    m("performance with $keys keys wide session using 20-op SCache->iov");
    $array =
	Array(Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'));
	      

    $start = utime_float();
    for($i = ($count / $checkpoint); $i > 0; $i--) {
	for($n = ($checkpoint / sizeof($array)); $n > 0; $n--) {
	    $res = $conn->iov($array);
	}
	$rounds += $checkpoint;
	m("$rounds rounds done", 2);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);

    m("ok, passed: $keys ops/sec");
    $conn->drop();
}

function get_set_high_performance_iov_1key() {
    get_set_high_performance_iov_r(1);
}

function get_set_high_performance_iov_100000key() {
    get_set_high_performance_iov_r(100000);
}


/* example of operating shared counters. counters are internally
 64-bit integers both in backend and in transit but typecasted to
 php-integer before returning from module. 

 operating counters that might overflow 0x7fffffff with both
 32- and 64-bit clients, might not be simple thing */
function shared_counters() {

    $count = 50000;
    $counter = TEST_ID . '_counter';
    $checkpoint = ($count / 10);
    m("connect");
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    m("resetting counter");
    $conn->vset($counter, 0);

    m("incrementing counter");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("value $i", 2);
	if (($r = $conn->vadd($counter, 1)) != $i + 1) {
	    f("expected $i got $r");
	    var_dump($r);
	    die();
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");

    m("next, decrementing counter");
    $start = utime_float();
    for($i = $count; $i > 0; $i--) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("value $i", 2);
	if (($r = $conn->vsub($counter, 1)) != $i + -1) {
	    f("expected $i got $r");
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");

    m("next, setting and testing values");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("value $i", 2);
	$conn->vset("$counter", $i);
	$conn->vset("$counter$i", $i);
	if (($conn->vget("$counter") != $i) ||
	    ($conn->vget("$counter$i") != $i)) {

	    f("values mismatch");
	}
    }

    $duration = utime_float() - $start;
    $keys = round((($count * 4) / $duration), 2);
    m("ok, passed: $keys ops/sec");
    $conn->drop();
}

/* testing and example functionality for basic operations.
 there is also parallel and equivalent functions for operating
 shared storage (SCache->shadd, SCache->shreplace, SCache->shset,
 SCache->unset, SCache->shexist, SCache->shstat) 

 shared storage is common for all clients, not session-specific
 private storeage
*/
function set_add_replace_logic() {

    m('value operations');
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    m('SCache->set sets always');
    if (!($conn->set('value', 1) &&
	  $conn->set('value', 2) &&
	  ($conn->get('value') === 2))) {
	f('expected "2" got "' . $conn->get('value') . '"');
    }

    m('SCache->add adds if nonexistent');
    if (!($conn->add('addok', 'addok') ||
	  ($conn->get('addok') != 'addok'))) {
	f('SCache->add did not add');
    }

    m('SCache->add does not replace');
    if ($conn->add('value', 'addreplace') ||
	($conn->lasterr() != SCERR_EXIST) ||
	($conn->get('value') == 'addreplace')) {
	f('SCache->add replaced, thought not ought to');
    }

    m('SCache->replace replaces only already existing');
    if (!$conn->replace('value', 'replaced') ||
	($conn->get('value') !== 'replaced')) {
	f('SCache->replace did not replace');
    }

    m('SCache->replace does not add values');
    if ($conn->replace('notexist', 'replace') ||
	($conn->lasterr() != SCERR_NOTEXIST) ||
	($conn->get('notexist') == 'replace')) {
	f('SCache->replace did add value, thought not ought to');
    }

    m('ok, passed');
    $conn->drop();
}

/* collect */
function collect_operations() {

    $count = 100;
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    m('basic operations');
    for($i = 0; $i < $count; $i++) {
	$conn->set("prefix/key$i", "$i");
	$conn->shset("prefix/key$i", "$i");
    }

    $conn->coll("prefix", "into");
    $conn->shcoll("prefix", "into");
    $a1 = $conn->get("into");
    $a2 = $conn->shget("into");

    if ((sizeof($a1) != $count) ||
	(sizeof($a2) != $count)) {
	f("array sizes mismatch: " .
	  sizeof($a1) . "(p) " .
	  sizeof($a2) . "(s) " .
	  "expected: $count");
    }

    m('keys matches');
    for($i = 0; $i < $count; $i++) {
	if (!$a1["key$i"] || !$a2["key$i"]) {
	    f("key missing, key$i");
	}

	unset($a1["key$i"]);
	unset($a2["key$i"]);
    }

    if ((sizeof($a1) > 0) || (sizeof($a2) > 0)) {
	f("extra keys in result");
    }

    m('binary keys');
    for($i = 1; $i < 256; $i++) {
	if (chr($i) == '/') continue;

	$chr = chr($i) . chr($i) . chr($i);
	$conn->set("bin/$chr", true);
    }

    $conn->coll('bin', 'into');
    $a1 = $conn->get('into');
    for($i = 1; $i < 256; $i++) {
	if (chr($i) == '/') continue;

	$chr = chr($i) . chr($i) . chr($i);
	if (!$a1["$chr"]) {
	    f("binary key failed: $i");
	}
    }

    m('subtype checking');
    $conn->set('subtype/value', 1);
    $conn->set('subtype/subdir/node', 2);
    $conn->set('subtype/deleted', 3);
    $conn->clear('subtype/deleted');
    $conn->coll('subtype', 'subtype/list');
    $res = $conn->get('subtype/list');

    if ($res['value'] != SCNODE_VALUE) {
	f("type mismatch: expected " . SCNODE_VALUE .
	  " got: " . $res['value']);
    }
    if ($res['subdir'] != SCNODE_BRANCH) {
	f("type mismatch: expected " . SCNODE_BRANCH .
	  " got: " . $res['value']);
    }
    if (array_key_exists('deleted', $res)) {
	f("deleted value exists in keys");
    }

    m('iov operations');
    $arr = Array(Array(SCIOP_SET, 'iov/key1', 1),
		 Array(SCIOP_SET, 'iov/key2', 2),
		 Array(SCIOP_COLL, 'iov/', 'iov/iov3'),
		 Array(SCIOP_COLL, '//iov/', 'iov//iov4'),
		 Array(SCIOP_COLL, '//iov//', 'iov//iov5'),
		 Array(SCIOP_COLL, 'iov', 'iov/sub/1'),
		 Array(SCIOP_COLL, 'iov', 'iov/sub/2'),
		 Array(SCIOP_COLL, 'iov', 'iov/list'),
		 Array(SCIOP_GET, 'iov/list'),
		 Array(SCIOP_COLL, 'iov/sub', 'iov/sub'),
		 Array(SCIOP_GET, 'iov/sub'),
		 Array(SCIOP_COLL, 'iov', 'iov/final'),
		 Array(SCIOP_GET, 'iov/final'));
    $res = $conn->iov($arr);
    foreach($res as $row) {
	if ($row == 1 || is_array($row)) continue;
	f("unexpcted results on iov");
    }

    $conn->drop();
}


/* example code and functionality testing for scached's cache
 features.

 unlike other functionality, cache code is really a cache that
 cant drop values in any time. private and shared storage quarantees
 all values stored to exist, cache does not and drops them freely
 on memory shortage

 cache features are common to all clients.
*/
function cache_features() {

    $count = 10000;

    //m('cache is really shared cache?');
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $conn1 = new SCache(TEST_PREFIX . '_cache1', TEST_HOST);
    $conn1->reset();

    $conn2 = new SCache(TEST_PREFIX . '_cache2', TEST_HOST);
    $conn2->reset();

    $key = TEST_ID;

    /* ...common to alla sessions */
    $conn->chput("$key-testkey", 'ishere');
    if (($conn1->chget("$key-testkey") != 'ishere') ||
	($conn2->chget("$key-testkey") != 'ishere')) {
	f('shared access to keys');
    }

    m('basic cache functionality');
    m('populating values', 2);
    for($i = 0; $i < $count; $i++) {
	$conn->chput("$key$i", "is $i");
    }

    m('verifing values', 2);
    for($i = 0; $i < $count; $i++) {
	if (($val = $conn->chget("$key$i")) !== "is $i") {
	    f("invalid value for key $i '$val'");
	}
    }


    m('stressing backends cache-ring, part 1');
    for($i = 0; $i < $count; $i++) {
	$idx = rand() % $count;
	if (($val = $conn->chget("$key$i")) !== "is $i") {
	    f("invalid value for key $i '$val'");
	}
    }

    m('stressing backends cache-ring, part 2');
    for($i = 1; $i < $count; $i++) {
	$prev = $i - 1;
	$conn->chclear("$key$prev");
    }

    m('verifying removed values');
    for($i = 0; $i < ($count - 1); $i++) {
	if (($val = $conn->chget("$key$i")) ||
	    $conn->lasterr() != SCERR_NOTEXIST) {
	    f("keys still exists: key$key: $val");
	}
    }

    m('subtree expiration');
    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	$conn->chput($key, $key);
	if ($conn->chget($key) != $key)
	    f("'$key' != '" . $conn->chget($key) . "'", 3);
    }
    
    m('expire by unput', 2);
    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	$conn->chclear("$i");
	if ($conn->chget($key) == $key)
	    f("'$key' != '" . $conn->chget($key) . "'", 3);
    }

    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	$conn->chput($key, $key);
	if ($conn->chget($key) != $key)
	    f("'$key' != '" . $conn->chget($key) . "'", 3);
    }
    m('expire by put(false)', 2);
    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	$conn->chput("$i", false);
	if ($conn->chget($key) == $key)
	    f("'$key' != '" . $conn->chget($key) . "'", 3);
    }

    m('complete expire from root');
    $start = utime_float(); $count = 0;
    for($a = 0; $a < 10; $a++) {
	for($b = 0; $b < 10; $b++)
	    for($c = 0; $c < 10; $c++)
		for($d = 0; $d < 10; $d++) 
		    for($e = 0; $e < 10; $e++, $count++) {
			$key = "$a/$b/$c/$d/$e";
			$conn->chput($key, $key);
			if ($conn->chget($key) != $key)
			    f("'$key' != '" . 
			      $conn->chget($key) . "'", 3);
		    }

	if ($a > 0) m(($a * 10000) . ' values populated', 2);
    }

    m('expire from root', 2);
    $conn->chclear('/');

    m('verify', 2);
    if ($conn->chget('1') ||
	$conn->chget('1/1') ||
	$conn->chget('1/1/1') ||
	$conn->chget('1/1/1/1')) {

	f("cache values still exist", 3);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");

    m('constant storing, constant expiring');
    $val = 
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa';
    $count = 100000;
    $checkpoint = ($count / 10);
    $start = utime_float(); 
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values cached", 2);
	$conn->chput("key$i", $val);
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");


    m("random cache banging");
    $count = 100000;
    $checkpoint = ($count / 10);
    $start = utime_float(); 
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values cached", 2);

	$a = rand() % 10;
	$b = rand() % 10;
	$c = rand() % 10;
	$d = rand() % 10;
	$key = "$a/$b/$c/$d";

	if (($val = $conn->chget($key))) {
	    if ($val !== $key) {
		f("value mismatch: $val != $key", 2);
		return;
	    }

	    $conn->chclear($key);
	} else {
	    $conn->chput($key, $key);
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");
    $conn->drop();
}

function random_banging_r(&$rec, $pnt, $level) {

    if ($level > 100) return;
    if ($level > $rec['depth']) $rec['depth'] = $level;

    $arr = Array();
    $conn = $rec['conn'];
    $count = $rec['count'];
    $checkpoint = $count / 10;

    /* for faster loop */
    $ops = 0;
    $gets = 0;
    $unsets = 0;
    $newbrs = 0;
    $newnodes = 0;
    $descends = 0;

    while(($rec['round']++  < $count)) {

	if (($rec['round'] % $checkpoint) == 0) 
	    m("round ${rec['round']}", 2);

	$rnd = rand() % 100; /* spread */
	$key = "key$rnd";
	$path = "$pnt/$key";

	$stat = $conn->stat($path); $ops++;
	if ($stat == SCNODE_BRANCH) {

	    /* descent or unlink */
	    if (!(rand() % 2)) {
		random_banging_r($rec, $path, ($level + 1));
		$descends++;
	    }
	    $conn->clear($path); $ops++;
	    $unsets++;
	    
	} else if ($stat == SCNODE_VALUE) {

	    if ($conn->get($path) != $path) {
		f("value mismatch: should '$path', is " . 
		  $conn->get($path));
	    }
	    $conn->clear($path); $ops += 2;
	    $gets++;
	    $unsets++;
	    
	} else {
	
	    /* not-exist, create file or sub */
	    if (!(rand() % 2)) {
		$conn->set($path, "$path"); $ops++;
		$newnodes++;
	    } else {
		$conn->set("$path/dummy", "$path"); $ops++;
		$newbrs++;
	    }
	}

	if ($level > 1) {
	    if (!(rand() % 25)) {
		$rec['ops'] += $ops;
		$rec['gets'] += $gets;
		$rec['unsets'] += $unsets;
		$rec['newbrs'] += $newbrs;
		$rec['newnodes'] += $newnodes;
		$rec['descends'] += $descends;
		return;
	    }
	}
    }
}

function random_banging() {

    $rec = 
	Array('conn' => new SCache(TEST_PREFIX, TEST_HOST),
	      'round' => 0,
	      'depth' => 0,
	      'ops' => 0,
	      'gets' => 0,
	      'unsets' => 0,
	      'newbrs' => 0,
	      'newnodes' => 0,
	      'descends' => 0,
	      'count' => 100000);

    $rec['conn']->reset();
    $start = utime_float();

    random_banging_r($rec, '', 0);
    $duration = utime_float() - $start;
    $keys = round(($rec['ops'] / $duration), 2);
    m("ok, $keys ops/sec");
    m(" ops          : " . $rec['ops']);
    m(" gets         : " . $rec['gets']);
    m(" unsets       : " . $rec['unsets']);
    m(" newbranches  : " . $rec['newbrs']);
    m(" newnodes     : " . $rec['newnodes']);
    m(" descends     : " . $rec['descends']);    
    m(" maxdepth     : " . $rec['depth']);
    m("passed");

    $rec['conn']->drop();
}

/* scached's internal io-loop performance testing */
function io_multiplexing() {

    $conn_count = 50;
    $reqs_count = 100000;
    $spread = $conn_count * 10;
    $checkpoint = ($reqs_count / 10);
    $conns = Array();

    m('allocating connections');
    for($i = 0; $i < $conn_count; $i++) {
	$conns[$i] = $conn = new SCache(TEST_PREFIX, TEST_HOST_NAME);
	$conn->reset();
    }

    m('random access over connections');
    for($i = 0; $i < $reqs_count; $i++) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("round $i", 2);

	$r = rand() % $spread;
	$conn = $conns[$r % $conn_count];
	$conn->set("$r", "$r");
    }

    m("ok, passed");
    $conn->drop();
}

/* testing and example code for using SCache->iov. basically SCache->iov
 is the function to use always whenever possible.

 absolutely largest bottleneck in scaches performance is socket io and
 it's delays, so packing as many requests as possible on single query
 results very large improvements on performance as previous iov-tests
 show.

 you should do something like 

 list($username, $fullname, $surname) = 
     SCache->iov($conn, Array(Array(SCIOP_GET, 'userinfo/login'),
                             Array(SCIOP_GET, 'userinfo/fname'),
                             Array(SCIOP_GET, 'userinfo/sname')));

 doing several requests in one query is practically as fast
 as doing one request. socket-delays are 90% of latencies
*/


function scache_iov_swiss_army_knife() {

    /* conn */
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $counter = 'cntr' . TEST_ID;

    /* ensure old values are removed, integer are not
     in session-specific storage so SCache->reset don't
     reset shared values */
    $conn->vclear($counter);

    /* shared counters */
    m('testing shared counters');
    $ops = 
	Array(Array(SCIOP_VSET, $counter, 12), /* returns oldval */
	      Array(SCIOP_VADD, $counter, 2), /* returns newval */
	      Array(SCIOP_VSUB, $counter, 1), /* return newval */
	      Array(SCIOP_VGET, $counter), /* return current val */
	      Array(SCIOP_VADD, $counter, 10), /* returns newval */
	      Array(SCIOP_VUNSET, $counter)); /* returns lastval */

    $res = $conn->iov($ops);
    if (sizeof($res) != sizeof($ops)) {
	f("result array size mismatch");
	return;
    }

    if (($res[0] != 0) ||
	($res[1] != 14) ||
	($res[2] != 13) ||
	($res[3] != 13) ||
	($res[4] != 23) ||
	($res[5] != 23)) {

	print_r($res);
	f("counter results mismatch");
    }


    /* session-private data */
    m("ok, testing session private operations");
    $path = "path";
    $node = "$path/node";
    $ops =
	Array(Array(SCIOP_SET, $node, '1st'), /* returns success */
	      Array(SCIOP_SET, $node, '2nd'), /* returns success */
	      Array(SCIOP_REPLACE, $node, '3rd'), /* returns success */
	      Array(SCIOP_GET, $node), /* returns '3rd' */
	      Array(SCIOP_ADD, $node, '4th'), /* returns failure */
	      Array(SCIOP_STAT, $node), /* returns SCNODE_VALUE */
	      Array(SCIOP_STAT, $path), /* returns SCNODE_BRANCH */
	      Array(SCIOP_UNSET, $node)); /* returns success */

    $res = $conn->iov($ops);
    if (sizeof($res) != sizeof($ops)) {
	f("result array size mismatch");
	return;
    }

    if (($res[0] != true) ||
	($res[1] != true) ||
	($res[2] != true) ||
	($res[3] != '3rd') ||
	($res[4] != false) ||
	($res[5] != SCNODE_VALUE) ||
	($res[6] != SCNODE_BRANCH) ||
	($res[7] != true)) {

	f("node values mismatch");
	return;
    }


    /* accessing shared data */
    m("ok, testing shared data operations");
    $path = "path";
    $node = "$path/node";
    $ops =
	Array(Array(SCIOP_SHSET, $node, '1st'), /* returns success */
	      Array(SCIOP_SHSET, $node, '2nd'), /* returns success */
	      Array(SCIOP_SHREPLACE, $node, '3rd'), /* returns success */
	      Array(SCIOP_SHGET, $node), /* returns '3rd' */
	      Array(SCIOP_SHADD, $node, '4th'), /* returns failure */
	      Array(SCIOP_SHSTAT, $node), /* returns SCNODE_VALUE */
	      Array(SCIOP_SHSTAT, $path), /* returns SCNODE_BRANCH */
	      Array(SCIOP_SHUNSET, $node)); /* returns success */

    $res = $conn->iov($ops);
    if (sizeof($res) != sizeof($ops)) {
	f("result array size mismatch");
	return;
    }

    if (($res[0] != true) ||
	($res[1] != true) ||
	($res[2] != true) ||
	($res[3] != '3rd') ||
	($res[4] != false) ||
	($res[5] != SCNODE_VALUE) ||
	($res[6] != SCNODE_BRANCH) ||
	($res[7] != true)) {

	f("node values mismatch on shared table");
	return;
    }

    $conn->drop();
}

function large_iov() {

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $reqs = Array(Array(SCIOP_VSET, 'counter', 1));
    for($i = 0; $i < 100; $i++) 
	$reqs[] = Array(SCIOP_VADD, 'counter', 1);
    $reqs[] = Array(SCIOP_VGET, 'counter');

    /* multiple times */
    $res = $conn->iov($reqs);
    $res = $conn->iov($reqs);
    $res = $conn->iov($reqs);
    $res = $conn->iov($reqs);
    $res = $conn->iov($reqs);

    if ($res[sizeof($res) - 1] != 101) {
	f('values mismatch, expect 101 got : ' . $res[sizeof($res) - 1]);
    } else {
	m('ok, 100 iops, result matches');
    }
    
    $conn->drop();
}

function ring_basics() {

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();
    
    m('push values');
    if (!($conn->rnpush(TEST_ID, 0) &&
	  $conn->rnpush(TEST_ID, 1) &&
	  $conn->rnpush(TEST_ID, 2))) {
	f($conn->strerror());
    }

    m('unshift values');
    if (!($conn->rnunshift(TEST_ID, 5) &&
	  $conn->rnunshift(TEST_ID, 4) &&
	  $conn->rnunshift(TEST_ID, 3))) {
	f($conn->strerror());
    }

    m('ring size');
    if (!($conn->rnsize(TEST_ID) == 6)) {
	f('should be 6, got ' . $conn->rnsize(TEST_ID));
    }

    m('rotate forward');
    if (!($conn->rnrotf(TEST_ID, 3) &&
	  $conn->rnrotf(TEST_ID, -4))) {
	f($conn->strerror());
    }

    m('rotage backwards');
    if (!($conn->rnrotb(TEST_ID, 1) &&
	  $conn->rnrotb(TEST_ID, -5))) {
	f($conn->strerror());
    }

    m('internal ordering');
    for($i = 0; $i < $conn->rnsize(TEST_ID); $i++) {
	if (($n = $conn->rnget(TEST_ID)) !== $i) {
	    f("expected $i got $n");
	} 
	$conn->rnrotf(TEST_ID, 1);
    }

    m('pop/unshift');
    if ((($n = $conn->rnpop(TEST_ID)) !== ($e = 5)) ||
	(($n = $conn->rnshift(TEST_ID)) !== ($e = 0)) ||
	(($n = $conn->rnpop(TEST_ID)) !== ($e = 4))) {
	f("expected $e, got $n");
    }

    m('remaining size');
    if (($n = $conn->rnsize(TEST_ID)) != 3) {
	f("expected 4, got $n");
    }

    m('clearing');
    if (!$conn->rnclear(TEST_ID)) {
	f($conn->strerror());
    }

    m('clearing unexistent');
    if (!$conn->rnclear(TEST_ID) &&
	(($e = $conn->lasterr()) != SCERR_NOTEXIST)) {

	f($conn->strerror($e));
    }

    m('getting, shifting, popping unexistent');
    if ($conn->rnget(TEST_ID) ||
	(($e = $conn->lasterr()) != SCERR_NOTEXIST) ||
	$conn->rnshift(TEST_ID) ||
	(($e = $conn->lasterr()) != SCERR_NOTEXIST) ||
	$conn->rnpop(TEST_ID) ||
	(($e = $conn->lasterr()) != SCERR_NOTEXIST)) {
	
	f($conn->strerror());
    }
}

function singlecall_ring_stressing() {

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $count = 100000;
    $checkpoint = ($count / 10);
    m('populating ring');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values pushed", 2);
	$conn->rnpush(TEST_ID, $i);
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    $rcount = 200000;
    $checkpoint = ($rcount / 10);
    m('rotate and verify ring');
    for($i = 0; $i < $rcount; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i rounds", 2);
	if (($val = $conn->rnget(TEST_ID)) != ($i % $count)) {
	    $conn->rnclear(TEST_ID);
	    f("expected $val, got " . ($i % $count));
	}
	$conn->rnrotf(TEST_ID, 1);
    }

    $duration = utime_float() - $start;
    $keys = round((($rcount * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m('clearing ring');
    $conn->rnclear(TEST_ID);
}

function ring_stressing_with_iov() {

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $count = 200000;
    $checkpoint = ($count / 10);
    m('populating ring (10op iov)');
    $start = utime_float();
    for($i = 0; $i < $count; $i += 10) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values pushed", 2);
	$conn->iov(Array(Array(SCIOP_RNPUSH, TEST_ID, ($i + 0)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 1)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 2)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 3)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 4)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 5)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 6)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 7)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 8)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 9))));
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    m('purge ring (10op iov)');
    $start = utime_float();
    for($i = 0; $i < $count; $i += 10) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values shifted", 2);
	$res = $conn->iov(Array(Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID)));
	if ($res[0] !== $i) {
	    f("expected $i, got " . $res[0]);
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    m('verify that ring is empty');
    if ($conn->rnsize(TEST_ID) !== 0) {
	f("expected 0, got " . $conn->rnsize(TEST_ID));
    }

    m('clearing ring');
    $conn->rnclear(TEST_ID);
}

function protocol_limits() {

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $key = str_repeat('a', 16384);
    $lval = str_repeat('a', 524288);
    $sval = '1';

    m('uplimit key');
    $conn->set($key, $sval);
    if ($conn->get($key) !== $sval) {
	f('retval mismatch');
    }

    m('uplimit value');
    $conn->set($key, $lval);
    if ($conn->get($key) !== $lval) {
	f('retval mismatch');
    }

    $count = 10000;
    $checkpoint = ($count / 10);

    m('memleak detect and pushing very large packets with iov');
    $start = utime_float();
    for($i = 0; $i < $count; $i += 10) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values pushed", 2);
	$res =
	    $conn->iov(Array(Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key)));
	if (($res[1] !== $lval)) {
	    f('retval mismatch');
	}
    }
    $duration = utime_float() - $start;
    $keys = round((($count) / $duration), 2);
    m("ok, $keys ops/sec");
}

function accumulator() {

    $id = TEST_ID;
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $conn->set("$id/a1", TEST_ID);

    $iops = Array(Array(SCIOP_GET, "$id/a1"),
		  Array(SCIOP_SET, "$id/a2"),
		  Array(SCIOP_SHSET, "$id/a3"),
		  Array(SCIOP_CHPUT, "$id/a4"),
		  Array(SCIOP_RNPUSH, "$id/a5"),
		  Array(SCIOP_GET, "$id/a2"),
		  Array(SCIOP_SHGET, "$id/a3"),
		  Array(SCIOP_CHGET, "$id/a4"),
		  Array(SCIOP_RNPOP, "$id/a5"));

    $rv = $conn->iov($iops);

    if (($rv[0] != TEST_ID) ||
	($rv[5] != TEST_ID) ||
	($rv[6] != TEST_ID) ||
	($rv[7] != TEST_ID) ||
	($rv[8] != TEST_ID)) {
	f('retval mismatch');
    } else {
	m("ok");
    }

    $conn->shclear($id);
    $conn->drop();
}

function merry_go_round() {

}

function escape_now() {
    m("woo hoo, i'm free!!!");
    die();
}

function scache_iov_lasterr() {

    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    $iops = Array(Array(SCIOP_SET, '/exist', 'true'),
		  Array(SCIOP_LASTERR),
		  Array(SCIOP_GET, '/not-exist'),
		  Array(SCIOP_LASTERR),
		  Array(SCIOP_ADD, '/exist', 'add-never-overwrites'),
		  Array(SCIOP_LASTERR),
		  Array(SCIOP_REPLACE, '/not-exist', 'existing-required'),
		  Array(SCIOP_LASTERR));

    $res = $conn->iov($iops);
    if (($res[1] == SCERR_SUCCESS) &&
	($res[3] == SCERR_NOTEXIST) &&
	($res[5] == SCERR_EXIST) &&
	($res[7] == SCERR_NOTEXIST)) {

	m('ok, passed');
    } else {
	f('failed: got ' .
	  '1:' . $conn->strerror($res[1]) .
	  '3:' . $conn->strerror($res[3]) .
	  '5:' . $conn->strerror($res[5]) .
	  '7:' . $conn->strerror($res[7]));
    }
}

function final_statistics() {
    
    $conn = new SCache(TEST_PREFIX, TEST_HOST);
    $conn->reset();

    if (is_array($conn->srvstats())) {
	m('ok, statistics');
    } else {
	f('statistics');
    }
}

function run_tests() {

    $tests = 
	Array(
	      'dump_functions',
	      'connect_unix',
	      'connect_ipv4',
	      //'connect_ipv6',
	      'connect_byname_localhost',
	      'connect_persistent_index',
	      'fast_connections',
	      'flood_connections',
	      'open_with_secret_key',
	      'session_status_result',
	      'large_session_tree',
	      'getset_options',
	      'session_timeout',
	      'modifying_session_timeout',
	      'session_renaming',
	      'session_dropping',
	      'unsetting_values',
	      'collect_operations',
	      'wide_keyspace',
	      'deep_tree',
	      'stat_paths',
	      'error_codes',
	      'random_access_flat',
	      'get_set_performance_unix',
	      'get_set_performance_localhost',
	      'get_set_high_performance_iov_1key',
	      'get_set_high_performance_iov_100000key',
	      'shared_counters',
	      'set_add_replace_logic',
	      'cache_features',
	      'random_banging',
	      'io_multiplexing',
	      'scache_iov_swiss_army_knife',
	      'large_iov',
	      'protocol_limits',
	      'ring_basics',
	      'singlecall_ring_stressing',
	      'ring_stressing_with_iov',
	      'accumulator',
	      'scache_iov_lasterr',
	      'final_statistics');

    foreach($tests as $test) {
	echo $test . "....\n";
	call_user_func($test);
    }

}

run_tests();

/*
  Local variables: ***
  mode: php ***
  c-basic-offset: 4 ***
  End: ***
*/
