<?php

@ini_set('memory_limit', '256M');

  /* for correct performance measurement you should run
   couple simultaneous tests and combine theirs results.
   otherwise latencies keep scached idling without 
   nothing to do */

define('TEST_ID', uniqid());
define('TEST_PREFIX', 'scache-test-' . TEST_ID);

if (0) {
    define('TEST_HOST', false);
    //define('TEST_HOST', 'remote.host.name.that.listens');
    define('TEST_HOST_NAME', false);
    define('TEST_HOST_IPV4', false);
    define('TEST_HOST_IPV6', false);
 } else {
    define('TEST_HOST', false);
    //define('TEST_HOST', 'remote.host.name.that.listens');
    // localhost might map to disabled ipv6
    define('TEST_HOST_NAME', '127.0.0.1' /* or 'localhost' */);
    define('TEST_HOST_IPV4', '127.0.0.1');
    define('TEST_HOST_IPV6', '::1');
 }

/* common functions and logging */
function utime_float() {
    list($usec, $sec) = explode(" ", microtime());
    return ((float)$usec + (float)$sec);
}

function m($line, $indent = 1) {
    echo str_repeat('.', ($indent * 4)) . "$line\n";
    return true;
}

function f($line, $nodie = false) {
    m("failed: $line", 2);
    if (!$nodie) die();
}

/* dump all functions defined by scache-module */
function dump_functions() {
    m('enumerating defined functions');
    $functions = get_extension_funcs('scache');
    foreach($functions as $func) {
	m("$func", 2);
    }
}

/* connectivity with different sockets */
function connect_unix() {
    $conn = scache_connect(TEST_PREFIX . '_unix', TEST_HOST);
    if (!$conn) f('connection failed');
    else m('ok');
}

function connect_ipv4() {
    $conn = scache_connect(TEST_PREFIX . '_ipv4', TEST_HOST_IPV4);
    if (!$conn) f('connection failed');
    else m('ok');
}

function connect_ipv6() {
    $conn = scache_connect(TEST_PREFIX . '_ipv6', TEST_HOST_IPV6);
    if (!$conn) f('connection failed');
    else m('ok');
}


/* getindex internals, same session must return same index slot.
 scache_open need be used for testing, because scache_connect
 does not query server and hence dont get hand to server returned
 current index slot */
function connect_persistent_index() {

    m("persistent connections and index logic");
    $pfx1 = TEST_PREFIX . '_idx1';
    $pfx2 = TEST_PREFIX . '_idx2';

    scache_preset($pfx1, TEST_HOST);
    scache_preset($pfx2, TEST_HOST);
    $conn1a = scache_open($pfx1, TEST_HOST);
    $conn1b = scache_open($pfx2, TEST_HOST);
    $conn2a = scache_popen($pfx1, TEST_HOST);
    $conn2b = scache_popen($pfx2, TEST_HOST);

    if (!((scache_getindex($conn1a) !=
	   scache_getindex($conn1b)) ||
	  (scache_getindex($conn2a) !=
	   scache_getindex($conn2b)))) {
	f("indexes mismatch");
    } else {
	m("ok, passed");
	scache_drop($conn1a);
	scache_drop($conn1b);
	scache_drop($conn2a);
	scache_drop($conn2b);
    }
}

/* example index hinting for backend, might have some performance 
 impact when using with very large session tables.

 usage scenario in scache_connect_fast()'s examples */
function fast_connections() {

    m("fast connections");
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    $index = scache_getindex($conn);
    $newconn = scache_connect_fast(TEST_PREFIX, $index, TEST_HOST);
    if (!scache_status($newconn)) {
	f('should not fail');
    } else {
	m('ok, passed');
	scache_drop($conn);
    }
}


/* flood server with multiple connections. when opened with
 nonpersisten functions like scache_open, scache_reset, module opens
 always new socket-connection to backend..
 
 when using persistent versions (scache_popen, scache_pconnect,
 scache_preset) module reuses existing connection if host (path) and
 port matches.

 this happens even if connection is originally initialized by another
 separate unrelated client. every request bunch to backend is
 self-contained by protocol
*/

function flood_connections() {
    
    $count = 512;
    $conns = Array();

    m("flooding $count connections");
    for($i = 0; $i < $count; $i++) {
	$conns[] = $conn = scache_open(TEST_PREFIX, TEST_HOST_NAME);
	if (!$conn) {
	    f('connection failed');
	    return;
	}
    }
    m('ok, passed');
    scache_drop($conn);
}

/* if secret is provided on first initialization, it must
 be matched on all subsequent accesses.

 this might offer some extra security when sharing same
 scached-backend with multiple different clients */
function open_with_secret_key() {

    $prefix = TEST_PREFIX . '_secret';

    m("scache_open with secret key");
    $conn = scache_reset($prefix, TEST_HOST, 'MySecret');
    if (($conn = @scache_reset($prefix, TEST_HOST, 'InvalidSecret')) ||
	($conn = @scache_reset($prefix, TEST_HOST))) {

	f("got connection despite of secret mismatch");
    } else {
	m("ok, passed");
    }
}

/* merely "ping" to check socket is still connected.
 leftover from debugging and no sensible use known */
function session_status_result() {

    m('connection round-trip with scache_status()');
    $conn = scache_reset(TEST_PREFIX . '_status_exists', TEST_HOST);
    if (!scache_status($conn)) {
	f('should not fail');
    } else {
	m('ok, passed');
	scache_drop($conn);
    }
}

/* large number of parallel sessions. might result odd performance
 fluxuation, because sessions are not immediately destroyed and 
 freed from backends memory. */
function large_session_tree() {

    $count = 20000;
    $checkpoint = ($count / 10);

    m('populating large session tree');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i sessions created", 2);
	$conn = scache_preset(TEST_PREFIX . "_large_session$i", TEST_HOST);
	scache_set($conn, 'value', $i);
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m('verifying sessions');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i sessions verified", 2);
	$conn = scache_pconnect(TEST_PREFIX . "_large_session$i", TEST_HOST);
	if (scache_get($conn, 'value') !== $i) {
	    f("key unexistent");
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m('dropping sessions');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i sessions dropped", 2);
	scache_drop(scache_pconnect(TEST_PREFIX . "_large_session$i",
				    TEST_HOST));
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");
}

/* connect-by-name */
function connect_byname_localhost() {
    $conn = @scache_connect(TEST_PREFIX . '_byname', TEST_HOST_NAME);
    if (!$conn) {
	f('connection failed');
    } else {
	m('ok');
	scache_drop($conn);
    }
}

/* session options */
function getset_options() {

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    for($i = 1000; $i < 2000; $i++) {
	if (!scache_setopt($conn, SCACHE_OPT_EXPIRES, $i)) {
	    f('set expires failed: ' . 
	      scache_strerror(scache_lasterr($conn)));
	    return;
	}

	if (!($exp = scache_getopt($conn, SCACHE_OPT_EXPIRES))) {
	    f('set expires failed: ' . 
	      scache_strerror(scache_lasterr($conn)));
	    return;
	}

	if ($exp != $i) {
	    f("set expires mismatch, was set as $i, returned $exp");
	    return;
	}
    }

    m("ok, get and set session expires");

    scache_setopt($conn, SCACHE_OPT_KEEPONELIMIT, true);
    if ((scache_getopt($conn, SCACHE_OPT_KEEPONELIMIT) === true) &&
	scache_setopt($conn, SCACHE_OPT_KEEPONELIMIT, false) &&
	(scache_getopt($conn, SCACHE_OPT_KEEPONELIMIT) === false) &&
	(scache_lasterr($conn) == SCERR_SUCCESS)) {

	m("ok, get and set keeponelimit");
    } else {
	f("get/set keeponelimit failed: " .
	      scache_strerror(scache_lasterr($conn)));
    }

    scache_drop($conn);
}


/* minimum session timeout is currently 5 seconds, maximum
 is about year. values are rounded to limits if overflown */

function session_timeout() {

    $tag = TEST_PREFIX . '_session_timeout';

    m('reset and populate with short timeout');
    $conn1 = scache_reset($tag, TEST_HOST, false, 5);
    scache_set($conn1, 'test', 'test');

    m('test immediate reconnection');
    $conn2 = scache_open($tag, TEST_HOST);
    if (scache_get($conn2, 'test') == 'test') {
	m('check value ok');
    } else {
	f('check value fail');
    }

    m('sleep 6 seconds and expect to be expired');
    sleep(6);

    m('reconnect and assume expired');
    $conn3 = scache_open($tag, TEST_HOST);
    if (scache_get($conn3, 'test') == 'test') {
	f('session did not expire', true);
    } else {
	m('expired, good');
    }

    scache_drop($conn1);
    scache_drop($conn2);
    scache_drop($conn3);
}

/* example of modifying session timeout afterwards */
function modifying_session_timeout() {

    m('reset with 60 second timeout');
    $conn = scache_reset(TEST_PREFIX, TEST_HOST, false, 60);
    scache_set($conn, 'test', 'test');

    m('sleep 6 seconds and verify not expired');
    sleep(6);
    $conn = scache_open(TEST_PREFIX, TEST_HOST);
    if (scache_get($conn, 'test') != 'test') {
	f('session expired prematurely');
	return;
    }

    m('session ok, modify timeout to 5 seconds');
    if (($ov = scache_setopt($conn, SCACHE_OPT_EXPIRES, 5)) != 60) {
	f("scache_setopt() returned wrong old timeout: $ov");
    }

    m('sleep another 6 seconds and verify session expired');
    sleep(6);
    $conn = scache_open(TEST_PREFIX, TEST_HOST);
    if (scache_exists($conn, 'test')) {
	f('session expired prematurely');
	return;
    }
    
    m('ok, passed');
    scache_drop($conn);
}

function session_renaming() {

    $count = 100000;
    $checkpoint = ($count / 10);

    $prefix = TEST_PREFIX . posix_getpid();
    if (($sess = scache_open("$prefix-$count", TEST_HOST))) {
	scache_drop($sess);
    }

    m("setup first session");
    $sess = scache_reset("$prefix", TEST_HOST);
    scache_set($sess, "testvalue", "testvalue");

    m("start renaming");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i rounds", 2);

	$new = "$prefix-$i";
	/*
	if (($exist = scache_open("$new"))) {
	    scache_drop($exist);
	}
	*/

	if (!scache_rename($sess, "$new")) {
	    f("failed: " . scache_strerror(scache_lasterr($sess)));
	    return;
	}
    }

    $duration = utime_float() - $start;
    $keys = round((($count * 3) / $duration), 2);
    m("ok, $keys ops/sec");

    scache_drop($sess);
}

function session_dropping() {

    $tag = 'value';

    m("reset session and sanitycheck");
    $conn = scache_preset(TEST_PREFIX, TEST_HOST);
    scache_set($conn, $tag, true);
    if (scache_get($conn, $tag) !== true) {
	f("value mismatch: " . scache_get($conn, $tag));
    }

    m("drop session");
    if (!scache_drop($conn)) {
	f("drop failed: " . scache_strerror(scache_lasterr($conn)));
	return;
    }

    m("re-open session with same name and check values does not exist");
    $conn = scache_open(TEST_PREFIX, TEST_HOST);
    if (($val = scache_get($conn, $tag))) {
	f("value exists: $val");
    }

    if (scache_exists($conn, $tag)) {
	f("value exists");
    }

    m("ok, all passed");
    scache_drop($conn);
}

/* set and unset values */
function unsetting_values() {

    $count = 50000;

    m("unsetting values");
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    m("populating session", 2);
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	scache_set($conn, "$i", "$i");
	scache_set($conn, "path/$i", "$i");
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m("next, unsetting values", 2);
    $start = utime_float();
    scache_unset($conn, "path");
    for($i = 0; $i < $count; $i++) {
	scache_unset($conn, "$i");
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    m("next, verifying they are gone", 2);
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if (scache_exists($conn, "$i") ||
	    scache_exists($conn, "path/$i")) {
	    
	    f("value still exists");
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round((($count * 2) / $duration), 2);
    m("ok, passed: $keys ops/sec");
    
    scache_drop($conn);
}

/* error codes */
function error_codes() {

    $sesid1 = TEST_PREFIX;
    $sesid2 = TEST_PREFIX . '_notexist';
    $sesid3 = TEST_PREFIX . '_dummy';


    $conn = scache_reset($sesid1, TEST_HOST);
    $null = scache_reset($sesid2, TEST_HOST);
    scache_drop($null);
    $null = scache_connect($sesid2, TEST_HOST);
    $dummy = scache_reset($sesid3, TEST_HOST);

    /* expect no-session */
    if (!scache_status($null) &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('session status');
    } else {
	f('session status: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect no-session */
    if (!scache_rename($null, "$sesid2-rename") &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('rename unexistent session');
    } else {
	f('rename unexistent session: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect exists */
    if (!scache_rename($conn, $sesid3) &&
	(scache_lasterr($conn) == SCERR_EXIST)) {
	m('rename to existing');
    } else {
	f('rename to existing: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect no-session */
    if (!scache_drop($null) &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('drop unexistent session');
    } else {
	f('drop unexistent session: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect no-session */
    if (!scache_set($null, 'value', 'null') &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('no session on set');
    } else {
	f('no session on set: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    scache_set($conn, 'value', 'value');

    /* expect unexisting */
    if (!scache_get($conn, 'unexisting') &&
	(scache_lasterr($conn) == SCERR_NOTEXIST)) {
	m('get unexisting');
    } else {
	f('get unexisting: ' .
	  scache_strerror(scache_lasterr($conn)));
    }

    /* expect exists */
    if (!scache_add($conn, 'value', 'add') &&
	(scache_lasterr($conn) == SCERR_EXIST)) {
	m('add to existing');
    } else {
	f('add to existing: ' .
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect not-exists */
    if (!scache_replace($conn, 'unexisting', 'add') &&
	(scache_lasterr($conn) == SCERR_NOTEXIST)) {
	m('replace unexistent');
    } else {
	f('replace unexistent: ' .
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect unexisting */
    if (scache_unset($conn, 'unexisting') &&
	(scache_lasterr($conn) == SCERR_NOTEXIST)) {
	m('unsetting unexisting');
    } else {
	f('unsetting unexisting: ' .
	  scache_strerror(scache_lasterr($conn)));
    }

    /* expect unexisting */
    if (scache_vunset($conn, 'unexisting') &&
	(scache_lasterr($conn) == SCERR_NOTEXIST)) {
	m('vunset unexisting');
    } else {
	f('vunset unexisting: ' .
	  scache_strerror(scache_lasterr($conn)));
    }

    /* expect no-session */
    if (!scache_chget($null, 'value') &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('cache get without session');
    } else {
	f('cache get session: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect no-session */
    if (!scache_chput($null, 'value', 'value') &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('cache put without session');
    } else {
	f('cache put session: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect no-session */
    if (!scache_chclear($null, 'value') &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('cache clear without session');
    } else {
	f('cache clear session: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    /* expect unexisting */
    if (!scache_chget($conn, 'unexisting') &&
	(scache_lasterr($conn) == SCERR_NOTEXIST)) {
	m('cache get unexisting');
    } else {
	f('cache get unexisting: ' .
	  scache_strerror(scache_lasterr($conn)));
    }

    /* expect no-session */
    if (!scache_getopt($null, SCACHE_OPT_EXPIRES) &&
	(scache_lasterr($null) == SCERR_NO_SESSION)) {
	m('getopt without session');
    } else {
	f('getopt session: ' . 
	  scache_strerror(scache_lasterr($null)));
    }

    scache_drop($conn);
    scache_drop($null);
    scache_drop($dummy);
}

/* performance testing wide flat keyspace */
function wide_keyspace() {
    
    $count = 100000;
    $checkpoint = ($count / 10);
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    /* store */
    m("store values ($count rounds)");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values stored", 2);
	scache_set($conn, "$i", $i);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    /* check */
    m("next, read and check values");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values read", 2);
	if (scache_get($conn, "$i") != $i) {
	    f("invalid result: $i == " . scache_get($conn, "$i"));
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");

    scache_drop($conn);
}


/* performance testing deep multilevel sessiontable */
function deep_tree() {

    $count = 20000;
    $checkpoint = ($count / 10);
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    /* store */
    m("store values ($count rounds)");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	$key = "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i" .
	    "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i";
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values stored", 2);
	scache_set($conn, $key, $i);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    /* check */
    m("read and check values");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	$key = "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i" .
	    "/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i/$i";
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values read", 2);
	if (scache_get($conn, $key) != $i) {
	    f("invalid result: $i == " . scache_get($conn, $key));
	    return;
	}
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    scache_drop($conn);
}


/* example code of path testing. note that return values
 SCNODE_NONEXISTENT and SCNODE_EMPTY mean effectively same thing
 and can be both returned when no path exists 

 (actually, I have no idea what that did mean...)
*/
function stat_paths() {

    m("testing scache_stat -operation");
    $conn = scache_preset(TEST_PREFIX, TEST_HOST);

    scache_set($conn, 'path/to/node', 'value');
    if (!(($path = 'path') ||
	  (scache_stat($conn, $path) != SCNODE_BRANCH) ||

	  ($path = 'path/to') ||
	  (scache_stat($conn, $path) != SCNODE_BRANCH) ||

	  ($path = 'path/to/node') ||
	  (scache_stat($conn, $path) != SCNODE_VALUE) ||

	  ($path = 'path/to/nowhere') ||
	  (scache_stat($conn, $path) != SCNODE_NONEXISTENT))) {

	f("failed on $path");
    }

    m('ok, passed');
    scache_drop($conn);
}

/* performance testing with flat keyspace */
function random_access_flat() {

    $count = 100000;
    $space = 1000;
    $reqs = 0;
    $checkpoint = ($count / 10);
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    $set_count = 0;
    $clr_count = 0;
    $array = Array();

    /* store */
    m("random storing and fetching on $space items wide keyspace");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i accesses", 2);
	
	$r = rand() % $space;
	if (@$array["$r"]) {
	    $reqs++;
	    if (scache_get($conn, "$r") != true) {
		f("invalid result: $i != true");
		return;
	    }

	    $reqs++;
	    scache_set($conn, "$r", false);
	    $clr_count++;
	    $array["$r"] = false;
	} else {

	    $reqs++;
	    scache_set($conn, "$r", true);
	    $set_count++;
	    $array["$r"] = true;
	}
    }
    $duration = utime_float() - $start;
    $keys = round(($reqs / $duration), 2);
    m("ok, $keys ops/sec, $set_count sets, $clr_count clears");

    scache_drop($conn);
}


/* performance test */
function get_set_performance_r($conn) {

    $count = 100000;
    $rounds = 0;
    $checkpoint = ($count / 10);
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    $start = utime_float();
    for($i = ($count / $checkpoint); $i > 0; $i--) {
	for($n = $checkpoint; $n > 0; $n--) {
	    scache_set($conn, 'key', $i);
	    scache_get($conn, 'key');
	}
	$rounds += $checkpoint;
	m("$rounds rounds done", 2);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);

    m("ok, passed: $keys ops/sec");
    scache_drop($conn);
}

function get_set_performance_unix() {
    m("performance with single key over unix socket");
    get_set_performance_r(scache_reset(TEST_PREFIX, TEST_HOST));
}

function get_set_performance_localhost() {
    m("performance with single key over connection to localhost ");
    get_set_performance_r(scache_reset(TEST_PREFIX, TEST_HOST_NAME));
}

function get_set_high_performance_iov_r($keys) {

    $count = 500000;
    $rounds = 0;
    $checkpoint = ($count / 10);
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    m("populating session");
    for($i = 0; $i < $keys; $i++) {
	scache_set($conn, "$i", $i);
    }

    m("performance with $keys keys wide session using 20-op scache_iov");
    $array =
	Array(Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'),

	      Array(SCIOP_SET, 'key', 1),
	      Array(SCIOP_GET, 'key'));
	      

    $start = utime_float();
    for($i = ($count / $checkpoint); $i > 0; $i--) {
	for($n = ($checkpoint / sizeof($array)); $n > 0; $n--) {
	    $res = scache_iov($conn, $array);
	}
	$rounds += $checkpoint;
	m("$rounds rounds done", 2);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);

    m("ok, passed: $keys ops/sec");
    scache_drop($conn);
}

function get_set_high_performance_iov_1key() {
    get_set_high_performance_iov_r(1);
}

function get_set_high_performance_iov_100000key() {
    get_set_high_performance_iov_r(100000);
}


/* example of operating shared counters. counters are internally
 64-bit integers both in backend and in transit but typecasted to
 php-integer before returning from module. 

 operating counters that might overflow 0x7fffffff with both
 32- and 64-bit clients, might not be simple thing */
function shared_counters() {

    $count = 50000;
    $counter = TEST_ID . '_counter';
    $checkpoint = ($count / 10);
    m("connect");
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    m("resetting counter");
    scache_vset($conn, $counter, 0);

    m("incrementing counter");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("value $i", 2);
	if (($r = scache_vadd($conn, $counter, 1)) != $i + 1) {
	    f("expected $i got $r");
	    var_dump($r);
	    die();
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");

    m("next, decrementing counter");
    $start = utime_float();
    for($i = $count; $i > 0; $i--) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("value $i", 2);
	if (($r = scache_vsub($conn, $counter, 1)) != $i + -1) {
	    f("expected $i got $r");
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");

    m("next, setting and testing values");
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("value $i", 2);
	scache_vset($conn, "$counter", $i);
	scache_vset($conn, "$counter$i", $i);
	if ((scache_vget($conn, "$counter") != $i) ||
	    (scache_vget($conn, "$counter$i") != $i)) {

	    f("values mismatch");
	}
    }

    $duration = utime_float() - $start;
    $keys = round((($count * 4) / $duration), 2);
    m("ok, passed: $keys ops/sec");
    scache_drop($conn);
}

/* testing and example functionality for basic operations.
 there is also parallel and equivalent functions for operating
 shared storage (scache_shadd, scache_shreplace, scache_shset,
 scache_unset, scache_shexist, scache_shstat) 

 shared storage is common for all clients, not session-specific
 private storeage
*/
function set_add_replace_logic() {

    m('value operations');
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    m('scache_set sets always');
    if (!(scache_set($conn, 'value', 1) &&
	  scache_set($conn, 'value', 2) &&
	  (scache_get($conn, 'value') === 2))) {
	f('expected "2" got "' . scache_get($conn, 'value') . '"');
    }

    m('scache_add adds if nonexistent');
    if (!(scache_add($conn, 'addok', 'addok') ||
	  (scache_get($conn, 'addok') != 'addok'))) {
	f('scache_add did not add');
    }

    m('scache_add does not replace');
    if (scache_add($conn, 'value', 'addreplace') ||
	(scache_lasterr($conn) != SCERR_EXIST) ||
	(scache_get($conn, 'value') == 'addreplace')) {
	f('scache_add replaced, thought not ought to');
    }

    m('scache_replace replaces only already existing');
    if (!scache_replace($conn, 'value', 'replaced') ||
	(scache_get($conn, 'value') !== 'replaced')) {
	f('scache_replace did not replace');
    }

    m('scache_replace does not add values');
    if (scache_replace($conn, 'notexist', 'replace') ||
	(scache_lasterr($conn) != SCERR_NOTEXIST) ||
	(scache_get($conn, 'notexist') == 'replace')) {
	f('scache_replace did add value, thought not ought to');
    }

    m('ok, passed');
    scache_drop($conn);
}

/* collect */
function collect_operations() {

    $count = 100;
    $conn = scache_preset(TEST_PREFIX, TEST_HOST);

    m('basic operations');
    for($i = 0; $i < $count; $i++) {
	scache_set($conn, "prefix/key$i", "$i");
	scache_shset($conn, "prefix/key$i", "$i");
    }

    scache_coll($conn, "prefix", "into");
    scache_shcoll($conn, "prefix", "into");
    $a1 = scache_get($conn, "into");
    $a2 = scache_shget($conn, "into");

    if ((sizeof($a1) != $count) ||
	(sizeof($a2) != $count)) {
	f("array sizes mismatch: " .
	  sizeof($a1) . "(p) " .
	  sizeof($a2) . "(s) " .
	  "expected: $count");
    }

    m('keys matches');
    for($i = 0; $i < $count; $i++) {
	if (!$a1["key$i"] || !$a2["key$i"]) {
	    f("key missing, key$i");
	}

	unset($a1["key$i"]);
	unset($a2["key$i"]);
    }

    if ((sizeof($a1) > 0) || (sizeof($a2) > 0)) {
	f("extra keys in result");
    }

    m('binary keys');
    for($i = 1; $i < 256; $i++) {
	if (chr($i) == '/') continue;

	$chr = chr($i) . chr($i) . chr($i);
	scache_set($conn, "bin/$chr", true);
    }

    scache_coll($conn, 'bin', 'into');
    $a1 = scache_get($conn, 'into');
    for($i = 1; $i < 256; $i++) {
	if (chr($i) == '/') continue;

	$chr = chr($i) . chr($i) . chr($i);
	if (!$a1["$chr"]) {
	    f("binary key failed: $i");
	}
    }

    m('subtype checking');
    scache_set($conn, 'subtype/value', 1);
    scache_set($conn, 'subtype/subdir/node', 2);
    scache_set($conn, 'subtype/deleted', 3);
    scache_unset($conn, 'subtype/deleted');
    scache_coll($conn, 'subtype', 'subtype/list');
    $res = scache_get($conn, 'subtype/list');

    if ($res['value'] != SCNODE_VALUE) {
	f("type mismatch: expected " . SCNODE_VALUE .
	  " got: " . $res['value']);
    }
    if ($res['subdir'] != SCNODE_BRANCH) {
	f("type mismatch: expected " . SCNODE_BRANCH .
	  " got: " . $res['value']);
    }
    if (array_key_exists('deleted', $res)) {
	f("deleted value exists in keys");
    }

    m('iov operations');
    $arr = Array(Array(SCIOP_SET, 'iov/key1', 1),
		 Array(SCIOP_SET, 'iov/key2', 2),
		 Array(SCIOP_COLL, 'iov/', 'iov/iov3'),
		 Array(SCIOP_COLL, '//iov/', 'iov//iov4'),
		 Array(SCIOP_COLL, '//iov//', 'iov//iov5'),
		 Array(SCIOP_COLL, 'iov', 'iov/sub/1'),
		 Array(SCIOP_COLL, 'iov', 'iov/sub/2'),
		 Array(SCIOP_COLL, 'iov', 'iov/list'),
		 Array(SCIOP_GET, 'iov/list'),
		 Array(SCIOP_COLL, 'iov/sub', 'iov/sub'),
		 Array(SCIOP_GET, 'iov/sub'),
		 Array(SCIOP_COLL, 'iov', 'iov/final'),
		 Array(SCIOP_GET, 'iov/final'));
    $res = scache_iov($conn, $arr);
    foreach($res as $row) {
	if ($row == 1 || is_array($row)) continue;
	f("unexpcted results on iov");
    }

    scache_drop($conn);
}


/* example code and functionality testing for scached's cache
 features.

 unlike other functionality, cache code is really a cache that
 cant drop values in any time. private and shared storage quarantees
 all values stored to exist, cache does not and drops them freely
 on memory shortage

 cache features are common to all clients.
*/
function cache_features() {

    $count = 10000;

    //m('cache is really shared cache?');
    $conn = scache_preset(TEST_PREFIX, TEST_HOST);
    $conn1 = scache_preset(TEST_PREFIX . '_cache1', TEST_HOST);
    $conn2 = scache_preset(TEST_PREFIX . '_cache2', TEST_HOST);
    $key = TEST_ID;

    /* ...common to alla sessions */
    scache_chput($conn, "$key-testkey", 'ishere');
    if ((scache_chget($conn1, "$key-testkey") != 'ishere') ||
	(scache_chget($conn2, "$key-testkey") != 'ishere')) {
	f('shared access to keys');
    }

    m('basic cache functionality');
    m('populating values', 2);
    for($i = 0; $i < $count; $i++) {
	scache_chput($conn, "$key$i", "is $i");
    }

    m('verifing values', 2);
    for($i = 0; $i < $count; $i++) {
	if (($val = scache_chget($conn, "$key$i")) !== "is $i") {
	    f("invalid value for key $i '$val'");
	}
    }


    m('stressing backends cache-ring, part 1');
    for($i = 0; $i < $count; $i++) {
	$idx = rand() % $count;
	if (($val = scache_chget($conn, "$key$i")) !== "is $i") {
	    f("invalid value for key $i '$val'");
	}
    }

    m('stressing backends cache-ring, part 2');
    for($i = 1; $i < $count; $i++) {
	$prev = $i - 1;
	scache_chclear($conn, "$key$prev");
    }

    m('verifying removed values');
    for($i = 0; $i < ($count - 1); $i++) {
	if (($val = scache_chget($conn, "$key$i")) ||
	    scache_lasterr($conn) != SCERR_NOTEXIST) {
	    f("keys still exists: key$key: $val");
	}
    }

    m('subtree expiration');
    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	scache_chput($conn, $key, $key);
	if (scache_chget($conn, $key) != $key)
	    f("'$key' != '" . scache_chget($conn, $key) . "'", 3);
    }
    
    m('expire by unput', 2);
    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	scache_chclear($conn, "$i");
	if (scache_chget($conn, $key) == $key)
	    f("'$key' != '" . scache_chget($conn, $key) . "'", 3);
    }

    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	scache_chput($conn, $key, $key);
	if (scache_chget($conn, $key) != $key)
	    f("'$key' != '" . scache_chget($conn, $key) . "'", 3);
    }
    m('expire by put(false)', 2);
    for($i = 0; $i < 10000; $i++) {
	$key = "$i/$i/$i/$i/$i/$i/$i";
	scache_chput($conn, "$i", false);
	if (scache_chget($conn, $key) == $key)
	    f("'$key' != '" . scache_chget($conn, $key) . "'", 3);
    }

    m('complete expire from root');
    $start = utime_float(); $count = 0;
    for($a = 0; $a < 10; $a++) {
	for($b = 0; $b < 10; $b++)
	    for($c = 0; $c < 10; $c++)
		for($d = 0; $d < 10; $d++) 
		    for($e = 0; $e < 10; $e++, $count++) {
			$key = "$a/$b/$c/$d/$e";
			scache_chput($conn, $key, $key);
			if (scache_chget($conn, $key) != $key)
			    f("'$key' != '" . 
			      scache_chget($conn, $key) . 
			      "'", 3);
		    }

	if ($a > 0) m(($a * 10000) . ' values populated', 2);
    }

    m('expire from root', 2);
    scache_chclear($conn, '/');

    m('verify', 2);
    if (scache_chget($conn, '1') ||
	scache_chget($conn, '1/1') ||
	scache_chget($conn, '1/1/1') ||
	scache_chget($conn, '1/1/1/1')) {

	f("cache values still exist", 3);
    }
    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");


    m('constant storing, constant expiring');
    $val = 
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa' .
	'aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa';
    $count = 100000;
    $checkpoint = ($count / 10);
    $start = utime_float(); 
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values cached", 2);
	scache_chput($conn, "key$i", $val);
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");


    m("random cache banging");
    $count = 100000;
    $checkpoint = ($count / 10);
    $start = utime_float(); 
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values cached", 2);

	$a = rand() % 10;
	$b = rand() % 10;
	$c = rand() % 10;
	$d = rand() % 10;
	$key = "$a/$b/$c/$d";

	if (($val = scache_chget($conn, $key))) {
	    if ($val !== $key) {
		f("value mismatch: $val != $key", 2);
		return;
	    }

	    scache_chclear($conn, $key);
	} else {
	    scache_chput($conn, $key, $key);
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, passed: $keys ops/sec");
    scache_drop($conn);
}

function random_banging_r(&$rec, $pnt, $level) {

    if ($level > 100) return;
    if ($level > $rec['depth']) $rec['depth'] = $level;

    $arr = Array();
    $conn = $rec['conn'];
    $count = $rec['count'];
    $checkpoint = $count / 10;

    /* for faster loop */
    $ops = 0;
    $gets = 0;
    $unsets = 0;
    $newbrs = 0;
    $newnodes = 0;
    $descends = 0;

    while(($rec['round']++  < $count)) {

	if (($rec['round'] % $checkpoint) == 0) 
	    m("round ${rec['round']}", 2);

	$rnd = rand() % 100; /* spread */
	$key = "key$rnd";
	$path = "$pnt/$key";

	$stat = scache_stat($conn, $path); $ops++;

	if ($stat == SCNODE_BRANCH) {

	    /* descent or unlink */
	    if (!(rand() % 2)) {
		random_banging_r($rec, $path, ($level + 1));
		$descends++;
	    }
	    scache_unset($conn, $path); $ops++;
	    $unsets++;
	    
	} else if ($stat == SCNODE_VALUE) {

	    if (scache_get($conn, $path) != $path) {
		f("value mismatch: should '$path', is " . 
		  scache_get($conn, $path));
	    }
	    scache_unset($conn, $path); $ops += 2;
	    $gets++;
	    $unsets++;
	    
	} else {
	
	    /* not-exist, create file or sub */
	    if (!(rand() % 2)) {
		scache_set($conn, $path, "$path"); $ops++;
		$newnodes++;
	    } else {
		scache_set($conn, "$path/dummy", "$path"); $ops++;
		$newbrs++;
	    }
	}

	if ($level > 1) {
	    if (!(rand() % 25)) {
		$rec['ops'] += $ops;
		$rec['gets'] += $gets;
		$rec['unsets'] += $unsets;
		$rec['newbrs'] += $newbrs;
		$rec['newnodes'] += $newnodes;
		$rec['descends'] += $descends;
		return;
	    }
	}
    }
}

function random_banging() {

    $rec = 
	Array('conn' => scache_reset(TEST_PREFIX, TEST_HOST),
	      'round' => 0,
	      'depth' => 0,
	      'ops' => 0,
	      'gets' => 0,
	      'unsets' => 0,
	      'newbrs' => 0,
	      'newnodes' => 0,
	      'descends' => 0,
	      'count' => 100000);

    $start = utime_float();
    random_banging_r($rec, '', 0);
    $duration = utime_float() - $start;
    $keys = round(($rec['ops'] / $duration), 2);
    m("ok, $keys ops/sec");
    m(" ops          : " . $rec['ops']);
    m(" gets         : " . $rec['gets']);
    m(" unsets       : " . $rec['unsets']);
    m(" newbranches  : " . $rec['newbrs']);
    m(" newnodes     : " . $rec['newnodes']);
    m(" descends     : " . $rec['descends']);    
    m(" maxdepth     : " . $rec['depth']);
    m("passed");

    scache_drop($rec['conn']);
}

/* scached's internal io-loop performance testing */
function io_multiplexing() {

    $conn_count = 50;
    $reqs_count = 100000;
    $spread = $conn_count * 10;
    $checkpoint = ($reqs_count / 10);
    $conns = Array();

    m('allocating connections');
    for($i = 0; $i < $conn_count; $i++) {
	$conns[$i] = scache_connect(TEST_PREFIX, TEST_HOST_NAME);
    }

    m('random access over connections');
    for($i = 0; $i < $reqs_count; $i++) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("round $i", 2);

	$r = rand() % $spread;
	$conn = $conns[$r % $conn_count];
	scache_set($conn, "$r", "$r");
    }

    m("ok, passed");
    scache_drop($conn);
}

/* testing and example code for using scache_iov. basically scache_iov
 is the function to use always whenever possible.

 absolutely largest bottleneck in scaches performance is socket io and
 it's delays, so packing as many requests as possible on single query
 results very large improvements on performance as previous iov-tests
 show.

 you should do something like 

 list($username, $fullname, $surname) = 
     scache_iov($conn, Array(Array(SCIOP_GET, 'userinfo/login'),
                             Array(SCIOP_GET, 'userinfo/fname'),
                             Array(SCIOP_GET, 'userinfo/sname')));

 doing several requests in one query is practically as fast
 as doing one request. socket-delays are 90% of latencies
*/


function scache_iov_swiss_army_knife() {

    /* conn */
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    $counter = 'cntr' . TEST_ID;

    /* ensure old values are removed, integer are not
     in session-specific storage so scache_reset don't
     reset shared values */
    scache_vunset($conn, $counter);

    /* shared counters */
    m('testing shared counters');
    $ops = 
	Array(Array(SCIOP_VSET, $counter, 12), /* returns oldval */
	      Array(SCIOP_VADD, $counter, 2), /* returns newval */
	      Array(SCIOP_VSUB, $counter, 1), /* return newval */
	      Array(SCIOP_VGET, $counter), /* return current val */
	      Array(SCIOP_VADD, $counter, 10), /* returns newval */
	      Array(SCIOP_VUNSET, $counter)); /* returns lastval */

    $res = scache_iov($conn, $ops);
    if (sizeof($res) != sizeof($ops)) {
	f("result array size mismatch");
	return;
    }

    if (($res[0] != 0) ||
	($res[1] != 14) ||
	($res[2] != 13) ||
	($res[3] != 13) ||
	($res[4] != 23) ||
	($res[5] != 23)) {

	print_r($res);
	f("counter results mismatch");
    }


    /* session-private data */
    m("ok, testing session private operations");
    $path = "path";
    $node = "$path/node";
    $ops =
	Array(Array(SCIOP_SET, $node, '1st'), /* returns success */
	      Array(SCIOP_SET, $node, '2nd'), /* returns success */
	      Array(SCIOP_REPLACE, $node, '3rd'), /* returns success */
	      Array(SCIOP_GET, $node), /* returns '3rd' */
	      Array(SCIOP_ADD, $node, '4th'), /* returns failure */
	      Array(SCIOP_STAT, $node), /* returns SCNODE_VALUE */
	      Array(SCIOP_STAT, $path), /* returns SCNODE_BRANCH */
	      Array(SCIOP_UNSET, $node)); /* returns success */

    $res = scache_iov($conn, $ops);
    if (sizeof($res) != sizeof($ops)) {
	f("result array size mismatch");
	return;
    }

    if (($res[0] != true) ||
	($res[1] != true) ||
	($res[2] != true) ||
	($res[3] != '3rd') ||
	($res[4] != false) ||
	($res[5] != SCNODE_VALUE) ||
	($res[6] != SCNODE_BRANCH) ||
	($res[7] != true)) {

	f("node values mismatch");
	return;
    }


    /* accessing shared data */
    m("ok, testing shared data operations");
    $path = "path";
    $node = "$path/node";
    $ops =
	Array(Array(SCIOP_SHSET, $node, '1st'), /* returns success */
	      Array(SCIOP_SHSET, $node, '2nd'), /* returns success */
	      Array(SCIOP_SHREPLACE, $node, '3rd'), /* returns success */
	      Array(SCIOP_SHGET, $node), /* returns '3rd' */
	      Array(SCIOP_SHADD, $node, '4th'), /* returns failure */
	      Array(SCIOP_SHSTAT, $node), /* returns SCNODE_VALUE */
	      Array(SCIOP_SHSTAT, $path), /* returns SCNODE_BRANCH */
	      Array(SCIOP_SHUNSET, $node)); /* returns success */

    $res = scache_iov($conn, $ops);
    if (sizeof($res) != sizeof($ops)) {
	f("result array size mismatch");
	return;
    }

    if (($res[0] != true) ||
	($res[1] != true) ||
	($res[2] != true) ||
	($res[3] != '3rd') ||
	($res[4] != false) ||
	($res[5] != SCNODE_VALUE) ||
	($res[6] != SCNODE_BRANCH) ||
	($res[7] != true)) {

	f("node values mismatch on shared table");
	return;
    }

    scache_drop($conn);
}

function large_iov() {

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    $reqs = Array(Array(SCIOP_VSET, 'counter', 1));
    for($i = 0; $i < 100; $i++) 
	$reqs[] = Array(SCIOP_VADD, 'counter', 1);
    $reqs[] = Array(SCIOP_VGET, 'counter');

    /* multiple times */
    $res = scache_iov($conn, $reqs);
    $res = scache_iov($conn, $reqs);
    $res = scache_iov($conn, $reqs);
    $res = scache_iov($conn, $reqs);
    $res = scache_iov($conn, $reqs);

    if ($res[sizeof($res) - 1] != 101) {
	f('values mismatch, expect 101 got : ' . $res[sizeof($res) - 1]);
    } else {
	m('ok, 100 iops, result matches');
    }
    
    scache_drop($conn);
    return;
}

function ring_basics() {

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    
    m('push values');
    if (!(scache_rnpush($conn, TEST_ID, 0) &&
	  scache_rnpush($conn, TEST_ID, 1) &&
	  scache_rnpush($conn, TEST_ID, 2))) {
	f(scache_strerror(scache_lasterr($conn)));
    }

    m('unshift values');
    if (!(scache_rnunshift($conn, TEST_ID, 5) &&
	  scache_rnunshift($conn, TEST_ID, 4) &&
	  scache_rnunshift($conn, TEST_ID, 3))) {
	f(scache_strerror(scache_lasterr($conn)));
    }

    m('ring size');
    if (!(scache_rnsize($conn, TEST_ID) == 6)) {
	f('should be 6, got ' . scache_rnsize($conn, TEST_ID));
    }

    m('rotate forward');
    if (!(scache_rnrotf($conn, TEST_ID, 3) &&
	  scache_rnrotf($conn, TEST_ID, -4))) {
	f(scache_strerror(scache_lasterr($conn)));
    }

    m('rotage backwards');
    if (!(scache_rnrotb($conn, TEST_ID, 1) &&
	  scache_rnrotb($conn, TEST_ID, -5))) {
	f(scache_strerror(scache_lasterr($conn)));
    }

    m('internal ordering');
    for($i = 0; $i < scache_rnsize($conn, TEST_ID); $i++) {
	if (($n = scache_rnget($conn, TEST_ID)) !== $i) {
	    f("expected $i got $n");
	} 
	scache_rnrotf($conn, TEST_ID, 1);
    }

    m('pop/unshift');
    if ((($n = scache_rnpop($conn, TEST_ID)) !== ($e = 5)) ||
	(($n = scache_rnshift($conn, TEST_ID)) !== ($e = 0)) ||
	(($n = scache_rnpop($conn, TEST_ID)) !== ($e = 4))) {
	f("expected $e, got $n");
    }

    m('remaining size');
    if (($n = scache_rnsize($conn, TEST_ID)) != 3) {
	f("expected 4, got $n");
    }

    m('clearing');
    if (!scache_rnclear($conn, TEST_ID)) {
	f(scache_strerror(scache_lasterr($conn)));
    }

    m('clearing unexistent');
    if (!scache_rnclear($conn, TEST_ID) &&
	(($e = scache_lasterr($conn)) != SCERR_NOTEXIST)) {
	f(scache_strerror($e));
    }

    m('getting, shifting, popping unexistent');
    if (scache_rnget($conn, TEST_ID) ||
	(($e = scache_lasterr($conn)) != SCERR_NOTEXIST) ||
	scache_rnshift($conn, TEST_ID) ||
	(($e = scache_lasterr($conn)) != SCERR_NOTEXIST) ||
	scache_rnpop($conn, TEST_ID) ||
	(($e = scache_lasterr($conn)) != SCERR_NOTEXIST)) {
	
	f(scache_strerror($e));
    }
}

function singlecall_ring_stressing() {

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    $count = 100000;
    $checkpoint = ($count / 10);
    m('populating ring');
    $start = utime_float();
    for($i = 0; $i < $count; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values pushed", 2);
	scache_rnpush($conn, TEST_ID, $i);
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    $rcount = 200000;
    $checkpoint = ($rcount / 10);
    m('rotate and verify ring');
    for($i = 0; $i < $rcount; $i++) {
	if ($i && ($i % $checkpoint) == 0) 
	    m("$i rounds", 2);
	if (($val = scache_rnget($conn, TEST_ID)) != ($i % $count)) {
	    scache_rnclear($conn, TEST_ID);
	    f("expected $val, got " . ($i % $count));
	}
	scache_rnrotf($conn, TEST_ID, 1);
    }

    $duration = utime_float() - $start;
    $keys = round((($rcount * 2) / $duration), 2);
    m("ok, $keys ops/sec");

    m('clearing ring');
    scache_rnclear($conn, TEST_ID);
}

function ring_stressing_with_iov() {

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    $count = 200000;
    $checkpoint = ($count / 10);
    m('populating ring (10op iov)');
    $start = utime_float();
    for($i = 0; $i < $count; $i += 10) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values pushed", 2);
	scache_iov($conn, 
		   Array(Array(SCIOP_RNPUSH, TEST_ID, ($i + 0)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 1)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 2)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 3)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 4)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 5)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 6)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 7)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 8)),
			 Array(SCIOP_RNPUSH, TEST_ID, ($i + 9))));
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    m('purge ring (10op iov)');
    $start = utime_float();
    for($i = 0; $i < $count; $i += 10) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values shifted", 2);
	$res = scache_iov($conn, 
			  Array(Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID),
				Array(SCIOP_RNSHIFT, TEST_ID)));
	if ($res[0] !== $i) {
	    f("expected $i, got " . $res[0]);
	}
    }

    $duration = utime_float() - $start;
    $keys = round(($count / $duration), 2);
    m("ok, $keys ops/sec");

    m('verify that ring is empty');
    if (scache_rnsize($conn, TEST_ID) !== 0) {
	f("expected 0, got " . scache_rnsize($conn, TEST_ID));
    }

    m('clearing ring');
    scache_rnclear($conn, TEST_ID);
}

function protocol_limits() {

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);

    $key = str_repeat('a', 16384);
    $lval = str_repeat('a', 524288);
    $sval = '1';

    m('uplimit key');
    scache_set($conn, $key, $sval);
    if (scache_get($conn, $key) !== $sval) {
	f('retval mismatch');
    }

    m('uplimit value');
    scache_set($conn, $key, $lval);
    if (scache_get($conn, $key) !== $lval) {
	f('retval mismatch');
    }

    $count = 10000;
    $checkpoint = ($count / 10);

    m('memleak detect and pushing very large packets with iov');
    $start = utime_float();
    for($i = 0; $i < $count; $i += 10) {

	if ($i && ($i % $checkpoint) == 0) 
	    m("$i values pushed", 2);
	$res =
	    scache_iov($conn, 
		       Array(Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key),
			     Array(SCIOP_SET, $key, $lval),
			     Array(SCIOP_GET, $key)));
	if (($res[1] !== $lval)) {
	    f('retval mismatch');
	}
    }
    $duration = utime_float() - $start;
    $keys = round((($count) / $duration), 2);
    m("ok, $keys ops/sec");
}

function accumulator() {

    $id = TEST_ID;
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    scache_set($conn, "$id/a1", TEST_ID);

    $iops = Array(Array(SCIOP_GET, "$id/a1"),
		  Array(SCIOP_SET, "$id/a2"),
		  Array(SCIOP_SHSET, "$id/a3"),
		  Array(SCIOP_CHPUT, "$id/a4"),
		  Array(SCIOP_RNPUSH, "$id/a5"),
		  Array(SCIOP_GET, "$id/a2"),
		  Array(SCIOP_SHGET, "$id/a3"),
		  Array(SCIOP_CHGET, "$id/a4"),
		  Array(SCIOP_RNPOP, "$id/a5"));

    $rv = scache_iov($conn, $iops);

    if (($rv[0] != TEST_ID) ||
	($rv[5] != TEST_ID) ||
	($rv[6] != TEST_ID) ||
	($rv[7] != TEST_ID) ||
	($rv[8] != TEST_ID)) {
	f('retval mismatch');
    } else {
	m("ok");
    }

    scache_shclear($conn, $id);
    scache_drop($conn);
}

function merry_go_round() {

}

function escape_now() {
    m("woo hoo, i'm free!!!");
    die();
}

function scache_iov_lasterr() {

    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    $iops = Array(Array(SCIOP_SET, '/exist', 'true'),
		  Array(SCIOP_LASTERR),
		  Array(SCIOP_GET, '/not-exist'),
		  Array(SCIOP_LASTERR),
		  Array(SCIOP_ADD, '/exist', 'add-never-overwrites'),
		  Array(SCIOP_LASTERR),
		  Array(SCIOP_REPLACE, '/not-exist', 'existing-required'),
		  Array(SCIOP_LASTERR));

    $res = scache_iov($conn, $iops);
    if (($res[1] == SCERR_SUCCESS) &&
	($res[3] == SCERR_NOTEXIST) &&
	($res[5] == SCERR_EXIST) &&
	($res[7] == SCERR_NOTEXIST)) {

	m('ok, passed');
    } else {
	f('failed: got ' .
	  '1:' . scache_strerror($res[1]) .
	  '3:' . scache_strerror($res[3]) .
	  '5:' . scache_strerror($res[5]) .
	  '7:' . scache_strerror($res[7]));
    }
}

function final_statistics() {
    
    $conn = scache_reset(TEST_PREFIX, TEST_HOST);
    if (is_array(scache_srvstats($conn))) {
	m('ok, statistics');
    } else {
	f('statistics');
    }
}

function run_tests() {

    $tests = 
	Array(
	      'dump_functions',
	      'connect_unix',
	      'connect_ipv4',
	      //'connect_ipv6',
	      'connect_byname_localhost',
	      'connect_persistent_index',
	      'fast_connections',
	      'flood_connections',
	      'open_with_secret_key',
	      'session_status_result',
	      'large_session_tree',
	      'getset_options',
	      'session_timeout',
	      'modifying_session_timeout',
	      'session_renaming',
	      'session_dropping',
	      'unsetting_values',
	      'collect_operations',
	      'wide_keyspace',
	      'deep_tree',
	      'stat_paths',
	      'error_codes',
	      'random_access_flat',
	      'get_set_performance_unix',
	      'get_set_performance_localhost',
	      'get_set_high_performance_iov_1key',
	      'get_set_high_performance_iov_100000key',
	      'shared_counters',
	      'set_add_replace_logic',
	      'cache_features',
	      'random_banging',
	      'io_multiplexing',
	      'scache_iov_swiss_army_knife',
	      'large_iov',
	      'protocol_limits',
	      'ring_basics',
	      'singlecall_ring_stressing',
	      'ring_stressing_with_iov',
	      'accumulator',
	      'scache_iov_lasterr',
	      'final_statistics');

    foreach($tests as $test) {
	echo $test . "....\n";
	call_user_func($test);
    }

}

run_tests();

/*
  Local variables: ***
  mode: php ***
  c-basic-offset: 4 ***
  End: ***
*/
