/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */

#ifndef _HAVE_PHP_SCACHE_PROTO_H_
#define _HAVE_PHP_SCACHE_PROTO_H_

#ifdef HAVE_CONFIG_H
# ifndef _CONFIG_H_
#  include "../config.h"
# endif
#endif

#define SC_HDR_PACK_H(x, len) ((((x) & 0xff) << 24) | ((len) & 0xffffff))
#define SC_HDR_PACK_R(x, len) ((((x) & 0xff) << 24) | ((len) & 0xffffff))


/* ==================================
 *  PROTOCOL DESCRIPTION
 * ==================================
 *
 * Protocol header
 *  
 *   U8     : number-of-request in packet. highest bit is reserved
 *            for future use, so number of requests is 1 - 127
 *   U24    : total-request-length including this header. backend
 *            reads request in complete before processing and expects
 *            total-request-length bytes from wire. 
 *   S32    : session-index-slot, -1 when unknown
 *   U16    : session-flags
 *   U16    : session-user-length including nul-termination
 *   U16    : session-tag-length including nul-termination
 *   U16    : session-secret-length including nul-termination
 *   U8 * session-tag-lenght : session-tag
 *
 * Request
 *   U8     : request-type, (SCREQ_TYPE_xx)
 *   U24    : request-length excluding request header
 *   U8 * requests-length : request data
 *            
 * */

#define SCACHE_PTO_MIN_TAGLEN       1
#define SCACHE_PTO_MAX_TAGLEN       4096
#define SCACHE_PTO_MIN_SECRETLEN    0
#define SCACHE_PTO_MAX_SECRETLEN    255
#define SCACHE_PTO_MIN_AUTHLEN      0
#define SCACHE_PTO_MAX_AUTHLEN      255

#define SCACHE_PTO_MAX_KEYLEN       16384
#define SCACHE_PTO_MAX_VALUELEN     524288
#define SCACHE_PTO_MAX_REQLEN       16777215

#define SCACHE_PTO_MIN_PORT         1
#define SCACHE_PTO_MAX_PORT         65535

/* idle session expire */
#define SCACHE_PTO_MIN_EXPIRE       5
#define SCACHE_PTO_DEF_EXPIRE       1800
#define SCACHE_PTO_MAX_EXPIRE       31536000 /* ~year */



/* strings */
#define SCACHE_PTO_UNKNOWN_CLIENT   "*"
#define SCACHE_PTO_UNKNOWN_OWNER    "*"

/* response flags */
#define SCACHE_QFLAG_MASK           0x0000ffff
#define SCACHE_RFLAG_MASK           0xffff0000
#define SCACHE_QFLAG_OVERWRITE      (1 << 0)

/* options */
#define SCACHE_OPT_EXPIRES          (1 << 1)
#define SCACHE_OPT_KEEPONELIMIT     (1 << 2)

#define SCACHE_IOP_RENAME           SCREQ_TYPE_SESSION_RENAME

#define SCACHE_IOP_SET              SCREQ_TYPE_SET
#define SCACHE_IOP_ADD              SCREQ_TYPE_ADD
#define SCACHE_IOP_REPLACE          SCREQ_TYPE_REPLACE
#define SCACHE_IOP_UNSET            SCREQ_TYPE_UNSET
#define SCACHE_IOP_GET              SCREQ_TYPE_GET
#define SCACHE_IOP_STAT             SCREQ_TYPE_STAT
#define SCACHE_IOP_COLL             SCREQ_TYPE_COLL

#define SCACHE_IOP_GETOPT           SCREQ_TYPE_GETOPT
#define SCACHE_IOP_SETOPT           SCREQ_TYPE_SETOPT

#define SCACHE_IOP_SHSET            SCREQ_TYPE_SHSET
#define SCACHE_IOP_SHADD            SCREQ_TYPE_SHADD
#define SCACHE_IOP_SHREPLACE        SCREQ_TYPE_SHREPLACE
#define SCACHE_IOP_SHUNSET          SCREQ_TYPE_SHUNSET
#define SCACHE_IOP_SHGET            SCREQ_TYPE_SHGET
#define SCACHE_IOP_SHSTAT           SCREQ_TYPE_SHSTAT
#define SCACHE_IOP_SHCOLL           SCREQ_TYPE_SHCOLL

#define SCACHE_IOP_CHGET            SCREQ_TYPE_CHGET
#define SCACHE_IOP_CHPUT            SCREQ_TYPE_CHPUT
#define SCACHE_IOP_CHCLEAR          SCREQ_TYPE_CHCLEAR

#define SCACHE_IOP_VADD             SCREQ_TYPE_VADD
#define SCACHE_IOP_VSUB             SCREQ_TYPE_VSUB
#define SCACHE_IOP_VGET             SCREQ_TYPE_VGET
#define SCACHE_IOP_VSET             SCREQ_TYPE_VSET
#define SCACHE_IOP_VUNSET           SCREQ_TYPE_VUNSET
#define SCACHE_IOP_VBRK             SCREQ_TYPE_VBRK

#define SCACHE_IOP_RNPUSH           SCREQ_TYPE_RNPUSH
#define SCACHE_IOP_RNPOP            SCREQ_TYPE_RNPOP
#define SCACHE_IOP_RNSHIFT          SCREQ_TYPE_RNSHIFT
#define SCACHE_IOP_RNUNSHIFT        SCREQ_TYPE_RNUNSHIFT
#define SCACHE_IOP_RNGET            SCREQ_TYPE_RNGET
#define SCACHE_IOP_RNROTF           SCREQ_TYPE_RNROTF
#define SCACHE_IOP_RNROTB           SCREQ_TYPE_RNROTB
#define SCACHE_IOP_RNSIZE           SCREQ_TYPE_RNSIZE
#define SCACHE_IOP_RNCLEAR          SCREQ_TYPE_RNCLEAR

#define SCACHE_IOP_LASTERR          SCREQ_TYPE_LASTERR

enum scache_nodetypes_t {
    SCNODE_NONEXISTENT = 0,
    SCNODE_EMPTY,
    SCNODE_BRANCH,
    SCNODE_VALUE
};

enum scache_error_t {
    SCERR_SUCCESS          = 0,
    SCERR_PROTOCOL         = 1,
    SCERR_EXIST            = 2,
    SCERR_NOTEXIST         = 3,
    SCERR_SYSTEM           = 4,
    SCERR_INVALID_REQUEST  = 5,
    SCERR_UNKNOWN_REQUEST  = 6,
    SCERR_LIMITS_REACHED   = 7,
    SCERR_NOT_CONNECTED    = 8,
    SCERR_NO_SESSION       = 9,
    SCERR_NO_ACCESS        = 10,
    SCERR_BREAK            = 11,
    SCERR_INVALID_FORMAT   = 12
};

enum scache_reqtype_t {  /* TODO: kiinteiksi */
    SCREQ_TYPE_NONE            = 0,
    SCREQ_TYPE_SRVSTATS        = 1,

    SCREQ_TYPE_SESSION_CREATE  = 2,
    SCREQ_TYPE_SESSION_DROP    = 3,
    SCREQ_TYPE_SESSION_RENAME  = 4,

    SCREQ_TYPE_SET             = 5,
    SCREQ_TYPE_ADD             = 6,
    SCREQ_TYPE_REPLACE         = 7,
    SCREQ_TYPE_UNSET           = 8,
    SCREQ_TYPE_GET             = 9,
    SCREQ_TYPE_STAT            = 10,
    SCREQ_TYPE_COLL            = 11,

    SCREQ_TYPE_GETOPT          = 12,
    SCREQ_TYPE_SETOPT          = 13,

    SCREQ_TYPE_SHSET           = 14,
    SCREQ_TYPE_SHADD           = 15,
    SCREQ_TYPE_SHREPLACE       = 16, 
    SCREQ_TYPE_SHUNSET         = 17,
    SCREQ_TYPE_SHGET           = 18,
    SCREQ_TYPE_SHSTAT          = 19,
    SCREQ_TYPE_SHCOLL          = 20,

    SCREQ_TYPE_CHGET           = 21,
    SCREQ_TYPE_CHPUT           = 22,
    SCREQ_TYPE_CHCLEAR         = 23,

    SCREQ_TYPE_VADD            = 24,
    SCREQ_TYPE_VSUB            = 25,
    SCREQ_TYPE_VGET            = 26,
    SCREQ_TYPE_VSET            = 27,
    SCREQ_TYPE_VUNSET          = 28,
    SCREQ_TYPE_VBRK            = 29,

    SCREQ_TYPE_SESGET          = 30,
    SCREQ_TYPE_SESPUT          = 31,
    SCREQ_TYPE_SESDROP         = 32,

    SCREQ_TYPE_RNPUSH          = 33,
    SCREQ_TYPE_RNPOP           = 34,
    SCREQ_TYPE_RNSHIFT         = 35,
    SCREQ_TYPE_RNUNSHIFT       = 36,
    SCREQ_TYPE_RNGET           = 37,
    SCREQ_TYPE_RNROTF          = 38,
    SCREQ_TYPE_RNROTB          = 39,
    SCREQ_TYPE_RNSIZE          = 40,
    SCREQ_TYPE_RNCLEAR         = 41,

    SCREQ_TYPE_LASTERR         = 42,

    SCREQ_TYPE_LAST            = 43
};


#endif /* _HAVE_PHP_SCACHE_PROTO_H_ */
