/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */


/* 
 * NOTES :
 *
 * - parts of protocol widened afterwards to 64bit, some
 *   old 32bit comments still exist
 */


// TODO: input-tarkistusten varmistukset
// TODO: sessionhandlerista puuttuu expire -> repack

#include "scache.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>


struct sc_reqhdr_t {

    int index;
    int flags;

    int tag_len;
    int sec_len;
    int ath_len;

    char *tag;
    char *secret;
    char *auth;
    const char *peer;

    enum scache_error_t scerr;
    struct sc_session_t *session;
};

static enum scache_error_t sc_scerr;
static int sc_vbreak = 0;
static char *sc_accumulator = NULL;
static int sc_accumulator_len = 0;

static inline void *
accum_dup(void *orig, int len) {

    if (orig && (sc_accumulator == orig)) {

	/* already in accumulator, reference */
	assert((len == sc_accumulator_len));
	sc_ref(sc_accumulator);

    } else {
	
	/* replace previous */
	if (sc_accumulator) sc_unref(sc_accumulator);
	if (orig) {
	    sc_accumulator_len = len;
	    sc_accumulator = sc_memdup(orig, len);
	    sc_ref(sc_accumulator);
	} else {
	    sc_accumulator = NULL;
	    sc_accumulator_len = 0;
	}
    }

    return sc_accumulator;
}

static inline void *
accum_ref(void *addr, int len) {

    if (sc_accumulator) sc_unref(sc_accumulator);
    sc_accumulator = addr;
    sc_accumulator_len = len;
    if (addr) sc_ref(addr);

    return addr;
}


static struct sc_wblock_t *
wblock_alloc() {

    struct sc_wblock_t *wblock;

#ifdef STAT_ME_HARDER
    sc_stats->wblock_allocs++;
#endif

    wblock = sc_alloc(sizeof(struct sc_wblock_t));
    return memset(wblock, 0, sizeof(*wblock));
}


void
sc_dump_packet(int length, char *req) {

    int i;
    
    printf("DUMP: %i", length);
    for(i = 0; i < length; i++) {

	if (!(i % 8)) printf("\n%p: ", req + i);
	printf(" %.2x", (unsigned int)(req[i] & 0xff));
    }

    printf("\n");
}



static struct sc_wblock_t *
handle_session_create(struct sc_reqhdr_t *rhdr, int flags,
		      int expires) {

    /* if SCACHE_QFLAG_OVERWRITE specified, we end up
       overwriting session */
    if (rhdr->session && !(flags & SCACHE_QFLAG_OVERWRITE)) {

	rhdr->scerr = SCERR_SUCCESS;
	return NULL; /* ok, not error */
    }

    rhdr->session = session_create(rhdr->tag,
				   rhdr->secret,
				   rhdr->auth,
				   rhdr->peer,
				   flags,
				   expires);

    if (!rhdr->session) {

	/* create failed, map to protocol error */
	switch(errno) {
	case EINVAL:
	    rhdr->scerr = SCERR_PROTOCOL;
	    break;
	case ENOSPC:
	    rhdr->scerr = SCERR_LIMITS_REACHED;
	    break;
	case EEXIST:
	    rhdr->scerr = SCERR_EXIST;
	    break;
	default:
	    LOG_IMPOSSIBILITY;
	    rhdr->scerr = SCERR_SYSTEM;
	    break;
	}

    } else {

	rhdr->scerr = SCERR_SUCCESS;
    }

    /* no content */
    return NULL;
}


static struct sc_wblock_t *
decode_session_create(struct sc_reqhdr_t *rhdr,
		      char *req, int length) {
    
    int s_expires;
    int s_flags;

    const int header_static_length = 8;

    /* packet 
     *
     *  u32      : expires
     *  u32      : flags
     *
     */

    if (length < header_static_length) {
	rhdr->scerr = SCERR_PROTOCOL;
	return NULL;
    }

    /* parse */
    length -= header_static_length;
    s_expires  = sc_ntohl(req); req += 4;
    s_flags    = sc_ntohl(req); req += 4;

    /* verify - expires */
    if (!s_expires) s_expires = SCACHE_PTO_DEF_EXPIRE;
    if (s_expires < SCACHE_PTO_MIN_EXPIRE) s_expires = SCACHE_PTO_MIN_EXPIRE;
    if (s_expires > SCACHE_PTO_MAX_EXPIRE) s_expires = SCACHE_PTO_MAX_EXPIRE;
    
    /* exec */
    return handle_session_create(rhdr, s_flags, s_expires);
}

static struct sc_wblock_t *
handle_session_drop(struct sc_reqhdr_t *rhdr) {

    if (rhdr->session) {

	if (!session_drop(rhdr->session)) {
	    rhdr->scerr = SCERR_SUCCESS;
	} else {
	    LOG_IMPOSSIBILITY;
	    rhdr->scerr = SCERR_SYSTEM;
	}

    } else {
	rhdr->scerr = SCERR_NO_SESSION;
    }

    return NULL;
}

static struct sc_wblock_t *
decode_session_drop(struct sc_reqhdr_t *rhdr,
		    char *req, int length) {


    if (length == 0) {

	/* it's ok, very simple request you know... */
	return handle_session_drop(rhdr);
    }

    /* ok or not, just push scerror back */
    return NULL;
}

static struct sc_wblock_t *
handle_getopt(struct sc_reqhdr_t *rhdr, int op, int opt) {

    int rval;
    char *buffer;
    struct sc_wblock_t *wblock;

    switch(opt) {
    case SCACHE_OPT_EXPIRES:
	rval = rhdr->session->maxidle;
	break;
    case SCACHE_OPT_KEEPONELIMIT:
	rval = !!(rhdr->session->flags & SCACHE_OPT_KEEPONELIMIT);
	break;
    default:
	rhdr->scerr = SCERR_INVALID_REQUEST;
	return NULL;
    }

    wblock = wblock_alloc();
    wblock->length = 8;
    wblock->base = buffer = sc_alloc(wblock->length);

    buffer = sc_htonl(buffer, SC_HDR_PACK_R(op, wblock->length));
    buffer = sc_htonl(buffer, rval);
    return wblock;
}

static struct sc_wblock_t *
handle_setopt(struct sc_reqhdr_t *rhdr, int opt, int val) {

    struct sc_wblock_t *wblock;

    /* old values */
    wblock = handle_getopt(rhdr, SCREQ_TYPE_SETOPT, opt);
    if (!wblock) return NULL;

    switch(opt) {
    case SCACHE_OPT_EXPIRES:

	if (val < SCACHE_PTO_MIN_EXPIRE) val = SCACHE_PTO_MIN_EXPIRE;
	if (val > SCACHE_PTO_MAX_EXPIRE) val = SCACHE_PTO_MAX_EXPIRE;
	rhdr->session->maxidle = val;
	
	/* DEP:SESSION_EXPIRATION */
	rhdr->session->expires = sc_lazytime + rhdr->session->maxidle;
	break;

    case SCACHE_OPT_KEEPONELIMIT:
	if (val) {
	    rhdr->session->flags |= SCACHE_OPT_KEEPONELIMIT;
	} else {
	    rhdr->session->flags &= ~SCACHE_OPT_KEEPONELIMIT;
	}
	break;
    }

    return wblock;
}

static struct sc_wblock_t *
decode_getopt(struct sc_reqhdr_t *rhdr,
	      char *req, int length) {

    int opt;
    const int header_static_length = 4;

    /* requres session */
    if (!rhdr->session) {
	rhdr->scerr = SCERR_NO_SESSION;
	return NULL;
    }

    if (length < header_static_length) {
	rhdr->scerr = SCERR_PROTOCOL;
	return NULL;
    }

    opt = sc_ntohl(req); req += 4;
    return handle_getopt(rhdr, SCREQ_TYPE_GETOPT, opt);
}

static struct sc_wblock_t *
decode_setopt(struct sc_reqhdr_t *rhdr,
	      char *req, int length) {

    int opt, val;
    const int header_static_length = 8;
    
    /* requres session */
    if (!rhdr->session) {
	rhdr->scerr = SCERR_NO_SESSION;
	return NULL;
    }

    if (length < header_static_length) {
	rhdr->scerr = SCERR_PROTOCOL;
	return NULL;
    }

    opt = sc_ntohl(req); req += 4;
    val = sc_ntohl(req); req += 4;
    return handle_setopt(rhdr, opt, val);
}


static struct sc_wblock_t *
decode_srvstats(struct sc_reqhdr_t *rhdr,
		char *req, int length) {


    /* requres session */
    if (!rhdr->session) {
	rhdr->scerr = SCERR_NO_SESSION;
	return NULL;
    }
    
    if (rhdr->session->partition &&
	rhdr->session->partition->allow_stats) {

	char *stats = config_dump_stats(rhdr->session);
	if (stats) {

	    int stats_len;
	    struct sc_wblock_t *wblock;

	    /* header */
	    stats_len = strlen(stats) + 1;
	    wblock = wblock_alloc();
	    wblock->length = 4;
	    wblock->base = sc_alloc(wblock->length);
	    sc_htonl(wblock->base, 
		     SC_HDR_PACK_R(SCREQ_TYPE_SRVSTATS, 
				   wblock->length + stats_len));

	    /* fake-serialized array */
	    wblock->next = wblock_alloc();
	    wblock->next->length = strlen(stats) + 1;
	    wblock->next->base = stats;

	    return wblock;

	} else {

	    /* never happens */
	    LOG_IMPOSSIBILITY;
	    rhdr->scerr = SCERR_SYSTEM;
	}
    } else {
	rhdr->scerr = SCERR_NO_ACCESS;
    }

    return NULL;
}





/* ++++++++++++++++++++++++++++++++++++
 * SCREQ_TYPE_(SH)
 * ++++++++++++++++++++++++++++++++++++ */
#define CALL_KD_FCN(_FNAME, _CALL, _ERR)	\
    static struct sc_wblock_t *			\
    _FNAME(struct sc_reqhdr_t *req, int vlen,	\
	   char *key, char *val) {		\
						\
	if (_CALL(req->session, key,					\
		  accum_dup(val, vlen), vlen) < 0) {			\
	    switch(errno) {						\
	    case ENOSPC:						\
		req->scerr = SCERR_LIMITS_REACHED;			\
		break;							\
	    default:							\
		req->scerr = _ERR;					\
		break;							\
	    }								\
	} else {							\
	    req->scerr = SCERR_SUCCESS;					\
	}								\
	return NULL;							\
    }

#define CALL_KD_SHFCN(_FNAME, _CALL, _ERR)	\
    static struct sc_wblock_t *			\
    _FNAME(struct sc_reqhdr_t *req, int vlen,	\
	   char *key, char *val) {				     \
								     \
	struct sc_session_t *shared =					\
	    session_host_shared(req->session);				\
									\
	if (shared) {							\
	    req->scerr = _CALL(shared, key,				\
			       accum_dup(val, vlen), vlen) < 0 ?	\
		_ERR : SCERR_SUCCESS;					\
	} else {							\
	    req->scerr = (errno == ENOSPC) ?				\
		SCERR_LIMITS_REACHED : SCERR_NO_ACCESS;			\
	}								\
	return NULL;							\
    }


CALL_KD_FCN(handle_add, session_add, SCERR_EXIST);
CALL_KD_FCN(handle_set, session_set, SCERR_SYSTEM);
CALL_KD_FCN(handle_replace, session_replace, SCERR_NOTEXIST);

CALL_KD_SHFCN(handle_shadd, session_add, SCERR_EXIST);
CALL_KD_SHFCN(handle_shset, session_set, SCERR_SYSTEM);
CALL_KD_SHFCN(handle_shreplace, session_replace, SCERR_NOTEXIST);



static struct sc_wblock_t *
handle_chput(struct sc_reqhdr_t *req, int vlen,
	     char *key, char *val) {

    if (cache_put(req->session, key, 
		  accum_dup(val, vlen), vlen) < 0) {
	
	req->scerr = (errno == EACCES) ? 
	    SCERR_NO_ACCESS : SCERR_LIMITS_REACHED;
    } else {
	req->scerr = SCERR_SUCCESS;
    }

    /* not error, rely alse success returning to 
       automatic error creation mechanism */
    return NULL;
}


#define CALL_WRAP(_FNAME, _WRAP, _OP)		\
    static struct sc_wblock_t *			\
    _FNAME(struct sc_reqhdr_t *req, int vlen,	\
	   char *key, char *val) {		\
	return _WRAP(req->session, req, vlen, key, val, _OP);	\
    }

#define CALL_SHWRAP(_FNAME, _WRAP, _OP)			\
    static struct sc_wblock_t *				\
    _FNAME(struct sc_reqhdr_t *req, int vlen,		\
	   char *key, char *val) { \
									\
	struct sc_session_t *shared =					\
	    session_host_shared(req->session);				\
	if (shared) {							\
	    return _WRAP(shared, req, vlen, key, val, _OP);		\
	}								\
	req->scerr = SCERR_NO_ACCESS;					\
	return NULL;							\
    }


/* ++++++++++++++++++++++++++++++++++++
 * SCREQ_TYPE_UNSET
 * ++++++++++++++++++++++++++++++++++++ */
static struct sc_wblock_t *
handle_chclear(struct sc_reqhdr_t *req, int vlen,
	       char *key, char *val) {

    req->scerr = cache_clear(req->session, key) < 0 ?
	SCERR_NO_ACCESS : SCERR_SUCCESS;

    /* not error, rely alse success returning to 
       automatic error creation mechanism */
    return NULL;
}

static struct sc_wblock_t *
handle_unset_r(struct sc_session_t *session,
	       struct sc_reqhdr_t *req, int vlen,
	       char *key, char *val, int op) {

    if ((session_unset(session, key) < 0) &&
	(errno != ENOENT)) {
	
	LOG_IMPOSSIBILITY;
	req->scerr = SCERR_SYSTEM;

    } else {

	/* whether missing */
	if (errno == ENOENT) {
	    req->scerr = SCERR_NOTEXIST;
	} else {
	    req->scerr = SCERR_SUCCESS;
	}
    }

    /* rely also for success returning to 
       automatic error creation mechanism */
    return NULL;
}

CALL_WRAP(handle_unset, handle_unset_r, SCREQ_TYPE_UNSET);
CALL_SHWRAP(handle_shunset, handle_unset_r, SCREQ_TYPE_SHUNSET);

/* ++++++++++++++++++++++++++++++++++++
 * SCREQ_TYPE_STAT
 * ++++++++++++++++++++++++++++++++++++ */

static struct sc_wblock_t *
handle_stat_r(struct sc_session_t *session,
	      struct sc_reqhdr_t *req, int vlen,
	      char *key, char *val, int op) {

    int datalen;
    char *data;
    char *buffer;

    struct sc_wblock_t *wblock;
    enum scache_nodetypes_t nodetype;

    nodetype = session_stat(session, key, &data, &datalen);

    /* accumulator */
    accum_ref(data, datalen);

    wblock = wblock_alloc();
    wblock->length = 8;
    wblock->base = buffer = sc_alloc(wblock->length);

    buffer = sc_htonl(buffer, SC_HDR_PACK_R(op, wblock->length));
    buffer = sc_htonl(buffer, nodetype);

    req->scerr = SCERR_SUCCESS;
    return wblock;
}

CALL_WRAP(handle_stat, handle_stat_r, SCREQ_TYPE_STAT);
CALL_SHWRAP(handle_shstat, handle_stat_r, SCREQ_TYPE_SHSTAT);


/* ++++++++++++++++++++++++++++++++++++
 * SCREQ_TYPE_COLL
 * ++++++++++++++++++++++++++++++++++++ */

static struct sc_wblock_t *
handle_coll_r(struct sc_session_t *session,
	      struct sc_reqhdr_t *req, int vlen,
	      char *from, char *to, int shared) {

    char *res;

    /* if allowed */
    if (!session->partition ||
	(!shared && !session->partition->allow_collect) ||
	(shared && !session->partition->allow_shcollect)) {

	req->scerr = SCERR_NO_ACCESS;
	return NULL;
    }

    if (!(res = session_collect(session, from))) {
	req->scerr = SCERR_NOTEXIST;
    } else {

	int len = strlen(res);
	if (session_set(session, to, accum_dup(res, len), len) < 0) {
	    switch(errno) {
	    case ENOSPC:
		req->scerr = SCERR_LIMITS_REACHED;
		break;
	    default:
		req->scerr = SCERR_SYSTEM;
		break;
	    }
	} else {
	    req->scerr = SCERR_SUCCESS;
	}
    }

    return NULL;
}

CALL_WRAP(handle_coll, handle_coll_r, 0);
CALL_SHWRAP(handle_shcoll, handle_coll_r, 1);

/* ++++++++++++++++++++++++++++++++++++
 * SCREQ_TYPE_GET
 * ++++++++++++++++++++++++++++++++++++ */

static struct sc_wblock_t *
handle_get_packet(int op, int datalen, char *data) {

    char *buffer;
    struct sc_wblock_t *wblock;

    wblock = wblock_alloc();
    wblock->length = 4;
    wblock->base = buffer = sc_alloc(wblock->length);
    buffer = sc_htonl(buffer, 
		      SC_HDR_PACK_R(op, wblock->length + datalen));

    
    /* data in distinct wblock to avoid unnecessary inmem copying */
    wblock->next = wblock_alloc();
    wblock->next->length = datalen;
    wblock->next->base = data;
    sc_ref(data); /* writer unreferences */

    return wblock;
}

/* common response-packet */
static struct sc_wblock_t *
handle_get_r(struct sc_session_t *session,
	     struct sc_reqhdr_t *req, int vlen,
	     char *key, char *val, int op) {

    int datalen;
    char *data;

    if (session_get(session, key, &data, &datalen) < 0) {
	req->scerr = SCERR_NOTEXIST;
	return NULL;
    }

    assert((datalen < 0xffffff));

    req->scerr = SCERR_SUCCESS;
    return handle_get_packet(op, datalen, accum_ref(data, datalen));
}

CALL_WRAP(handle_get, handle_get_r, SCREQ_TYPE_GET);
CALL_SHWRAP(handle_shget, handle_get_r, SCREQ_TYPE_SHGET);

static struct sc_wblock_t *
handle_chget(struct sc_reqhdr_t *req, int vlen,
	     char *key, char *val) {
    
    int datalen;
    char *data;

    if (cache_get(req->session, key, &data, &datalen) < 0) {
	req->scerr = (errno == EACCES) ? 
	    SCERR_NO_ACCESS : SCERR_NOTEXIST;
	return NULL;
    }

    assert((datalen < 0xffffff));

    req->scerr = SCERR_SUCCESS;
    return handle_get_packet(SCREQ_TYPE_CHGET, datalen, 
			     accum_ref(data, datalen));
}

/* LVAL */
static struct sc_wblock_t *
handle_v_result(int op, int64_t val) {


    char *buffer;
    struct sc_wblock_t *wblock;

    wblock = wblock_alloc();
    wblock->length = 12;
    wblock->base = buffer = sc_alloc(wblock->length);

    //val = val & 0xffffffff;
    buffer = sc_htonl(buffer, SC_HDR_PACK_R(op, wblock->length));
    buffer = sc_htonll(buffer, val);
    return wblock;
}

static int64_t *
handle_v_getref(struct sc_reqhdr_t *req, char *key) {

    int slen;
    int64_t *stored;
    struct sc_session_t *lval;

    assert((key));

    lval = session_host_lval(req->session);
    if (!lval) {
	req->scerr = SCERR_NO_ACCESS;
	return NULL;
    }

    if (session_get(lval, key, (char **)&stored, &slen) < 0) {
	
	int64_t temp = 0;

	slen = sizeof(temp); 
	stored = (int64_t *)sc_memdup((char *)&temp, sizeof(temp));
	if (session_set(lval, key, (char *)stored, sizeof(temp)) < 0) {

	    LOG_IMPOSSIBILITY;
	    req->scerr = SCERR_SYSTEM; /* never */
	    return NULL;
	}
    }

    assert((slen == sizeof(int64_t)));
    return stored;
}

static int
handle_v_getarg(struct sc_reqhdr_t *req, int vlen, 
		char *val, int64_t *rv) {

    if (vlen != 8) {
	req->scerr = SCERR_PROTOCOL;
	return -1;
    }

    *rv = sc_ntohll(val);
    return 0;
}

static struct sc_wblock_t *
handle_vadd(struct sc_reqhdr_t *req, int vlen,
	    char *key, char *val) {

    int64_t arg;
    int64_t *stored;

    if ((handle_v_getarg(req, vlen, val, &arg) < 0) ||
	!(stored = handle_v_getref(req, key))) {

	/* no-access */
	return NULL;
    }

    *stored += arg;
    return handle_v_result(SCREQ_TYPE_VADD, (*stored));
}

static struct sc_wblock_t *
handle_vsub(struct sc_reqhdr_t *req, int vlen,
	    char *key, char *val) {

    int64_t arg;
    int64_t *stored;

    if ((handle_v_getarg(req, vlen, val, &arg) < 0) ||
	!(stored = handle_v_getref(req, key))) {

	/* no-access */
	return NULL;
    }

    *stored -= arg;
    return handle_v_result(SCREQ_TYPE_VSUB, (*stored));
}

static struct sc_wblock_t *
handle_vset(struct sc_reqhdr_t *req, int vlen,
	    char *key, char *val) {

    int64_t arg, prev;
    int64_t *stored;

    if ((handle_v_getarg(req, vlen, val, &arg) < 0) ||
	!(stored = handle_v_getref(req, key))) {

	/* no-access */
	return NULL;
    }

    prev = *stored;
    *stored = arg;
    return handle_v_result(SCREQ_TYPE_VSET, prev);
}

static struct sc_wblock_t *
handle_vget(struct sc_reqhdr_t *req, int vlen,
	    char *key, char *val) {

    int slen;
    int64_t *stored;
    struct sc_session_t *lval;

    lval = session_host_lval(req->session); 
    if (!lval) {
	req->scerr = SCERR_NO_ACCESS;
	return NULL;
    }

    if (!session_get(lval, key, (char **)&stored, &slen)) {

	assert((slen == sizeof(int64_t)));
	return handle_v_result(SCREQ_TYPE_VGET, (*stored));

    } else {

	return handle_v_result(SCREQ_TYPE_VGET, 0);
    }
}


static struct sc_wblock_t *
handle_vunset(struct sc_reqhdr_t *req, int vlen,
	      char *key, char *val) {

    int slen;
    int *stored;
    struct sc_session_t *lval;

    lval = session_host_lval(req->session);
    if (!lval) {
	req->scerr = SCERR_NO_ACCESS;
	return NULL;
    }

    if (!session_get(lval, key, (char **)&stored, &slen)) {

	assert((slen == sizeof(int64_t)));
	session_unset(lval, key);
	return handle_v_result(SCREQ_TYPE_VUNSET, (*stored));

    } else {

	/* never fails */
	req->scerr = SCERR_NOTEXIST;
	return NULL;
    }
}

static struct sc_wblock_t *
handle_vbrk(struct sc_reqhdr_t *req, int vlen,
	    char *key, char *val) {

    int slen;
    int64_t arg;
    int64_t *stored;
    struct sc_session_t *lval;

    assert((key));

    if (handle_v_getarg(req, vlen, val, &arg) < 0) {
	/* no-access */
	return NULL;
    }

    lval = session_host_lval(req->session);
    if (!lval) {
	req->scerr = SCERR_NO_ACCESS;
	return NULL;
    }

    if (session_get(lval, key, (char **)&stored, &slen) < 0) {
	req->scerr = SCERR_NOTEXIST;
	return NULL;
    }

    if (*stored != arg) {
	sc_vbreak = 1;
	req->scerr = SCERR_BREAK;
    } else {
	req->scerr = SCERR_SUCCESS;
    }

    return NULL;
}

/* ring, pushing data to */
static struct sc_wblock_t *
handle_rnadd_common(struct sc_reqhdr_t *req, int vlen,
		    char *key, char *val, int op) {

    /* symmetrical to handle_rnunshift */
    if (ring_add(req->session, key, 
		 accum_dup(val, vlen), vlen, op) < 0) {
	
	switch(errno) {
	case EACCES: req->scerr = SCERR_NO_ACCESS; break;
	case ENOSPC: req->scerr = SCERR_LIMITS_REACHED; break;
	}
    } else {
	req->scerr = SCERR_SUCCESS;
    }

    /* not error, rely alse success returning to 
       automatic error creation mechanism */
    return NULL;
}

static struct sc_wblock_t *
handle_rnpush(struct sc_reqhdr_t *req, int vlen,
	      char *key, char *val) {

    return handle_rnadd_common(req, vlen, key, val, 0);
}

static struct sc_wblock_t *
handle_rnunshift(struct sc_reqhdr_t *req, int vlen,
		 char *key, char *val) {

    return handle_rnadd_common(req, vlen, key, val, 1);
}

/* ring, acquiring data */
static struct sc_wblock_t *
handle_rnfetch_common(struct sc_reqhdr_t *req, int vlen,
		      char *key, char *val, int op, int from) {

    int datalen;
    char *data;
    struct sc_wblock_t *wblock;

    /* symmetrical to handle_rnshift */
    if (ring_fetch(req->session, key, &data, &datalen, from) < 0) {

	req->scerr = (errno == EACCES) ? 
	    SCERR_NO_ACCESS : SCERR_NOTEXIST;
	return NULL;
    }

    assert((datalen < 0xffffff));

    req->scerr = SCERR_SUCCESS;
    wblock = handle_get_packet(op, datalen, accum_ref(data, datalen));
    if (data) sc_unref(data); /* not to be left to ring, store.c
				 compensates on scache_rnget() */

    return wblock;
}

static struct sc_wblock_t *
handle_rnget(struct sc_reqhdr_t *req, int vlen,
	     char *key, char *val) {

    return handle_rnfetch_common(req, vlen, key, val, 
				 SCREQ_TYPE_RNGET, 0);
}

static struct sc_wblock_t *
handle_rnpop(struct sc_reqhdr_t *req, int vlen,
	     char *key, char *val) {

    return handle_rnfetch_common(req, vlen, key, val, 
				 SCREQ_TYPE_RNPOP, -1);
}

static struct sc_wblock_t *
handle_rnshift(struct sc_reqhdr_t *req, int vlen,
		 char *key, char *val) {

    return handle_rnfetch_common(req, vlen, key, val,
				 SCREQ_TYPE_RNSHIFT, 1);
}

/* ring, other operations */
static struct sc_wblock_t *
handle_rnclear(struct sc_reqhdr_t *req, int vlen,
	       char *key, char *val) {

    /* symmetrical to handle_rnshift */
    if (ring_purge(req->session, key) < 0) {
	req->scerr = (errno == EACCES) ? 
	    SCERR_NO_ACCESS : SCERR_NOTEXIST;
	return NULL;
    }
    
    req->scerr = SCERR_SUCCESS;
    return NULL;
}

static struct sc_wblock_t *
handle_rnsize(struct sc_reqhdr_t *req, int vlen,
	      char *key, char *val) {

    int size;

    /* symmetrical to handle_rnshift */
    if ((size = ring_size(req->session, key)) < 0) {
	req->scerr = (errno == EACCES) ? 
	    SCERR_NO_ACCESS : SCERR_NOTEXIST;
	return NULL;
    }

    return handle_v_result(SCREQ_TYPE_RNSIZE, size);
}

static struct sc_wblock_t *
handle_rnrotf(struct sc_reqhdr_t *req, int vlen,
	      char *key, char *val) {

    int64_t count;

    if (handle_v_getarg(req, vlen, val, &count)) return NULL;
    if (ring_rotate(req->session, key, count) < 0) {
	req->scerr = (errno == EACCES) ? 
	    SCERR_NO_ACCESS : SCERR_NOTEXIST;
    } else {
	req->scerr = SCERR_SUCCESS;
    }

    return NULL;
}
static struct sc_wblock_t *
handle_rnrotb(struct sc_reqhdr_t *req, int vlen,
	      char *key, char *val) {

    int64_t count;

    if (handle_v_getarg(req, vlen, val, &count)) return NULL;
    if (ring_rotate(req->session, key, -count) < 0) {
	req->scerr = (errno == EACCES) ? 
	    SCERR_NO_ACCESS : SCERR_NOTEXIST;
    } else {
	req->scerr = SCERR_SUCCESS;
    }

    return NULL;
}

/* session rename */
static struct sc_wblock_t *
handle_session_rename(struct sc_reqhdr_t *req, int vlen,
		      char *key, char *val) {


    /* size-verify */
    if (strlen(key) > SCACHE_PTO_MAX_TAGLEN) {
	req->scerr = SCERR_PROTOCOL;
    } else if (session_rename(req->session, key) < 0) {
	switch(errno) {
	case EEXIST:
	    req->scerr = SCERR_EXIST;
	    break;
	default:
	    req->scerr = SCERR_SYSTEM;
	    break;
	}
    } else {
	req->scerr = SCERR_SUCCESS;
    }

    return NULL;
}


static struct sc_wblock_t *
decode_2op(struct sc_reqhdr_t *rhdr, char *req, int length,
	   struct sc_wblock_t * (*hdlr)(struct sc_reqhdr_t *,
					int, char *, char *)) {
    
    int s_klen, s_vlen;
    char *s_key, *s_val;

    const int header_static_length = 8;

    /* requires session */
    if (!rhdr->session) {
	rhdr->scerr = SCERR_NO_SESSION;
	return NULL;
    }

    /* packet
     *
     * u16      : key-len
     * u16      : val-len (or 0 if not used)
     * char * key-length     : key
     * char * content-length : content (if used)
     */

    if (length < header_static_length) {
	rhdr->scerr = SCERR_PROTOCOL;
	return NULL;
    }

    /* parse */
    length -= header_static_length;
    s_klen = sc_ntohl(req); req += 4;
    s_vlen = sc_ntohl(req); req += 4;

    if (s_klen > (SCACHE_PTO_MAX_KEYLEN + 1)) {
	rhdr->scerr = SCERR_PROTOCOL;
	return NULL;
    }

    if (s_vlen == -1) { /* value from accumulator */

	if ((s_klen != length) || (s_klen < 1)) {
	    rhdr->scerr = SCERR_PROTOCOL;
	    return NULL;
	} else if (!sc_accumulator) {
	    rhdr->scerr = SCERR_INVALID_REQUEST;
	    return NULL;
	}

	s_key = req; req += s_klen;
	s_key[s_klen - 1] = 0; /* ensure null-termination */

	s_val = sc_accumulator;
	s_vlen = sc_accumulator_len;

    } else {

	if (s_vlen > (SCACHE_PTO_MAX_VALUELEN + 1)) {
	    rhdr->scerr = SCERR_PROTOCOL;
	    return NULL;
	} else if ((s_klen + s_vlen != length) || (s_klen < 1)) {
	    rhdr->scerr = SCERR_PROTOCOL;
	    return NULL;
	}
	
	s_key = req; req += s_klen;
	s_val = req; req += s_vlen;
	s_key[s_klen - 1] = 0; /* ensure null-termination */
    }

    /* exec */
    return hdlr(rhdr, s_vlen, s_key, s_val);
}


static struct sc_wblock_t *
handle_shdlrget(struct sc_reqhdr_t *rhdr, 
		struct sc_session_t *session, 
		int expires, char *key) {

    int datalen;
    char *data;

    if (session_get(session, key, &data, &datalen) < 0) {

	switch(errno) {
	case ENOENT:
	    rhdr->scerr = SCERR_NOTEXIST;
	    break;
	default:
	    rhdr->scerr = SCERR_SYSTEM;
	    break;
	}

	return NULL;
    }

    return handle_get_packet(SCREQ_TYPE_SESGET, datalen, data);
}

static struct sc_wblock_t *
handle_shdlrdrop(struct sc_reqhdr_t *rhdr, 
		 struct sc_session_t *session, char *key) {

    if (session_unset(session, key) < 0 && errno != ENOENT) {
	rhdr->scerr = SCERR_SYSTEM;
    } else {
	rhdr->scerr = SCERR_SUCCESS;
    }

    return NULL;
}

static struct sc_wblock_t *
handle_shdlrput(struct sc_reqhdr_t *rhdr, 
		struct sc_session_t *session,
		int vlen, char *key, char *val) {

    /* note: key might contain slashes, that make
       session_set to store data to subtree.

       but, should be no problem with that */
    if (session_set(session, key, val, vlen) < 0) {
	switch(errno) {
	case ENOSPC:
	    rhdr->scerr = SCERR_LIMITS_REACHED;
	    break;
	default:
	    rhdr->scerr = SCERR_SYSTEM;
	}
    } else {
	rhdr->scerr = SCERR_SUCCESS;
    }

    return NULL;
}

static struct sc_wblock_t *
decode_shdlrget(struct sc_reqhdr_t *rhdr, char *req, int length) {

    int s_klen, s_expires;
    char *s_key;
    struct sc_session_t *session;

    const int header_static_length = 8;

    /* dummy session */
    session = session_host_session(rhdr->auth, rhdr->peer);
    if (!session) {
	rhdr->scerr = SCERR_NO_SESSION;
	return NULL;
    }
    
    if (length < header_static_length) {
	rhdr->scerr = SCERR_INVALID_REQUEST;
	return NULL;
    }

    /* parse */
    length -= header_static_length;
    s_expires = sc_ntohl(req); req += 4;
    s_klen = sc_ntohl(req); req += 4;

    if ((s_klen != length) || (s_klen < 1)) {
	rhdr->scerr = SCERR_INVALID_REQUEST;
	return NULL;
    }

    
    s_key = req; req += s_klen;
    s_key[s_klen - 1] = 0; /* ensure null-termination */

    return handle_shdlrget(rhdr, session, s_expires, s_key);
}

static struct sc_wblock_t *
decode_shdlrdrop(struct sc_reqhdr_t *rhdr, char *req, int length) {

    int s_klen;
    char *s_key;
    struct sc_session_t *session;

    const int header_static_length = 4;

    /* dummy session */
    session = session_host_session(rhdr->auth, rhdr->peer);
    if (!session) {
	rhdr->scerr = SCERR_NO_SESSION;
	return NULL;
    }
    
    if (length < header_static_length) {
	rhdr->scerr = SCERR_INVALID_REQUEST;
	return NULL;
    }

    /* parse */
    length -= header_static_length;
    s_klen = sc_ntohl(req); req += 4;

    if ((s_klen != length) || (s_klen < 1)) {
	rhdr->scerr = SCERR_INVALID_REQUEST;
	return NULL;
    }

    
    s_key = req; req += s_klen;
    s_key[s_klen - 1] = 0; /* ensure null-termination */

    return handle_shdlrdrop(rhdr, session, s_key);
}

static struct sc_wblock_t *
decode_shdlrput(struct sc_reqhdr_t *rhdr, char *req, int length) {

    int s_expires; /* TODO: implementation missing */
    int s_klen, s_vlen;
    char *s_key, *s_val;
    struct sc_session_t *session;

    const int header_static_length = 12;

    /* dummy session */
    session = session_host_session(rhdr->auth, rhdr->peer);
    if (!session) {
	rhdr->scerr = SCERR_NO_SESSION;
	return NULL;
    }
    
    if (length < header_static_length) {
	rhdr->scerr = SCERR_INVALID_REQUEST;
	return NULL;
    }

    /* parse */
    length -= header_static_length;
    s_expires = sc_ntohl(req); req += 4;
    s_klen = sc_ntohl(req); req += 4;
    s_vlen = sc_ntohl(req); req += 4;

    if ((s_klen + s_vlen != length) || 
	(s_klen < 1) || (s_vlen < 1)) {

	rhdr->scerr = SCERR_INVALID_REQUEST;
	return NULL;
    }

    s_key = req; req += s_klen;
    s_val = req; req += s_vlen;
    s_key[s_klen - 1] = 0; /* ensure null-termination */

    return handle_shdlrput(rhdr, session, s_vlen, s_key, s_val);
}




/* ===================================
 * protocol / middle layer
 * =================================== */
static struct sc_wblock_t *
create_error(int req_type, struct sc_reqhdr_t *rhdr) {

    char *buffer;
    struct sc_wblock_t *wblock;

    wblock = wblock_alloc();
    wblock->length = 8;
    wblock->base = buffer = sc_alloc(wblock->length);
    
    buffer = sc_htonl(buffer, 
		      SC_HDR_PACK_R((req_type | 0x80), wblock->length));
    buffer = sc_htonl(buffer, rhdr->scerr);
    return wblock;
}

static struct sc_wblock_t *
proto_switch(struct sc_reqhdr_t *rhdr, char *req, 
	     int req_type, int req_len) {

    struct sc_wblock_t *wblock = NULL;

    sc_stats->io_requests++;
    rhdr->scerr = SCERR_SUCCESS; /* default */

    /* is breaked */
    if (sc_vbreak) {

	switch(req_type) {
	case SCREQ_TYPE_SETOPT:

	case SCREQ_TYPE_ADD:
	case SCREQ_TYPE_SET:
	case SCREQ_TYPE_REPLACE:
	case SCREQ_TYPE_UNSET:
	case SCREQ_TYPE_COLL:

	case SCREQ_TYPE_SHADD:
	case SCREQ_TYPE_SHSET:
	case SCREQ_TYPE_SHREPLACE:
	case SCREQ_TYPE_SHUNSET:
	case SCREQ_TYPE_SHCOLL:

	case SCREQ_TYPE_CHPUT:
	case SCREQ_TYPE_CHCLEAR:

	case SCREQ_TYPE_VADD:
	case SCREQ_TYPE_VSUB:
	case SCREQ_TYPE_VSET:
	case SCREQ_TYPE_VUNSET:

	case SCREQ_TYPE_RNCLEAR:
	case SCREQ_TYPE_RNPUSH:
	case SCREQ_TYPE_RNPOP:
	case SCREQ_TYPE_RNSHIFT:
	case SCREQ_TYPE_RNUNSHIFT:
	case SCREQ_TYPE_RNROTF:
	case SCREQ_TYPE_RNROTB:

	    sc_scerr = rhdr->scerr = SCERR_BREAK;
	    return create_error(req_type, rhdr);
	}
    }

    switch(req_type) {
    case SCREQ_TYPE_GETOPT:
	STAT_HARD(sc_stats->req_t_getopts++);
	wblock = decode_getopt(rhdr, req, req_len);
	break;
    case SCREQ_TYPE_SETOPT:
	STAT_HARD(sc_stats->req_t_setopts++);
	wblock = decode_setopt(rhdr, req, req_len);
	break;
	
    case SCREQ_TYPE_SRVSTATS:
	STAT_HARD(sc_stats->req_t_srvstats++);
	wblock = decode_srvstats(rhdr, req, req_len);
	break;


	/* data-operations */
    case SCREQ_TYPE_ADD:
	STAT_HARD(sc_stats->req_t_adds++);
	wblock = decode_2op(rhdr, req, req_len, handle_add);
	break;
    case SCREQ_TYPE_SET:
	STAT_HARD(sc_stats->req_t_sets++);
	wblock = decode_2op(rhdr, req, req_len, handle_set);
	break;
    case SCREQ_TYPE_REPLACE:
	STAT_HARD(sc_stats->req_t_replaces++);
	wblock = decode_2op(rhdr, req, req_len, handle_replace);
	break;
    case SCREQ_TYPE_UNSET:
	STAT_HARD(sc_stats->req_t_unsets++);
	wblock = decode_2op(rhdr, req, req_len, handle_unset);
	break;
    case SCREQ_TYPE_STAT:
	STAT_HARD(sc_stats->req_t_stats++);
	wblock = decode_2op(rhdr, req, req_len, handle_stat);
	break;
    case SCREQ_TYPE_GET:
	STAT_HARD(sc_stats->req_t_gets++);
	wblock = decode_2op(rhdr, req, req_len, handle_get);
	break;
    case SCREQ_TYPE_COLL:
	STAT_HARD(sc_stats->req_t_colls++);
	wblock = decode_2op(rhdr, req, req_len, handle_coll);
	break;

	/* session */
    case SCREQ_TYPE_SESGET:
	STAT_HARD(sc_stats->req_t_shdlrgets++);
	wblock = decode_shdlrget(rhdr, req, req_len);
	break;
    case SCREQ_TYPE_SESPUT:
	STAT_HARD(sc_stats->req_t_shdlrputs++);
	wblock = decode_shdlrput(rhdr, req, req_len);
	break;
    case SCREQ_TYPE_SESDROP:
	STAT_HARD(sc_stats->req_t_shdlrdrops++);
	wblock = decode_shdlrdrop(rhdr, req, req_len);
	break;

	/* shared */
    case SCREQ_TYPE_SHADD:
	STAT_HARD(sc_stats->req_t_shadds++);
	wblock = decode_2op(rhdr, req, req_len, handle_shadd);
	break;
    case SCREQ_TYPE_SHSET:
	STAT_HARD(sc_stats->req_t_shsets++);
	wblock = decode_2op(rhdr, req, req_len, handle_shset);
	break;
    case SCREQ_TYPE_SHREPLACE:
	STAT_HARD(sc_stats->req_t_shreplaces++);
	wblock = decode_2op(rhdr, req, req_len, handle_shreplace);
	break;
    case SCREQ_TYPE_SHUNSET:
	STAT_HARD(sc_stats->req_t_shunsets++);
	wblock = decode_2op(rhdr, req, req_len, handle_shunset);
	break;
    case SCREQ_TYPE_SHSTAT:
	STAT_HARD(sc_stats->req_t_shstats++);
	wblock = decode_2op(rhdr, req, req_len, handle_shstat);
	break;
    case SCREQ_TYPE_SHGET:
	STAT_HARD(sc_stats->req_t_shgets++);
	wblock = decode_2op(rhdr, req, req_len, handle_shget);
	break;
    case SCREQ_TYPE_SHCOLL:
	STAT_HARD(sc_stats->req_t_shcolls++);
	wblock = decode_2op(rhdr, req, req_len, handle_shcoll);
	break;

	/* cache */
    case SCREQ_TYPE_CHGET:
	STAT_HARD(sc_stats->req_t_chgets++);
	wblock = decode_2op(rhdr, req, req_len, handle_chget);
	break;
    case SCREQ_TYPE_CHPUT:
	STAT_HARD(sc_stats->req_t_chputs++);
	wblock = decode_2op(rhdr, req, req_len, handle_chput);
	break;
    case SCREQ_TYPE_CHCLEAR:
	STAT_HARD(sc_stats->req_t_chclears++);
	wblock = decode_2op(rhdr, req, req_len, handle_chclear);
	break;

	/* lval */
    case SCREQ_TYPE_VADD:
	STAT_HARD(sc_stats->req_t_vadds++);
	wblock = decode_2op(rhdr, req, req_len, handle_vadd);
	break;
    case SCREQ_TYPE_VSUB:
	STAT_HARD(sc_stats->req_t_vsubs++);
	wblock = decode_2op(rhdr, req, req_len, handle_vsub);
	break;
    case SCREQ_TYPE_VGET:
	STAT_HARD(sc_stats->req_t_vgets++);
	wblock = decode_2op(rhdr, req, req_len, handle_vget);
	break;
    case SCREQ_TYPE_VSET:
	STAT_HARD(sc_stats->req_t_vsets++);
	wblock = decode_2op(rhdr, req, req_len, handle_vset);
	break;
    case SCREQ_TYPE_VUNSET:
	STAT_HARD(sc_stats->req_t_vunsets++);
	wblock = decode_2op(rhdr, req, req_len, handle_vunset);
	break;
    case SCREQ_TYPE_VBRK:
	STAT_HARD(sc_stats->req_t_vbrks++);
	wblock = decode_2op(rhdr, req, req_len, handle_vbrk);
	break;

	/* ring */
    case SCREQ_TYPE_RNGET:
	STAT_HARD(sc_stats->req_t_rngets++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnget);
	break;
    case SCREQ_TYPE_RNSIZE:
	STAT_HARD(sc_stats->req_t_rnsizes++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnsize);
	break;
    case SCREQ_TYPE_RNCLEAR:
	STAT_HARD(sc_stats->req_t_rnclears++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnclear);
	break;
    case SCREQ_TYPE_RNPUSH:
	STAT_HARD(sc_stats->req_t_rnpushes++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnpush);
	break;
    case SCREQ_TYPE_RNPOP:
	STAT_HARD(sc_stats->req_t_rnpops++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnpop);
	break;
    case SCREQ_TYPE_RNSHIFT:
	STAT_HARD(sc_stats->req_t_rnshifts++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnshift);
	break;
    case SCREQ_TYPE_RNUNSHIFT:
	STAT_HARD(sc_stats->req_t_rnunshifts++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnunshift);
	break;
    case SCREQ_TYPE_RNROTF:
	STAT_HARD(sc_stats->req_t_rnrotfs++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnrotf);
	break;
    case SCREQ_TYPE_RNROTB:
	STAT_HARD(sc_stats->req_t_rnrotbs++);
	wblock = decode_2op(rhdr, req, req_len, handle_rnrotb);
	break;

	/* session-operations */
    case SCREQ_TYPE_SESSION_CREATE:
	STAT_HARD(sc_stats->req_t_creates++);
	wblock = decode_session_create(rhdr, req, req_len);
	break;

    case SCREQ_TYPE_SESSION_DROP:
	STAT_HARD(sc_stats->req_t_drops++);
	wblock = decode_session_drop(rhdr, req, req_len);
	break;

    case SCREQ_TYPE_SESSION_RENAME:
	STAT_HARD(sc_stats->req_t_renames++);
	wblock = decode_2op(rhdr, req, req_len, handle_session_rename);
	break;

    case SCREQ_TYPE_NONE:
	if (!rhdr->session) {
	    rhdr->scerr = SCERR_NO_SESSION;
	} else {
	    rhdr->scerr = SCERR_SUCCESS;
	}
	wblock = NULL;
	break;

    case SCREQ_TYPE_LASTERR:
	STAT_HARD(sc_stats->req_t_lasterrs++);
	rhdr->scerr = sc_scerr;
	wblock = handle_v_result(SCREQ_TYPE_LASTERR, rhdr->scerr);
	break;

    default:
	STAT_HARD(sc_stats->req_t_unknowns++);
	rhdr->scerr = SCERR_UNKNOWN_REQUEST;
	wblock = NULL;
	break;
    }

    /* store last error */
    sc_scerr = rhdr->scerr;

    if (!wblock) {
	wblock = create_error(req_type, rhdr);
    }

    return wblock;
}


static void
proto_update_statistics(struct sc_reqhdr_t *rhdr) {

    switch(rhdr->scerr) {
    case SCERR_SUCCESS:
	sc_stats->req_count_success++;
	break;
    case SCERR_INVALID_REQUEST:
	sc_stats->req_count_invalid++;
	break;
    case SCERR_UNKNOWN_REQUEST:
	sc_stats->req_count_unknown++;
	break;
    default:
	break;
    }
}

struct sc_wblock_t *
create_response_header(struct sc_reqhdr_t *rhdr, int numresp,
		       struct sc_wblock_t *queue) {

    int tlen = 0;
    char *buffer;
    struct sc_wblock_t *wblock;

    assert((numresp < 256));

    /* sum length */
    for(wblock = queue; wblock; wblock = wblock->next) {
	tlen += wblock->length;
	wblock->pos = 0;
    }

    wblock = wblock_alloc();
    wblock->length = 12; /* proto-headerlen */
    wblock->base = buffer = sc_alloc(wblock->length);
    wblock->next = queue;

    /* pack proto header */
    buffer = sc_htonl(buffer, 
		      SC_HDR_PACK_H(numresp, 
				    tlen + wblock->length));
    buffer = sc_htonl(buffer, rhdr->session ? rhdr->session->idx : -1);
    buffer = sc_htonl(buffer, rhdr->flags);
    return wblock;
}

static char *
proto_parse_header(struct sc_reqhdr_t *rhdr, char *req, int *length) {

    /* header */
    if (*length < 14) return NULL;
    rhdr->index = sc_ntohl(req); req += 4;
    rhdr->flags = sc_ntohs(req); req += 4;
    rhdr->tag_len = sc_ntohs(req); req += 2;
    rhdr->sec_len = sc_ntohs(req); req += 2;
    rhdr->ath_len = sc_ntohs(req); req += 2;
    *length -= 14;

    rhdr->flags = rhdr->flags & SCACHE_QFLAG_MASK;

    /* tag + verify */
    if ((rhdr->tag_len < SCACHE_PTO_MIN_TAGLEN) ||
	(rhdr->tag_len > SCACHE_PTO_MAX_TAGLEN) ||
	(rhdr->tag_len < 1) ||
	(*length < rhdr->tag_len)) return NULL;

    rhdr->tag = req; 
    rhdr->tag[rhdr->tag_len - 1] = 0;
    req += rhdr->tag_len;
    *length -= rhdr->tag_len;
    
    /* secret + verify */
    if ((rhdr->sec_len < SCACHE_PTO_MIN_SECRETLEN) ||
	(rhdr->sec_len > SCACHE_PTO_MAX_SECRETLEN) ||
	(rhdr->sec_len < 1) ||
	(*length < rhdr->sec_len)) return NULL;

    rhdr->secret = req;
    rhdr->secret[rhdr->sec_len - 1] = 0;
    req += rhdr->sec_len;
    *length -= rhdr->sec_len;

    /* auth */
    if ((rhdr->ath_len < SCACHE_PTO_MIN_AUTHLEN) ||
	(rhdr->ath_len > SCACHE_PTO_MAX_AUTHLEN) ||
	(rhdr->ath_len < 1) ||
	(*length < rhdr->ath_len)) return NULL;


    rhdr->auth = req;
    rhdr->auth[rhdr->ath_len - 1] = 0;
    req += rhdr->ath_len;
    *length -= rhdr->ath_len;

    /* fetch session */
    rhdr->session = 
	session_open(rhdr->tag, rhdr->secret, 
		     rhdr->auth, rhdr->peer,
		     rhdr->index);

    return req;
}


struct sc_wblock_t *
proto_request(const char *peer_id, int num_requests,
	      char *req, int length) {

    int i;
    int req_len;

    enum scache_reqtype_t req_type;

    struct sc_wblock_t *wblock = NULL;
    struct sc_wblock_t **wblockp = &wblock;
    struct sc_reqhdr_t *rhdr = alloca(sizeof(struct sc_reqhdr_t));

    /* initially non-breaked */
    sc_vbreak = 0;
    sc_scerr = SCERR_SUCCESS;

    /* accumulator-null */
    sc_accumulator = NULL;
    sc_accumulator_len = 0;

    rhdr->peer = peer_id;
    if ((req = proto_parse_header(rhdr, req, &length))) {
	
	for(i = 0; i < num_requests; i++) {

	    req_len = req_type = sc_ntohl(req); /* SC_PACK_HDR */
	    req_len = req_len & 0xffffff; /* maximum request length */
	    req_type = (req_type >> 24) & 0xff;
	    req += 4; length -= 4; req_len -= 4; 

	    if (length >= req_len) {

		/* prosess part */
		*wblockp = proto_switch(rhdr, req, req_type, req_len);
		length -= req_len;
		req += req_len;

		while(*wblockp) wblockp = &((*wblockp)->next);

	    } else { 

		/* length mismatch, die */
		rhdr->scerr = SCERR_PROTOCOL;
		break;
	    }
	}

	if (i != num_requests || rhdr->scerr == SCERR_PROTOCOL) {

	    rhdr->scerr = SCERR_PROTOCOL;
	    while(wblock) {

		struct sc_wblock_t *next = wblock->next;
		assert((wblock->base));
		sc_unref(wblock->base);
		sc_unref(wblock);
		wblock = next;
	    }
	}
    }

    /* leave point for all protocol */
    proto_update_statistics(rhdr);

    /* free accumulator */
    accum_dup(NULL, 0);

    if (wblock) {
	return create_response_header(rhdr, num_requests, wblock);
    } else {
	return NULL; /* will kill socket */
    }
}


int
proto_init(enum sc_bootstages_t stage) {

    return 0;
}


/*
  Local variables: ***
  mode: c ***
  c-basic-offset: 4 ***
  End: ***
*/
