/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */

/*
  TODO: $_SESSION-test
*/

#include "scache.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <syslog.h>
#include <signal.h>


#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>

/* local variables */
static enum {
    SC_CONTROL_NONE,
    SC_CONTROL_RESTART,
    SC_CONTROL_STOP
} main_control = SC_CONTROL_NONE;


static pid_t child;
static int   restart_child;
static int   dump_compiled_config = 0;

static int
main_init_syslog() {

    if (sc_config->debug) {
	openlog(sc_config->syslog_tag, LOG_PERROR, 
		sc_config->syslog_fac);
    } else {
	openlog(sc_config->syslog_tag, 0, 
		sc_config->syslog_fac);
    }

    return 0;
}

static int
main_priv_set(uid_t uid, gid_t gid, int effective_only) {
    
    if (effective_only) {

	if (setegid(gid) < 0) {
	    fprintf(stderr, "setegid(%i) failed: %s",
		    gid, strerror(errno));
	    return -1;
	}
	if (seteuid(uid) < 0) {
	    fprintf(stderr, "seteuid(%i) failed: %s",
		    uid, strerror(errno));
	    return -1;
	}

    } else { /* persistent drop */

	if (setgid(gid) < 0) {
	    fprintf(stderr, "setgid(%i) failed: %s",
		    gid, strerror(errno));
	    return -1;
	}
	if (setuid(uid) < 0) {
	    fprintf(stderr, "setuid(%i) failed: %s",
		    uid, strerror(errno));
	    return -1;
	}
    }

    return 0;
}

static int
main_open_pidfile(int openonly) {

    int rv;

    if (openonly) {

	rv = open(sc_config->pidfile_path, O_RDONLY);
	if (rv < 0) {
	    fprintf(stderr, "opening pidfile '%s' failed: %s\n",
		    sc_config->pidfile_path, strerror(errno));
	    return -1;
	} else {
	    return rv;
	}
    }

#ifdef NDEBUG
    if (sc_config->pidfile_path[0] != '/') {
	fprintf(stderr, "pidfile path needs to be absolute path. "
		"current path is '%s'\n", sc_config->pidfile_path);
	return -1;
    }
#endif

    /* else create */
    rv = unlink(sc_config->pidfile_path);
    if (rv < 0 && errno != ENOENT) {
	fprintf(stderr, "removing old pidfile '%s' failed: %s\n",
		sc_config->pidfile_path, strerror(errno));
	return -1;
    }

    rv = open(sc_config->pidfile_path, 
	      O_WRONLY | O_EXCL | O_CREAT, 
	      S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
    if (rv < 0) {
	fprintf(stderr, "creating pidfile '%s' failed: %s\n",
		sc_config->pidfile_path, strerror(errno));
	return -1;
    }

    return rv;
}

static int
main_daemonize() {

    int pidfd;
    char pidbuf[16];

    /* while we have stderr */
    if ((pidfd = main_open_pidfile(0)) < 0) return -1;

    /* pidfile opened, drop privileges permanently */
    if (!geteuid() || !getuid()) {
	if (main_priv_set(sc_config->runas_uid,
			  sc_config->runas_gid, 1) < 0) {
	    
	    fprintf(stderr, "dropping privileges failed. exiting\n");
	    exit(-1);
	}
    }

    /* prefork */
    switch(fork()) {
    case -1:
	fprintf(stderr, "first fork() failed: %s\n",
		strerror(errno));
	return -1;
    case 0: 
	break; /* we are the annoying child */
    default: 
	exit(0); /* we are the proud parent */
    }

    /* become to session leader */
    if (setsid() < 0) {
	fprintf(stderr, "setsid() failed: %s\n",
		strerror(errno));
	return -1;
    }

    /* refork */
    switch(fork()) {
    case -1:
	fprintf(stderr, "second fork() failed: %s\n",
		strerror(errno));
	return -1;
    case 0: 
	break; /* another annoying child */
    default: 
	exit(0); /* another hopeful parent */
    }
    
    chdir("/");
    umask(0);
    close(0);
    close(1);
    close(2);

    if (!freopen("/dev/null", "r", stdin) ||
	!freopen("/dev/null", "a", stdout) ||
	!freopen("/dev/null", "a", stderr)) {

	/* no use to fprintf any more... */
	syslog(LOG_ERR, "fdreopen() failed: %s",
	       strerror(errno));
	return -1;
    }

    /* store pid and return */
    sprintf(pidbuf, "%u\n", getpid());
    if (write(pidfd, pidbuf, strlen(pidbuf)) != strlen(pidbuf)) {
	syslog(LOG_ERR, "writing pid %u to pidfile failed",
	       getpid());
	return -1;
    }

    close(pidfd);
    return 0;
}

static void
main_signal(int signum) {
    
    restart_child = (signum == SIGHUP ? 1 : 0);
    kill(child, SIGTERM);
}

static int
main_watcher() { /* restart childs if they exit */

    int breakout = 100; /* max child-dies */

    signal(SIGHUP, main_signal);
    signal(SIGTERM, main_signal);


 child:

    /* fork child and return */
    switch((child = fork())) {
    case -1:
	syslog(LOG_ERR, "third fork() failed: %s",
	       strerror(errno));
	goto cleanup;
    case 0: /* oppressive child */
	return 0;
    default: /* middle-aged self */
	restart_child = 0;
    }

    /* we are parent (watcher), child returned eventually 
       to spin_sock() */

    syslog(LOG_NOTICE, "parent: started child %u", child);
    do {
	
	int status;
	pid_t pid;

	pid = waitpid(child, &status, 0);
	if (pid < 0 && errno == EINTR) {
	    /* got signal, ok */
	} else if (pid < 0) {
	    syslog(LOG_ERR, "parent: waitpid() failed: %s",
		   strerror(errno));
	    continue;
	} else if (pid == child) {

	    if (status == 0) {

		if (restart_child) {
		    syslog(LOG_NOTICE,
			   "parent: ALERT: child %u exited with code %u. "
			   "Restarting child.", child, status);
		    goto child;
		}

		/* else */
		breakout = 0;
		syslog(LOG_NOTICE, 
		       "parent: child %u exited with code %u. "
		       "Fine, I'll exit too.", 
		       child, status);
		
		break;

	    } else if (status == SIGSEGV) {

		syslog(LOG_ALERT,
		       "parent: child %i exited on SIGSEGV. "
		       "This suggests either defective hardware or "
		       "programming error on scached's code. If you "
		       "can replicate error, please inform developer.",
		       child);
		syslog(LOG_ALERT,
		       "parent: restarting child, sessions are gone");
		goto child;

	    } else {

		syslog(LOG_NOTICE,
		       "parent: child %u exited with error code %u. "
		       "Won't continue, exiting", child, status);

		/* continue; TODO.. why break??? Explain myself */
		break;
	    }

	} else {
	    syslog(LOG_ERR, "parent: unknown child %u exited with "
		   "code %u. Mine well-grown is %u", pid, status, child);
	    break;
	}

    } while(breakout-- > 0);


 cleanup:

    /* remove pidfile and socket */
    unlink(sc_config->pidfile_path);
    unlink(sc_config->usock_path);

    if (breakout < 0) {
	
	/* bailed out */
	syslog(LOG_ERR, "parent: breaked out through errors, "
	       "terminating child %u.", child);
	kill(child, SIGTERM);
	exit(-1);
    }

    exit(0);
}

static int
main_run_inits(enum sc_bootstages_t stage) {

    extern int config_init(enum sc_bootstages_t);
    extern int common_init(enum sc_bootstages_t);
    extern int sock_init(enum sc_bootstages_t);
    extern int proto_init(enum sc_bootstages_t);
    extern int store_init(enum sc_bootstages_t);

    if (common_init(stage) < 0 ||
	config_init(stage) < 0 ||
	sock_init(stage) < 0 ||
	proto_init(stage) < 0 ||
	store_init(stage) < 0) return -1;
	
    return 0;
}

static void
main_help() {

    fprintf(stdout,
	    "Session caching daemon\n"
	    "\n"
	    "    scached [options]\n"
	    "\n"
	    "Options:\n"
	    "\n"
	    "   --restart        signal restart to running daemon\n"
	    "   --hup\n"
	    "   --stop           signal exit to running daemon\n"
	    "   --quit\n"
	    "   --kill\n"
	    "   --config         path to configuration file\n"
	    "    -C\n"
	    "   --dump-config    \n"
	    "   --version        \n"
	    "   --help           this message\n"
	    "\n");


    config_dump_default_config();
    exit(0);

}

static void
main_dump_version() {

    fprintf(stdout,
	    "scached - version %s\n"
	    "\n"
	    "License GPLv3+: GNU GPL version 3 or later"
	    " <http://gnu.org/licenses/gpl.html>.\n"
	    "This is free software: you are free to change"
	    " and redistribute it.\n"
	    "There is NO WARRANTY, to the extent permitted by law.\n"
	    "\n"
	    "<http://scache.nanona.fi/>\n\n",
	    VERSION);

    exit(0);
}

static int
main_preparse_args(int argc, char **argv) {

    int i;

    ((struct sc_config_t *)sc_config)->program_argc = argc;
    ((struct sc_config_t *)sc_config)->program_argv = argv;

    for(i = 1; i < argc; i++) {

	if (!strcmp(argv[i], "--restart") ||
		   !strcmp(argv[i], "--hup")) {

	    main_control = SC_CONTROL_RESTART; /* restart if daemon */

	} else if (!strcmp(argv[i], "--stop") ||
		   !strcmp(argv[i], "--quit") ||
		   !strcmp(argv[i], "--kill")) {

	    main_control = SC_CONTROL_STOP; /* stop if daemon */

	} else if (!strcmp(argv[i], "--help") ||
		   !strcmp(argv[i], "-h")) {

	    /* help */
	    main_help();

	} else if (!strcmp(argv[i], "--dump-config")) {

	    /* config */
	    dump_compiled_config = 1;

	} else if (!strcmp(argv[i], "--version") ||
		   !strcmp(argv[i], "-v")) {

	    /* version + copyright */
	    main_dump_version();

	} else if (!strcmp(argv[i], "--")) {
	    break;
	}
    }

    return 0;
}

static int
main_do_control() {

    int fd;
    pid_t pid;
    char *cp, buffer[16];

    fd = main_open_pidfile(1);
    if (fd < 0) return -1;

    memset(buffer, 0, sizeof(buffer));
    if (read(fd, buffer, sizeof(buffer)) < 0) {
	fprintf(stderr, "reading pid from pidfile failed: %s\n",
		strerror(errno));
	return -1;
    }

    pid = strtol(buffer, &cp, 10);
    if (!pid || (*cp && !isspace((int)*cp))) {
	fprintf(stderr, "invalid pid '%s' in pidfile.\n", buffer);
	return -1;
    }

    if (main_control == SC_CONTROL_RESTART) {
	kill(pid, SIGHUP);
    } else if (main_control == SC_CONTROL_STOP) {
	kill(pid, SIGTERM);
    }

    return 0;
}

static void
main_child_exit(int signum) {

    exit(0);
}

int main (int argc, char **argv) {

    /* extract control commands */
    if (main_preparse_args(argc, argv) < 0) return -1;

    /* try to drop to generic privileges while parsing 
       and initializing */
    if (!geteuid() || !getuid()) {
	if (main_priv_set(sc_config->runas_uid,
			  sc_config->runas_gid, 1) < 0) {
	    
	    fprintf(stderr, "dropping privileges during initialization "
		    "failed. continuing shamelessy.\n");
	}
    }

    if (main_init_syslog() < 0) return -1;
    
    /* prepare internals and collect and verify configuratinon */
    if (main_run_inits(SC_STAGE_PREPARE) < 0 ||
	main_run_inits(SC_STAGE_CONFIG) < 0) {

	exit(-1);
    }

    /* dump compiled config */
    if (dump_compiled_config) {
	config_dump_compiled_config();
	exit(0);
    }
    
    /* command to existing process, after we have resolved
       correct pidfiles */
    if (main_control != SC_CONTROL_NONE) {
	exit(main_do_control());
    }

    /* reopen syslog with real values */
    if (main_init_syslog() < 0) return -1;
    
    /* raise privileges during start */
    if (!geteuid() || !getuid()) {
	if (main_priv_set(0, 0, 1) < 0) {
	    
	    fprintf(stderr, "raising privileges during initialization "
		    "failed. continuing, but little skeptical for "
		    "success.\n");
	}
    }

    /* do privileged internals and finalize */
    if (main_run_inits(SC_STAGE_START) < 0) {
	exit(-1);
    }
    

    if (!sc_config->nodetach) {

	/* need to keep privileges until pidfile opened */
	if (main_daemonize() < 0 ||
	    main_watcher() < 0) return -1;

    } else {

	/* drop privileges permanently */
	if (!geteuid() || !getuid()) {
	    if (main_priv_set(sc_config->runas_uid,
			      sc_config->runas_gid, 1) < 0) {
		
		fprintf(stderr, "dropping privileges failed. exiting\n");
		exit(-1);
	    }
	}
    }


#ifndef NDEBUG
    /* watcher left behind, we are the operating child */
#endif

    /* exit-signal */
    signal(SIGHUP, main_child_exit);
    signal(SIGINT, main_child_exit);
    signal(SIGTERM, main_child_exit);

    /* pipe-ignore */
    signal(SIGPIPE, SIG_IGN);

    /* THE main */
    sock_spin();

    return 0;
}

/*
  Local variables: ***
  mode: c ***
  c-basic-offset: 4 ***
  End: ***
*/

