/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */

#ifndef _SCACHE_H_
#define _SCACHE_H_

#ifdef HAVE_CONFIG_H
# define _CONFIG_H_
# include "config.h"
#endif

#ifndef ENABLE_DEVEL
# define NDEBUG 1 /* before assert.h */
#endif

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>

#ifdef HAVE_STDINT_H
# include <stdint.h>
#endif
#ifdef HAVE_INTTYPES_H
# include <inttypes.h>
#endif

#ifdef HAVE_STRING_H
# include <string.h>
#endif
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif

#ifdef TIME_WITH_SYS_TIME
# include <sys/time.h>
# include <time.h>
#else
# ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

#ifdef HAVE_ERROR_H
# include <error.h>
#endif
#ifdef HAVE_ERRNO_H
# include <errno.h>
#endif

#include "ext/php-scache-proto.h"
#include "ext/php-scache-paths.h"

//#define ENABLE_EXTENDED_STATS 1
/* #undef ENABLE_EXTENDED_STATS */

/* +++++++++++++++++++++++++++++++++++++++++++++++++
 * featuring depending config.h
 * +++++++++++++++++++++++++++++++++++++++++++++++++ */

/* enough functionality to enable ipv6 */
#if defined(HAVE_GETADDRINFO) && \
    defined(HAVE_FREEADDRINFO) && \
    defined(HAVE_NETINET_IP6_H) && \
    defined(HAVE_SOCKADDR_STORAGE)
# define ENABLE_IPV6 1
#endif


/* polling method depending on config */
#if defined(HAVE_EPOLL_CREATE) && \
    defined(HAVE_SYS_EPOLL_H) 
# define ENABLE_EPOLL 1 /* can epoll */
#elif defined(HAVE_POLL) && \
      defined(HAVE_POLL_H)
# define ENABLE_POLL 1  /* can poll */
#elif !defined(ENABLE_POLL) && !defined(ENABLE_EPOLL)
# include <sys/select.h>
# define ENABLE_SELECT 1 /* fallback to select */
#endif

#ifdef ENABLE_EXTENDED_STATS
# define STAT_ME_HARDER                1
# define STAT_HARD(x) (x)
#else
# define STAT_HARD(x)
#endif

#ifndef DISABLE_NODES_ACCOUNTING
# define SESSION_NODES_ACCOUNTING      1
#endif

#ifndef DISABLE_MEMORY_ACCOUNTING
# define SESSION_MEMORY_ACCOUNTING     1
#endif

#ifdef ENABLE_FULL_PARTITIONING
# define SEPARATE_SESSIONTREES         1
#endif

#ifndef NDEBUG
# define M() fprintf(stderr, "MARK: %s:%u\n", __FUNCTION__, __LINE__)
#else
# define M() ()
#endif

#define LOG_IMPOSSIBILITY \
    syslog(LOG_ERR, \
           "IMPOSSIBLE RESULT, we really shouldn't be here. " \
	   "File: %s, line: %i, Function: %s\n", \
	    __FILE__, __LINE__, __FUNCTION__);

#define sc_asize(x)  (sizeof(x) / sizeof(x[0]))

#define _NMTOBO                    64    /* need-more-than-off-by-one */


/* logging levels */
#define SC_LOG_NOTHING             0
#define SC_LOG_FATAL               1
#define SC_LOG_ERROR               2
#define SC_LOG_NOTICE              3
#define SC_LOG_DEBUG               4

extern time_t   sc_lazytime;

struct sc_tree_t;


/* bootstrap */
enum sc_bootstages_t {
    
    SC_STAGE_CONFIG,
    SC_STAGE_PREPARE,
    SC_STAGE_START
};

/* generic linked-list node */
struct sc_list_t {
    void *data;
    struct sc_list_t *next;
};


/* writeout-node */
struct sc_wblock_t {
    void *base;
    int pos;
    int length;
    struct sc_wblock_t *next;
};


/* refcounted memory-allocation element. when allocated
   by sc_alloc, sc_strdup or sc_memdup, refcount is 
   set to 1. 
   
   freeing must be through sc_unref, that actually frees
   element when refcount drops to 0. don't sc_free 
   directly, sc_free asserts if refcount is not 0

   small element are aquired from obstack and
   queued on free, free():d, large elements (queue == NULL) are
   generic malloc(), free() operations */

struct sc_memb_t {
    int refcount;

#ifdef STAT_ME_HARDER
    struct sc_memqueue_t *queue;
#else
    struct sc_list_t **queue;
#endif

#ifndef NDEBUG
    /* 0xb00f when out, 0xbeef when in */
    void *b00f;
#endif
};


/* session data / branch node for storing data
 *
 * crearing data does not remove node from it's tree
 * just clears content and marks it empty to avoid tree balance.
 * missing entry and SC_NODE_EMPY should be so considered
 * equal */

#define SC_NODE_EMPTY          0x0
#define SC_NODE_VALUE          0x1
#define SC_NODE_SUBTREE        0x2

struct sc_node_t {
    int flags;    /* actually one of SC_NODE_s */
#ifdef SESSION_MEMORY_ACCOUNTING
    int keylen;   /* for avoiding strlen():s */
#endif
    union {
	struct { /* SC_NODE_VALUE */
	    int data_len;
	    char *data;
	} data;
	struct sc_tree_t *tree; /* SC_NODE_SUBTREE */
    } u;
};

/* ring */
struct sc_ring_t {

    int64_t cur_nodes;
    int64_t cur_memory;

    struct sc_rnode_t *root_node;
    struct sc_partition_t *partition;
};

/* cache */
struct sc_cache_t {

    struct sc_cnode_t *root_node;
    struct sc_partition_t *partition;

    int64_t cur_size;

    struct sc_cnode_t *ring;
};

struct sc_partition_t {

    const char *id;

    int allow_ring;
    int allow_stats;
    int allow_lvals;
    int allow_cache;
    int allow_shared;
    int allow_session;
    int allow_collect;
    int allow_shcollect;

    struct sc_ring_t *ring;
    struct sc_cache_t *cache;
    struct sc_session_t *shsession;
    struct sc_session_t *lvsession;
    struct sc_session_t *snsession;

    int64_t cache_hits;
    int64_t cache_misses;
    int64_t cache_adds;
    int64_t cache_dels;
    int64_t cache_expires;

    int64_t cache_max;

    int64_t ring_nodes_limit;
    int64_t ring_memory_limit;

#ifdef SESSION_MEMORY_ACCOUNTING
    int64_t memory_allocated;
    int64_t memory_limit;
#endif

#ifdef SESSION_NODES_ACCOUNTING
    int64_t nodes_allocated;
    int64_t nodes_limit;
#endif

#ifdef SEPARATE_SESSIONTREES
    struct sc_tree_t *session_tree;
#endif
};


/* session container */
struct sc_session_t {

    int flags;
    int idx;    /* position in store.c:sessionidx. (which is for)
		   shortcutting session_open_by_idx() and avoidin
		   searching from global session tree by tag */

    time_t expires;
    time_t maxidle;

    char *tag;
    char *secret;

#ifdef SESSION_NODES_ACCOUNTING
    int64_t nodes_allocated;
#endif
#ifdef SESSION_MEMORY_ACCOUNTING
    int64_t memory_allocated;
#endif 

    struct sc_node_t *root_node;
    struct sc_partition_t *partition;
};


/* global statistics */
struct sc_stats_t {

    int64_t sessiontable_max;

    int64_t obstack_size;
    int64_t mallocs_allocated;
    int64_t mallocs_freed;
    int64_t mallocs_total;

    int64_t req_count_success;
    int64_t req_count_invalid;
    int64_t req_count_unknown;

    int64_t session_creates;
    int64_t session_reuses;
    int64_t session_destroys;
    int64_t session_expires;
    int64_t session_clears;

    int64_t session_hit_by_idx;
    int64_t session_miss_by_idx;
    int64_t session_miss_by_search;
    int64_t session_secret_mismatches;

    int64_t all_cache_hits;
    int64_t all_cache_misses;
    int64_t all_cache_adds;
    int64_t all_cache_dels;
    int64_t all_cache_expires;

    int64_t io_requests;
    int64_t io_bytes_read;
    int64_t io_bytes_written;
    int64_t io_socks_accepted;
    int64_t io_socks_accept_fails;
    int64_t io_socks_closed;

#ifdef STAT_ME_HARDER
    int64_t io_polls;

    int64_t req_t_getopts;
    int64_t req_t_setopts;
    int64_t req_t_srvstats;

    int64_t req_t_adds;
    int64_t req_t_sets;
    int64_t req_t_replaces;
    int64_t req_t_unsets;
    int64_t req_t_stats;
    int64_t req_t_colls;
    int64_t req_t_gets;

    int64_t req_t_shadds;
    int64_t req_t_shsets;
    int64_t req_t_shreplaces;
    int64_t req_t_shunsets;
    int64_t req_t_shstats;
    int64_t req_t_shcolls;
    int64_t req_t_shgets;

    int64_t req_t_chgets;
    int64_t req_t_chputs;
    int64_t req_t_chclears;

    int64_t req_t_vadds;
    int64_t req_t_vsubs;
    int64_t req_t_vgets;
    int64_t req_t_vsets;
    int64_t req_t_vunsets;
    int64_t req_t_vbrks;

    int64_t req_t_rngets;
    int64_t req_t_rnsizes;
    int64_t req_t_rnclears;
    int64_t req_t_rnpushes;
    int64_t req_t_rnpops;
    int64_t req_t_rnshifts;
    int64_t req_t_rnunshifts;
    int64_t req_t_rnrotfs;
    int64_t req_t_rnrotbs;
    
    int64_t req_t_shdlrgets;
    int64_t req_t_shdlrputs;
    int64_t req_t_shdlrdrops;

    int64_t req_t_creates;
    int64_t req_t_renames;
    int64_t req_t_drops;
    
    int64_t req_t_lasterrs;
    int64_t req_t_nones;
    int64_t req_t_unknowns;

    int64_t wblock_allocs;

    int64_t trees_created;
    int64_t trees_freed;
    int64_t trees_out;

    int64_t tree_nodes_created;
    int64_t tree_nodes_freed;
    int64_t tree_nodes_out;
#endif
};
extern struct sc_stats_t *sc_stats;


/* local configuration */
#define SC_CONF_PROGNAME               PROGNAME
#define SC_CONF_RUNAS_USER             "nobody"
#define SC_CONF_RUNAS_GROUP            "nogroup"
#ifdef ENABLE_DEVEL
# define SC_CONF_CONFIG_PATH           "./" PROGNAME ".conf"
# define SC_CONF_PIDFILE_PATH          "/tmp/" PROGNAME ".pid"
#else
# define SC_CONF_CONFIG_PATH           SYSCONFDIR "/" PROGNAME ".conf"
# define SC_CONF_PIDFILE_PATH          LOCALSTATEDIR "/" PROGNAME ".pid"
#endif

/* socket */
#define SC_CONF_LISTEN_BACKLOG         100
#define SC_CONF_READ_BUFFERLEN         (80 * 1024)
#define SC_CONF_SOCK_RCVBUF            (64 * 1024)
#define SC_CONF_SOCK_SNDBUF            (64 * 1024)
#define SC_CONF_USOCK_OWNER            NULL
#define SC_CONF_USOCK_GROUP            NULL
#define SC_CONF_USOCK_MASK             0666

/* sockets and ports */
#if defined(ENABLE_DEVEL)
# define SC_CONF_USOCK_PATH            "/tmp/scached-devel.sock"
#elif defined(SCACHED_SOCK)
# define SC_CONF_USOCK_PATH            SCACHED_SOCK
#else
# define SC_CONF_USOCK_PATH            "/var/run/" PROGNAME ".sock"
#endif

#if defined(SCACHED_PORT)
# define SCACHE_PTO_LISTEN_PORT        SCACHED_PORT
#else
# define SCACHE_PTO_LISTEN_PORT        50659
#endif

/* debug */
#ifdef ENABLE_DEVEL
# define SC_CONF_DEBUG                  1
# define SC_CONF_NODETACH               1
#else
# define SC_CONF_DEBUG                  0
# define SC_CONF_NODETACH               0
#endif
#define SC_CONF_LOGLEVEL               SC_LOG_NOTICE
#define SC_CONF_SYSLOG_FAC             LOG_USER
#define SC_CONF_SYSLOG_TAG             SC_CONF_PROGNAME

/* misc */
#define SC_CONF_EXPIRE_DELAY           1800
#define SC_CONF_OBSTACK_CHUNK          (128 * 1024)
#define SC_CONF_MAX_REQSIZE            SCACHE_PTO_MAX_REQLEN

/* partitions */
#define SC_CONF_ENABLE_AUTHPARTITION    1
#define SC_CONF_ENABLE_HOSTPARTITION    1
#define SC_CONF_DEFPART_ALLOW_RING      1
#define SC_CONF_DEFPART_ALLOW_CACHE     1
#define SC_CONF_DEFPART_ALLOW_LVALS     1
#define SC_CONF_DEFPART_ALLOW_SHARED    1
#define SC_CONF_DEFPART_ALLOW_SESSION   1
#define SC_CONF_DEFPART_ALLOW_STATS     1
#define SC_CONF_DEFPART_ALLOW_COLLECT   1
#define SC_CONF_DEFPART_ALLOW_SHCOLLECT 1
#define SC_CONF_DEFPART_CACHEMAX        (16 * 1024 * 1024)
#define SC_CONF_DEFPART_RINGNODESLIMIT  (16 * 1024 * 1024)
#define SC_CONF_DEFPART_RINGMEMORYLIMIT (16 * 1024 * 1024)
#define SC_CONF_DEFPART_MLIMIT          ((int64_t)1 << 40) /* 1TiB */
#define SC_CONF_DEFPART_NLIMIT          ((int64_t)1 << 40) /* 1TiB */


struct sc_config_t {

    /* prog */
    int program_argc;
    char **program_argv;

    uid_t runas_uid;
    gid_t runas_gid;

    const char *progname;
    const char *config_path;
    const char *pidfile_path;

    /* socket */
    int read_bufferlen;
    int sock_sndbuf;
    int sock_rcvbuf;

    uid_t usock_owner;
    uid_t usock_group;
    int usock_mask;
    const char *usock_path;

    int listen_backlog;
    int listen_port;
    int disable_unix;
    int disable_ipv4;
    struct sc_list_t *listen_addrs_ipv4;
#ifdef ENABLE_IPV6
    int disable_ipv6;
    struct sc_list_t *listen_addrs_ipv6;
#endif

    /* debug */
    int debug;
    int nodetach;
    int loglevel;
    int syslog_fac;
    const char *syslog_tag;

    /* misc */
    int obstack_chunk;
    int expire_delay;
    int max_reqsize;

    /* partitions */
    int defpart_ring;
    int defpart_cache;
    int defpart_lvals;
    int defpart_stats;
    int defpart_shared;
    int defpart_session;
    int defpart_collect;
    int defpart_shcollect;

    int64_t defpart_cachemax;
    int64_t defpart_ringnodeslimit;
    int64_t defpart_ringmemorylimit;
#ifdef SESSION_MEMORY_ACCOUNTING
    int64_t defpart_mlimit;
#endif
#ifdef SESSION_NODES_ACCOUNTING
    int64_t defpart_nlimit;
#endif

    int enable_host_partitions;
    int enable_auth_partitions;


}; 

extern const struct sc_config_t *sc_config;


/* global variables */
/* extern int config_debug; */

#ifdef STAT_ME_HARDER
struct sc_memqueue_t {
    int64_t out;
    int64_t total;
    int64_t allocs;
    const char *token;
    struct sc_list_t *queue;
};

extern struct sc_memqueue_t memq16;
extern struct sc_memqueue_t memq32;
extern struct sc_memqueue_t memq64;
extern struct sc_memqueue_t memq128;
extern struct sc_memqueue_t memq256;
extern struct sc_memqueue_t memq512;
extern struct sc_memqueue_t memq1024;

typedef struct sc_memqueue_t sc_allocq_t;
#else
extern struct sc_list_t *memq16;
extern struct sc_list_t *memq32;
extern struct sc_list_t *memq64;
extern struct sc_list_t *memq128;
extern struct sc_list_t *memq256;
extern struct sc_list_t *memq512;
extern struct sc_list_t *memq1024;

typedef struct sc_list_t * sc_allocq_t;
#endif



int sock_spin();


struct sc_wblock_t *proto_request(const char *, int, char *, int);


typedef void (sc_tree_callback_t)(const char *, void *, void *);

struct sc_tree_t *tree_create();
void *tree_add(struct sc_tree_t *, char *, void *, int);
void *tree_set(struct sc_tree_t *, char *, void *, int);
void *tree_repl(struct sc_tree_t *, char *, void *, int);

void *tree_add_referable_d(struct sc_tree_t *, char *, void *);
void *tree_set_referable_d(struct sc_tree_t *, char *, void *);
void *tree_repl_referable_d(struct sc_tree_t *, char *, void *);

void *tree_add_referable_k(struct sc_tree_t *, char *, void *, int);
void *tree_set_referable_k(struct sc_tree_t *, char *, void *, int);
void *tree_repl_referable_k(struct sc_tree_t *, char *, void *, int);

void *tree_add_referable_kd(struct sc_tree_t *, char *, void *);
void *tree_set_referable_kd(struct sc_tree_t *, char *, void *);
void *tree_repl_referable_kd(struct sc_tree_t *, char *, void *);


void *tree_remove(struct sc_tree_t *, const char *);
void *tree_fetch(struct sc_tree_t *, const char *);
void *tree_fetch_both(struct sc_tree_t *, const char *,
		      char **keyslot, void *dataslot);
void tree_destroy_unref(struct sc_tree_t *);
void tree_destroy(struct sc_tree_t *,
		  sc_tree_callback_t,
		  void *);
void tree_enum_asc(struct sc_tree_t *,
		   sc_tree_callback_t,
		   void *);
void tree_enum_desc(struct sc_tree_t *,
		    sc_tree_callback_t,
		    void *);

int tree_isempty(struct sc_tree_t *tree);

/* ========================
 * store.c
 * ======================== */

int store_init();
void store_unref(char *buffer);
int store_request(char *, int, char **);

struct sc_session_t *session_create(const char *tag,
				    const char *secret,
				    const char *auth,
				    const char *host,
				    int flags,
				    int expires);

struct sc_session_t *session_open(const char *tag,
				  const char *secret,
				  const char *auth,
				  const char *host,
				  int index);

struct sc_session_t *session_host_lval(struct sc_session_t *session);
struct sc_session_t *session_host_shared(struct sc_session_t *session);
struct sc_session_t *
session_host_session(const char *auth, const char *host);

int session_rename(struct sc_session_t *, const char *newtag);

int session_drop(struct sc_session_t *);

int session_reset(struct sc_session_t *);

int session_add(struct sc_session_t *, char *, char *, int);
int session_set(struct sc_session_t *, char *, char *, int);
int session_replace(struct sc_session_t *,char *, char *, int);
int session_unset(struct sc_session_t *, char *);
int session_get(struct sc_session_t *, char *, char **, int *);
char *session_collect(struct sc_session_t *, char *);

/* cache */
int cache_put(struct sc_session_t *session,
	      char *key, char *data, int datalen);
int cache_get(struct sc_session_t *session,
	      char *key, char **dataptr, int *datalen);
int cache_clear(struct sc_session_t *session,
		char *key);

/* ring */
int ring_size(struct sc_session_t *session, char *key);
int ring_purge(struct sc_session_t *session, char *key);
int ring_add(struct sc_session_t *session,
	     char *key, char *data, int datalen, int op);
int ring_fetch(struct sc_session_t *session,
	       char *key, char **dataptr, int *datalen, int op);
int ring_rotate(struct sc_session_t *session,
		char *key, int count);

int config_dump_default_config();
int config_dump_compiled_config();

char *config_dump_stats(struct sc_session_t *);



enum scache_nodetypes_t
session_stat(struct sc_session_t *,
	     char *, char **, int *);


void session_expire();

#ifdef STAT_ME_HARDER
void *sc_alloc_s(struct sc_memqueue_t *, int);
#else
void *sc_alloc_s(struct sc_list_t **, int);
#endif

void sc_free(void *);
void sc_membinfo(void *);

void *sc_obs_alloc(int);
void *sc_obs_calloc(int);


void stats_print();


struct sc_partition_t *config_get_hostinfo(const char *id);
struct sc_partition_t *config_get_authinfo(const char *id);

struct sc_partition_t *config_get_default_partition();
struct sc_partition_t *config_get_unlimited_partition();

/* network-to-host */
static inline int64_t
sc_ntohll(char *buffer) {
    return (((int64_t)buffer[0] << 56) & 0xff00000000000000LL) |
	(((int64_t)buffer[1] << 48) & 0xff000000000000LL) |
	(((int64_t)buffer[2] << 40) & 0xff0000000000LL) |
	(((int64_t)buffer[3] << 32) & 0xff00000000LL) |
	(((int64_t)buffer[4] << 24) & 0xff000000LL) |
	(((int64_t)buffer[5] << 16) & 0xff0000LL) |
	(((int64_t)buffer[6] << 8) & 0xff00LL) |
	((int64_t)buffer[7] & 0xffLL);
	
}

static inline int32_t
sc_ntohl(char *buffer) {
    return ((buffer[0] << 24) & 0xff000000) |
	((buffer[1] << 16) & 0xff0000) |
	((buffer[2] << 8) &0xff00) |
	(buffer[3] & 0xff);
}

static inline int16_t
sc_ntohs(char *buffer) {
    
    return ((buffer[0] << 8) & 0xff00) |
		(buffer[1] & 0xff);
}

/* host-to-network */
static inline char *
sc_htonll(char *buffer, int64_t val) {

    buffer[7] = (char)(val & 0xff);
    buffer[6] = (char)((val >> 8) & 0xff);
    buffer[5] = (char)((val >> 16) & 0xff);
    buffer[4] = (char)((val >> 24) & 0xff);
    buffer[3] = (char)((val >> 32) & 0xff);
    buffer[2] = (char)((val >> 40) & 0xff);
    buffer[1] = (char)((val >> 48) & 0xff);
    buffer[0] = (char)((val >> 56) & 0xff);
    return buffer + 8;
}

static inline char *
sc_htonl(char *buffer, int32_t val) {

    buffer[3] = (char)(val & 0xff); 
    buffer[2] = (char)((val >> 8) & 0xff);
    buffer[1] = (char)((val >> 16) & 0xff);
    buffer[0] = (char)((val >> 24) & 0xff);
    return buffer + 4;
}

static inline char *
sc_htons(char *buffer, int16_t val) {

    buffer[1] = (char)(val & 0xff); 
    buffer[0] = (char)((val >> 8) & 0xff);
    return buffer + 2;
}

/* memref */
static inline int
sc_refcount(void *addr) {
    struct sc_memb_t *memb = (struct sc_memb_t *)addr;
    return (--memb)->refcount;
}

static inline void *
sc_ref(void *addr) {

    struct sc_memb_t *memb = (struct sc_memb_t *)addr;
    memb--;
    
    assert((memb->refcount > 0));
    assert((memb->b00f == (void *)0xb00f));

    memb->refcount++;
    return addr;
}

static inline void
sc_unref(void *addr) {

    struct sc_memb_t *memb = (struct sc_memb_t *)addr;
    memb--;

    assert((addr));
    assert((memb->refcount > 0));
    assert((memb->b00f == (void *)0xb00f));

    if ((--memb->refcount) < 1) sc_free(addr);
}

/* sc_alloc is static here, so that compiler
   can optimize if out when allocating compile time
   known constant sizes */
static void *
sc_alloc(int size) {

    struct sc_memb_t *memb;

    /* probable */
    if (size <= 16) return sc_alloc_s(&memq16, 16);
    if (size <= 32) return sc_alloc_s(&memq32, 32);
    if (size <= 64) return sc_alloc_s(&memq64, 64);

    /* rest from queue */
    if (size <= 256) {
	if (size <= 128) return sc_alloc_s(&memq128, 128);
	else return sc_alloc_s(&memq256, 256);
    } else if (size <= 512) {
	return sc_alloc_s(&memq512, 512);
    } else if (size <= 1024) {
	return sc_alloc_s(&memq1024, 1024);
    }


    /* malloc */
    memb = (struct sc_memb_t *)malloc(size + sizeof(struct sc_memb_t));
    memb->refcount = 1;
    memb->queue = NULL;
    sc_stats->mallocs_allocated++;
    sc_stats->mallocs_total += size + sizeof(struct sc_memb_t);

    /* printf("sc_alloc: %p (%i)\n", memb, size); */
    assert((memb->b00f = (void *)0xb00f));
    return (void *)(++memb);
}

static inline void *
sc_memdup(const void *src, int len) {

    char *addr = sc_alloc(len);
    return memcpy(addr, src, len);
}

static inline char *
sc_strdup(const char *src) {

    char *addr = sc_alloc(strlen(src) + 1);
    return strcpy(addr, src);
}

#endif /* _SCACHE_H_ */


/*
  Local variables: ***
  mode: c ***
  c-basic-offset: 4 ***
  End: ***
*/

