/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */


/* 
 * NOTES:
 *
 * excuses for certain parts:
 * - tried to avoid syscalls 
 * - tried to avoid inmem copying 
*/

#include "scache.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <alloca.h>
#include <fcntl.h>
#include <syslog.h>
#include <limits.h>
#include <sys/uio.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <netdb.h>

#ifdef ENABLE_POLL
# include <poll.h>
#endif

#ifdef ENABLE_EPOLL
# include <sys/epoll.h>
# define EPOLL_FDSIZE         128
# define EPOLL_WAITSIZE       128
#endif

#ifdef ENABLE_IPV6
# include <netinet/ip6.h>
#endif


#define PEER_ARRAY_INCREMENT 64



struct sc_sock_t {

    int fd;
    int pfd;
    int outpos;
    int outlen;

    int in_pos;
    int in_reqcnt;
    int in_length;
    int in_remainder;

    char in_hdr[4];
    char *in_buf;

#ifdef ENABLE_EPOLL
    enum {
	EPFD_NONE = 0,
	EPFD_WRITE,
	EPFD_INFD
    } epoll_state;
#endif /* ENABLE EPOLL */

#ifdef ENABLE_SELECT
    int select_wrflag;
#endif /* ENABLE_SELECT */

    struct sc_wblock_t *out_wblocks;

    char *sockname;
#ifdef ENABLE_IPV6
    struct sockaddr_storage sa;
#else
    struct sockaddr_in sa;
#endif /* ENABLE_IPV6 */

    int (*sock_pollin)(int);
    int (*sock_pollout)(int);
};

static struct sc_accept_blocklist_t {
    int idx;
    struct sc_accept_blocklist_t *next;
} *sc_accept_blocklist = NULL;

/* predef */
static int sock_read(int);
static int sock_write(int);
static int sock_accept(int);

/* local */
#if defined(ENABLE_EPOLL) || defined(ENABLE_POLL)
static int epollfd    = -1;
#endif
static int iov_max    = 16;

static int peer_array_size = 0;
static struct sc_sock_t *peer_array = NULL;

#ifdef ENABLE_POLL
static int peer_count = 0;
static struct pollfd *peer_pfds = NULL;
#endif /* ENABLE_POLL */

#ifdef ENABLE_SELECT
static fd_set peer_rdset;
static fd_set peer_wrset;
#endif


/* iowait-loop breakout */
static int breakout = 0;


#ifdef ENABLE_POLL
static int
sock_add_poll(int fd, const char *id,
	      struct sockaddr *sa,
	      int (*sock_pollin)(int),
	      int (*sock_pollout)(int)) {

    if (peer_count >= peer_array_size) {

	peer_array_size += PEER_ARRAY_INCREMENT;
	peer_pfds =  realloc(peer_pfds, 
			     sizeof(peer_pfds[0]) * peer_array_size);
	
	peer_array = realloc(peer_array, 
			     sizeof(peer_array[0]) * peer_array_size);
    }
    
    memset(&peer_pfds[peer_count], 0, sizeof(peer_pfds[0]));
    peer_pfds[peer_count].fd = fd;
    peer_pfds[peer_count].events = POLLIN;
    
    memset(&peer_array[peer_count], 0, sizeof(peer_array[0]));
    peer_array[peer_count].fd = fd;
    
    peer_array[peer_count].sockname = id ?
	sc_strdup(id) : sc_strdup("");
    
    peer_array[peer_count].sock_pollin = sock_pollin;
    peer_array[peer_count].sock_pollout = sock_pollout;
    
    return peer_count++;
}
#endif /* ENABLE_POLL */

#if defined(ENABLE_EPOLL) || defined(ENABLE_SELECT)
static int
sock_add_fill_peer(int fd, const char *id,
		   struct sockaddr *sa,
		   int (*sock_pollin)(int),
		   int (*sock_pollout)(int)) {

    if (fd >= peer_array_size) {
	peer_array_size += PEER_ARRAY_INCREMENT;
	peer_array = realloc(peer_array, 
			     sizeof(peer_array[0]) * peer_array_size);
    }

    memset(&peer_array[fd], 0, sizeof(peer_array[0]));
    peer_array[fd].fd = fd;
    peer_array[fd].sockname = id ?
	sc_strdup(id) : sc_strdup("");

    peer_array[fd].sock_pollin = sock_pollin;
    peer_array[fd].sock_pollout = sock_pollout;

    return fd;
}
		   
#endif /* ENABLE_EPOLL || ENABLE_SELECT */

#ifdef ENABLE_EPOLL
static int
sock_add_epoll(int fd, const char *id,
	       struct sockaddr *sa,
	       int (*sock_pollin)(int),
	       int (*sock_pollout)(int)) {

    struct epoll_event ev;

    ev.events = EPOLLIN;
    ev.data.fd = fd;
    if (epoll_ctl(epollfd, EPOLL_CTL_ADD, fd, &ev) < 0) {

	syslog(LOG_ERR, "%s: epoll_ctl(%i, EPOLLIN) failed: %s",
	       __FUNCTION__, fd, strerror(errno));

	return -1;
    }

    return sock_add_fill_peer(fd, id, sa, sock_pollin, sock_pollout);
}
#endif /* ENABLE_EPOLL */

#ifdef ENABLE_SELECT
static int
sock_add_select(int fd, const char *id,
		struct sockaddr *sa,
		int (*sock_pollin)(int),
		int (*sock_pollout)(int)) {
 
    if (fd >= FD_SETSIZE) {
	syslog(LOG_ERR, "%s: fd %i larger than FD_SETSIZE",
	       __FUNCTION__, fd);
	return -1;
    }

    FD_SET(fd, &peer_rdset);
    return sock_add_fill_peer(fd, id, sa, sock_pollin, sock_pollout);
}
#endif /* ENABLE_SELECT */

static int
sock_add(int fd, struct sockaddr *sa,
	 int (*sock_pollin)(int),
	 int (*sock_pollout)(int)) {

    const char *id;

    /* peer-id */
    if (((struct sockaddr_un *)sa)->sun_family == AF_UNIX) {
	id = "<unix>";
    } else {
#ifdef ENABLE_IPV6
	id = inet_ntop(((struct sockaddr_in6 *)sa)->sin6_family, sa,
		       alloca(INET6_ADDRSTRLEN), INET6_ADDRSTRLEN);
#else 
	id = inet_ntoa(((struct sockaddr_in *)sa)->sin_addr);
#endif /* ENABLE_IPV6 */
    }


#if defined(ENABLE_POLL) || defined(ENABLE_EPOLL)
 #ifdef ENABLE_POLL
    if (epollfd < 0) {
	return sock_add_poll(fd, id, sa, sock_pollin, sock_pollout);

    }
 #endif /* ENABLE_POLL */

 #ifdef ENABLE_EPOLL
    if (epollfd >= 0) {
	return sock_add_epoll(fd, id, sa, sock_pollin, sock_pollout);
    }
 #endif /* ENABLE_EPOLL */

    return -1;

#elif defined(ENABLE_SELECT)
    return sock_add_select(fd, id, sa, sock_pollin, sock_pollout);
#endif
}

static int
sock_remove(int idx) {

    struct sc_sock_t *peer;

    /* sanity-check */
#ifdef ENABLE_POLL
    //assert((idx <= peer_count));
#else
    assert((idx < peer_array_size));
#endif

    peer = &(peer_array[idx]);

    /* free */
    if (peer->in_buf) sc_unref(peer->in_buf);
    if (peer->sockname) sc_unref(peer->sockname);
    while(peer->out_wblocks) {

	struct sc_wblock_t *next = peer->out_wblocks->next;

	sc_unref(peer->out_wblocks->base);
	sc_unref(peer->out_wblocks);
	peer->out_wblocks = next;
    }

#ifdef ENABLE_SELECT
    FD_CLR(peer_array[idx].fd, &peer_rdset);
    FD_CLR(peer_array[idx].fd, &peer_wrset);
#endif

    /* close-sock */
    while(close(peer_array[idx].fd) < 0 && errno == EINTR);
    peer_array[idx].fd = -1;

#ifdef ENABLE_POLL
    if (epollfd < 0) {
	if (peer_count-- > idx) {
	    peer_array[idx] = peer_array[peer_count];
	    peer_pfds[idx] = peer_pfds[peer_count];
	    peer_pfds[idx].revents = 0;
	}
    }
#endif /* ENABLE_POLL */

    return 0;
}

static int
sock_block_accept(int idx) {

    struct sc_accept_blocklist_t *item;

#ifdef ENABLE_EPOLL
    struct epoll_event ev;

    ev.events = 0;
    ev.data.fd = peer_array[idx].fd;
    if (epoll_ctl(epollfd, EPOLL_CTL_MOD, peer_array[idx].fd, &ev) < 0) {
	syslog(LOG_ERR, "%s: epoll_ctl(%i, 0) failed: %s",
	       __FUNCTION__, peer_array[idx].fd, strerror(errno));
	return -1;
    }

#elif ENABLE_POLL
    
    /* events to null */
    peer_pfds[idx].events = 0;

#elif ENABLE_SELECT

    /* clear bitmask */
    FD_CLR(peer_array[idx].fd, &peer_rdset);
#endif

    /* queue to be unblocked when fds become available */
    item = sc_alloc(sizeof(struct sc_accept_blocklist_t) + _NMTOBO);
    item->idx = idx;
    item->next = sc_accept_blocklist;
    sc_accept_blocklist = item;

    return 0;
}

static int
sock_unblock_accept() {

    int rv = 0;
    struct sc_accept_blocklist_t *item;

    while((item = sc_accept_blocklist)) {
	
#ifdef ENABLE_EPOLL
	struct epoll_event ev;

	ev.events = EPOLLIN;
	ev.data.fd = peer_array[item->idx].fd;
	if (epoll_ctl(epollfd, EPOLL_CTL_MOD, 
		      peer_array[item->idx].fd, &ev) < 0) {

	    rv = -1; /* mark-err */
	    syslog(LOG_ALERT, "%s: re-enabling epoll() on listening "
		   "socket %i failed: %s, daemon has became at least "
		   "partly unavailable. Oops! This is not good...", 
		   __FUNCTION__, peer_array[item->idx].fd,
		   strerror(errno));

	    /* TODO: exit(restart) */
	}
#elif ENABLE_POLL

	/* re-eanble events */
	peer_pfds[item->idx].events = POLLIN;
	
#elif ENABLE_SELECT

	/* re-enable bit */
	FD_SET(peer_array[item->idx].fd, &peer_rdset);
#endif

	/* next and release */
	sc_accept_blocklist = item->next;
	sc_unref(item); 
    }

    if (sc_config->loglevel >= SC_LOG_ERROR) {
	syslog(LOG_ERR, "re-enabling listening socket. accepts are "
	       "not blocked any more");
    }
    return rv;
}

static int
sock_kill(int idx) {

    assert((peer_array[idx].fd >= 0));

    if (peer_array[idx].sock_pollin == sock_accept) {
	/* don't kill accept sock */
	return 0;
    }

    while(shutdown(peer_array[idx].fd, 2) < 0 &&  errno == EINTR);
    sock_remove(idx);

    /* socket freed, re-enable accept socket if blocked on 
       fd-limit overflow */
    if (sc_accept_blocklist) sock_unblock_accept();

    return 0;
}

static inline void
wblock_ff(struct sc_wblock_t **wp) {
    while(*wp && (*wp)->next) wp = &((*wp)->next);
}


static int
sock_accept(int idx) {

    int val;
    int nsock;
    socklen_t sa_len;
    struct linger li;
#ifdef ENABLE_IPV6
    struct sockaddr_storage sa;
#else
    struct sockaddr_in sa;
#endif /* ENABLE_IPV6 */

    do {
	sa_len = sizeof(sa);
	nsock = accept(peer_array[idx].fd, 
		       (struct sockaddr *)&sa, &sa_len);
    } while(nsock < 0 && errno == EINTR);


    if (nsock < 0 && errno == EAGAIN) {
	/* EWOULDBLOCK */
	return -1;

    } else if (nsock < 0) {

	if ((errno == EMFILE) || (errno == ENFILE)) {

	    if (sc_config->loglevel >= SC_LOG_ERROR) {
		syslog(LOG_ERR, "accepting new connection failed. "
		       "too many open files. blocking accepts "
		       "temporarily");
	    }

	    /* remove pollin temporarily from socket */
	    sock_block_accept(idx);
	    

	} else if (sc_config->loglevel >= SC_LOG_ERROR) {
	    syslog(LOG_ERR, "sock_accept() failed: %s",
		   strerror(errno));
	}

	return -1;
	
    } else {
	
	/* tausta-close() plle */
	li.l_onoff = 1;
	li.l_linger = 5;
	do {
	    val = setsockopt(nsock, SOL_SOCKET, SO_LINGER,
			     (void *)&li, sizeof(li));
	} while(val < 0 && errno == EINTR);
	
	if (!val) {

	    if (setsockopt(nsock, SOL_SOCKET, SO_SNDBUF, 
			   &(sc_config->sock_sndbuf),
			   sizeof(sc_config->sock_sndbuf)) < 0) {

		/* not hazard */
		syslog(LOG_ERR, 
		       "%s: setsockopt(%i, SO_SNDBUF, %i): failed",
		       __FUNCTION__, nsock, sc_config->sock_sndbuf);
	    }


	    if (setsockopt(nsock, SOL_SOCKET, SO_RCVBUF, 
			   &(sc_config->sock_rcvbuf),
			   sizeof(sc_config->sock_rcvbuf)) < 0) {

		/* not hazard */
		syslog(LOG_ERR, 
		       "%s: setsockopt(%i, SO_RCVBUF, %i): failed",
		       __FUNCTION__, nsock, sc_config->sock_rcvbuf);
	    }


	    val = fcntl(nsock, F_GETFL);
	    if (val >= 0) {
		val |= O_NONBLOCK;
		if (!fcntl(nsock, F_SETFL, val)) {
		    
		    val = sock_add(nsock, (struct sockaddr *)&sa,
				   sock_read, sock_write);
		    if (val >= 0) {

			if (sc_config->loglevel >= SC_LOG_DEBUG) {
			    syslog(LOG_DEBUG, "%s: accepted fd %i\n", 
				   __FUNCTION__, nsock);
			}
    
			sc_stats->io_socks_accepted++;
			peer_array[val].sa = sa;
			return val;

		    } else {
			goto failout;
		    }
		}
	    }
	}

	if (sc_config->loglevel >= SC_LOG_ERROR) {
	    syslog(LOG_ERR, "%s: setoptions failed: %s",
		   __FUNCTION__, strerror(errno));
	}
    }

 failout:

    while(shutdown(nsock, 2) < 0 && errno == EINTR);
    while(close(nsock) < 0 && errno == EINTR);

    sc_stats->io_socks_accept_fails++;
    return -1;
}

static void
parse_reqhdr(struct sc_sock_t *peer) {

    assert((peer->in_pos == 4));

    peer->in_length = peer->in_reqcnt = sc_ntohl(peer->in_hdr);
    peer->in_length = (peer->in_length & 0xffffff) - 4 /* self */;
    peer->in_reqcnt = (peer->in_reqcnt >> 24) & 0x7f;

    assert((peer->in_length < SCACHE_PTO_MAX_REQLEN));
}


static struct sc_wblock_t *
handle_read_req(struct sc_sock_t *peer, char *req, int length) {


    struct sc_wblock_t *wblock = NULL;
    struct sc_wblock_t **wblockp = &wblock;

    while(length > 0) {

	while((peer->in_pos < 4) && (length-- > 0)) {
	    peer->in_hdr[peer->in_pos++] = *req++;
	}

	if (peer->in_pos == 4) {

	    parse_reqhdr(peer);
	    if (peer->in_length > sc_config->max_reqsize) {
		syslog(LOG_ERR, "oversized request from peer '%s'",
		       peer->sockname);
		break;
	    }
	}

	if (length < peer->in_length) {
	    
	    /* need more */
	    if (length > 0) {

		assert((!peer->in_buf));
		peer->in_buf = sc_alloc(peer->in_length + _NMTOBO);
		peer->in_pos = length;
		peer->in_remainder = peer->in_length - peer->in_pos;
		memcpy(peer->in_buf, req, length);
	    }

	    break;

	} else {

	    /* handle-req */
	    *wblockp = proto_request(peer->sockname, 
				     peer->in_reqcnt, req, 
				     peer->in_length);
	    if (!*wblockp) break; /* fatal */
	    wblock_ff(wblockp);

	    /* reset header */
	    length -= peer->in_length;
	    peer->in_pos = peer->in_length = peer->in_remainder = 0;

	    assert((!length));
	}
    }

    if (length > 0) {

	/* fatal, all must be consumed here release completed and
	   signal to kill socket */
	while(wblock) {
	    *wblockp = wblock->next;
	    sc_unref(wblock->base);
	    sc_unref(wblock);
	    wblock = *wblockp;
	}

	return NULL;
    }

    return wblock;
}

static int
sock_read(int idx) { 

    int rv, iov_count;
    struct iovec iov[2];
    struct sc_sock_t *peer;
    struct sc_wblock_t **wblockp;

    /* rewind to end of write-q */
    peer = &peer_array[idx];
    assert((peer));

    wblockp = &(peer->out_wblocks);
    wblock_ff(wblockp); /* fast-forward to end to add data */

    /* avoid syscalls... avoid inmem copying... */

    if (peer->in_remainder) {

	/* when previous read did not get whole request */
	iov[0].iov_base = peer->in_buf + peer->in_pos;
	iov[0].iov_len = peer->in_remainder;
	iov[1].iov_base = alloca(sc_config->read_bufferlen + _NMTOBO);
	iov[1].iov_len = sc_config->read_bufferlen;
	iov_count = 2;

    } else {

	/* but in most common case we get it in one read
	   and can complete everything in stack */
	iov[0].iov_base = alloca(sc_config->read_bufferlen + _NMTOBO);
	iov[0].iov_len = sc_config->read_bufferlen;
	iov_count = 1;

    }

    /* repeat as long as there is something to read */
    for(;;) {

	do { /* EINTR */
	    rv = readv(peer->fd, iov, iov_count);
	} while(rv < 0 && errno == EINTR);
	
	if (rv > 0) { /* got something */

	    sc_stats->io_bytes_read += rv;

	    if (iov_count == 2) { /* had incomplete content,ie.
				     need to appened to previous*/
		if (rv >= peer->in_remainder) { 

		    /* but now it is full, so we handle it out */
		    *wblockp = proto_request(peer->sockname, 
					     peer->in_reqcnt,
					     peer->in_buf, 
					     peer->in_length);
		    
		    if (!*wblockp) {

			/* it should have returned something,
			   so this is fatal */
			return -1;
		    }

		    /* returned might always be queue, so forward to end */
		    wblock_ff(wblockp);

		    rv -= peer->in_remainder;
		    peer->in_remainder = 
			peer->in_length = peer->in_pos = 0;

		    sc_unref(peer->in_buf);
		    peer->in_buf = NULL;

		    iov[0] = iov[1]; /* emulate stack-read */

		} else {

		    /* still not enought, so append to previous */
		    assert((peer->in_buf));
		    peer->in_pos += rv;
		    peer->in_remainder -= rv;
		    break; /* back to iowait */
		}
	    }

	    /* (now) everything is completely in stack */
	    if (rv > 0) {

		*wblockp = handle_read_req(peer, iov[0].iov_base, rv);
		if (peer->in_remainder || peer->in_pos < 4) {

		    /* there is probably nothing waiting. stack buffer
		       was sized to get enough

		       but we still need to get EWOULDBLOCK to prevent
		       empty poll event */

		    if (*wblockp) wblock_ff(wblockp);

		    if (peer->in_remainder > 0) {
			iov[1] = iov[0]; /* iov[0] is stackbuf */
			iov[0].iov_base = peer->in_buf + peer->in_pos;
			iov[0].iov_len = peer->in_remainder;
			iov_count = 2;
		    }

		} else if (!*wblockp) {

                    /* no partial so should have returned
                       something */
                    return -1;
		}
	    }

	    /* ...all here. buf is consumed, no hints for
	       more data waiting and no errors. so
	       break back in iowait */

	    /* break; wrong, need to read once to mark socket
	       empty, otherwise poll singals fd readable 
	       thought empty (linux at least) */
	    continue; 

	} else if (rv == 0) { /* EOF from client */

	    /* ok */
	    
	} else if (rv < 0) {
	    
	    if (errno == EAGAIN) { /* !EWOULDBLOCK */
		break;
	    } else if (errno == ENOTCONN || /* EOF */
		       errno == ECONNRESET) { 
		break;
	    } else {
		syslog(LOG_ERR, "%s: readv() failed: %s (%i)",
		       __FUNCTION__, strerror(errno), errno);
	    }
	}
	
	/* most fails end up here */
	return -1;
    }

    /* all ok's break to here */
    if (peer->out_wblocks) {

	/* we probably can write, so we fake returned
	   events */
#ifdef ENABLE_POLL
	if (epollfd < 0) {
	    peer_pfds[idx].events |= POLLOUT;
	    peer_pfds[idx].revents |= POLLOUT;
	}
#endif  /* ENABLE_POLL */

#ifdef ENABLE_EPOLL
	peer_array[idx].epoll_state = EPFD_WRITE;
#endif /* ENABLE_EPOLL */

#ifdef ENABLE_SELECT
	peer_array[idx].select_wrflag = 1;
#endif
    }

    return 0;
}

static int
sock_write(int idx) {

    int i, rv, count = 0;
    struct iovec *iov;
    struct sc_sock_t *peer;
    struct sc_wblock_t *wblock;


    peer = &peer_array[idx];
    assert((peer));

    wblock = peer->out_wblocks;
    assert((wblock));

    /* count wqueue length */
    while(wblock) {
	count++;
	wblock = wblock->next;
    }

    /* alloc iovex */
    if (count > iov_max) count = iov_max;
    iov = alloca(count * sizeof(struct iovec));


    while(peer->out_wblocks) { /* loop until all written or EWOULDBLOCK */

	/* alloc and fill iovec */
	wblock = peer->out_wblocks;
	for(i = 0; i < count && wblock; i++) {

	    assert((wblock));
	    assert((wblock->length > wblock->pos));

	    iov[i].iov_base = wblock->base + wblock->pos;
	    iov[i].iov_len = wblock->length - wblock->pos;
	    wblock = wblock->next;

	}

	do { /* EINTR */
	    rv = writev(peer->fd, iov, i);
	} while(rv < 0 && errno == EINTR);

	if (rv > 0) {

	    sc_stats->io_bytes_written += rv;
	    while(rv > 0) {

		wblock = peer->out_wblocks;
		i = wblock->length - wblock->pos;

		if (rv >= i) {

		    /* all out from this wblock */
		    rv -= i;
		    peer->out_wblocks = wblock->next;
		    sc_unref(wblock->base);
		    sc_unref(wblock);

		} else {

		    wblock->pos += rv;
		    return 0;
		}
	    }

	} else if (rv == 0) { /* EOF */

	    syslog(LOG_ERR, "%s: client '%s' disconnected",
		   __FUNCTION__, peer->sockname);
	    break;

	} else if (rv < 0) {
	    if (errno == EAGAIN) { /* EWOULDBLOCK */
		return 0;
	    } else {

		syslog(LOG_ERR, "%s: writev() failed: %s (%i)",
		       __FUNCTION__, strerror(errno), errno);

		break;
	    }
	}
    }

    return peer->out_wblocks ? -1 : 0;
}

#ifdef ENABLE_EPOLL
static int
sock_spin_epoll() {

    int rv;
    int spincounter = 0;
    int timeout = 1000;

    struct epoll_event events[EPOLL_WAITSIZE];

    do {

	rv = epoll_wait(epollfd, events, EPOLL_WAITSIZE, timeout);
	if (rv < 0) { /* got error */
	    if (errno == EINTR) { 
		/* sig */
		continue;
	    } else {
		syslog(LOG_ERR, "%s: epoll() failed: %s",
		       __FUNCTION__, strerror(errno));
		continue;
	    }
	}

	/* sc_lazytime mangling do avoid extra syscalls */
	if (rv == 0) { /* got timeout */
	    
	    /* sc_lazytime need updating on next ioevent,
	       but now we can wait it forever */
	    sc_lazytime = 0; 
	    timeout = -1; 
	    continue;

	} else if (--spincounter < 0) {

	    sc_lazytime = 0; 
	    spincounter = 10; /* max spins between sc_lazytime update */
	}

	timeout = 1000;
	if (!sc_lazytime) { /* shared time */

	    sc_lazytime = time(NULL);
	    spincounter = 10;
	    session_expire(); /* maintainance */
	}

	STAT_HARD(sc_stats->io_polls++);

	if (rv > 0) {

	    int fd;
	    
	    while(rv-- > 0) {

		fd = events[rv].data.fd;

		/* something to read */
		if (events[rv].events & EPOLLIN) {
		    if (peer_array[fd].sock_pollin(fd) < 0) { /* sock_read
								 sock_accept
							      */
			sock_kill(fd);
			continue;
		    }
		}

		/* something to write */
		if ((events[rv].events & EPOLLOUT) ||
		    (peer_array[fd].epoll_state == EPFD_WRITE)) {

		    assert((peer_array[fd].out_wblocks));
		    if (peer_array[fd].sock_pollout(fd) < 0) {
			sock_kill(fd);
			continue;
		    }
		}


		/* nothing left in write-queue */
		if (!peer_array[fd].out_wblocks) {

		    struct epoll_event ev;
		    
		    /* not in epollfd */
		    if (peer_array[fd].epoll_state == EPFD_NONE) continue;

		    /* else remove */
		    ev.events = EPOLLIN;
		    ev.data.fd = fd;
		    if (epoll_ctl(epollfd, EPOLL_CTL_MOD, 
				  fd, &ev) < 0) {
			
			syslog(LOG_ERR,
			       "%s: removing EPOLLOUT for %i failed: %s",
			       __FUNCTION__, fd, strerror(errno));
			sock_kill(fd);

		    } else {
			/* and mark removed */
			peer_array[fd].epoll_state = EPFD_NONE;
		    }

		    continue;
		}


		/* something in queue, but not waiting EPOLLOUT.
		   add to epollfd */
		if (peer_array[fd].out_wblocks && 
		    peer_array[fd].epoll_state == EPFD_WRITE) {

		    struct epoll_event ev;
		    ev.events = EPOLLIN | EPOLLOUT;
		    ev.data.fd = fd;

		    if (epoll_ctl(epollfd, EPOLL_CTL_MOD, fd, &ev) < 0) {

			syslog(LOG_ERR, 
			       "%s: adding EPOLLOUT for %i failed: %s",
			       __FUNCTION__, fd, strerror(errno));
			sock_kill(fd);

		    } else {
			peer_array[fd].epoll_state = EPFD_INFD;
		    }
		}
	    }
	}

	
    } while(!breakout);


    return 0;
}
#endif /* ENABLE_EPOLL */

#ifdef ENABLE_POLL
static int
sock_spin_poll() {

    int i, rv;
    int spincounter = 0;
    int timeout = 1000;

    do {

	rv = poll(peer_pfds, peer_count, timeout);
	if (rv < 0) { /* got error */
	    if (errno == EINTR) {
		/* sig */
		continue;
	    } else {
		syslog(LOG_ERR, "%s: poll(): %s", 
		       __FUNCTION__, strerror(errno));
		continue;
	    }
	}

	/* sc_lazytime mangling do avoid extra syscalls */
	if (rv == 0) { /* got timeout */
	    
	    /* sc_lazytime need updating on next ioevent,
	       but now we can wait it forever */
	    sc_lazytime = 0; 
	    timeout = -1; 
	    continue;

	} else if (--spincounter < 0) {

	    sc_lazytime = 0; 
	    spincounter = 10; /* max spins between sc_lazytime update */
	}


	timeout = 1000;
	if (!sc_lazytime) { /* shared time */

	    sc_lazytime = time(NULL);
	    spincounter = 10;
	    session_expire(); /* maintainance */
	}

	STAT_HARD(sc_stats->io_polls++);

	for(i = 0; i < peer_count; i++) {

	    if ((peer_pfds[i].revents & 
		 peer_pfds[i].events) & POLLIN) {

		/* sock_read */
		if (peer_array[i].sock_pollin(i) < 0) { /* sock_read
							   sock_accept */
		    sock_kill(i);
		    continue;
		}
	    }

	    if ((peer_pfds[i].revents & 
		 peer_pfds[i].events) & POLLOUT) {

		/* sock_write */
		if (peer_array[i].sock_pollout(i) < 0) {
		    sock_kill(i);
		    continue;
		}
	    }

	    if (!peer_array[i].out_wblocks) {
		/* nothing to write */
		peer_pfds[i].events &= ~POLLOUT;
	    }
	}
    } while(!breakout);

    return 0;
}
#endif /* ENABLE_POLL */


#ifdef ENABLE_SELECT
static int
sock_spin_select() {

    int i, rv;
    int spincounter = 0;
    fd_set rdset, wrset;
    struct timeval tv;

    tv.tv_sec = 0;
    tv.tv_usec = 1000;
    do {

	/* very stupid select, but who uses it anymore */
	rdset = peer_rdset;
	wrset = peer_wrset;

	rv = select(FD_SETSIZE /* :b */,
		    &rdset, &wrset, NULL, 
		    tv.tv_usec ? &tv : NULL);

	if (rv < 0) { /* got error */
	    if (errno == EINTR) {
		/* sig */
		continue;
	    } else {
		syslog(LOG_ERR, "%s: select(): %s",
		       __FUNCTION__, strerror(errno));
		continue;
	    }
	}

	/* sc_lazytime mangling do avoid extra syscalls */
	if (rv == 0) { /* got timeout */
	    
	    /* sc_lazytime need updating on next ioevent,
	       but now we can wait it forever */
	    sc_lazytime = 0; 
	    tv.tv_usec = 0;
	    continue;

	} else if (--spincounter < 0) {

	    sc_lazytime = 0; 
	    spincounter = 10; /* max spins between sc_lazytime update */
	}


	tv.tv_usec = 1000;
	if (!sc_lazytime) { /* shared time */

	    sc_lazytime = time(NULL);
	    spincounter = 10;
	    session_expire(); /* maintainance */
	}

	STAT_HARD(sc_stats->io_polls++);

	for(i = 0; i < peer_array_size && rv > 0; i++) {

	    /* in readset */
	    if (FD_ISSET(i, &rdset)) {
		rv--;
		if (peer_array[i].sock_pollin(i) < 0) { /* sock_read
							   sock_accept */
		    sock_kill(i);
		    continue;
		}
	    }

	    /* write-set */
	    if (FD_ISSET(i, &wrset)) {
		rv--;
		if (peer_array[i].sock_pollout(i) < 0) {
		    sock_kill(i);
		    continue;
		}
	    }

	    /* faked in read to avoid syscall */
	    if (peer_array[i].select_wrflag) {

		/* and here we don't decrement rv */
		peer_array[i].select_wrflag = 0;
		if (peer_array[i].sock_pollout(i) < 0) {
		    sock_kill(i);
		    continue;
		}
	    }


	    if (peer_array[i].out_wblocks) {
		FD_SET(i, &peer_wrset);
	    } else {
		FD_CLR(i, &peer_wrset);
	    }
	}

    } while(!breakout);

    return 0;
}
#endif /* ENABLE_SELECT */


static int
sock_create(int pf, struct sockaddr *sa, int sa_len, char *tag) {

    int one;
    int sock;

    sock = socket(pf, SOCK_STREAM, PF_UNSPEC);
    if (sock < 0) {

	/* omit complains for ipv6's unexistence */
	if (pf != AF_INET6 && errno != EAFNOSUPPORT ) {
	    fprintf(stderr, "%s: %s: socket(): %s\n",
		    __FUNCTION__, tag, strerror(errno));
	}

    } else if (((one = fcntl(sock, F_GETFL)) < 0) ||
	       (fcntl(sock, F_SETFL, one | O_NONBLOCK) < 0)) {

	fprintf(stderr, "%s: %s: fcntl(O_NONBLOCK): %s\n",
		__FUNCTION__, tag, strerror(errno));

    } else if (((pf == AF_INET) || (pf == AF_INET6)) &&
	       ((one = 1) &&
		(setsockopt(sock, IPPROTO_TCP,
			    TCP_NODELAY, (void *)&one,
			    sizeof(one)) < 0))) {

	fprintf(stderr, "%s: %s: setsockopt(TCP_NODELAY): %s\n",
		__FUNCTION__, tag, strerror(errno));

    } else if ((pf != AF_UNIX) &&
	       (one = 1,
		setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, 
			   &one, sizeof(one)) < 0)) {

	fprintf(stderr, "%s: %s: setsockopt(REUSEADDR): %s\n",
		__FUNCTION__, tag, strerror(errno));
#ifdef ENABLE_IPV6
    } else if ((pf == AF_INET6) &&
	       (one = 1,
		setsockopt(sock, IPPROTO_IPV6, IPV6_V6ONLY,
			   &one, sizeof(one)) < 0)) {

	fprintf(stderr, "%s: %s: setsockopt(IPV6_V6ONLY): %s\n",
		__FUNCTION__, tag, strerror(errno));

#endif /* ENABLE_IPV6 */
    } else if (bind(sock, sa, sa_len) < 0) {

	fprintf(stderr, "%s: %s: bind(): %s\n",
		__FUNCTION__, tag, strerror(errno));
	
    } else if (listen(sock, sc_config->listen_backlog) < 0) {

	fprintf(stderr, "%s: %s: listen(): %s\n",
		__FUNCTION__, tag, strerror(errno));

    } else {

	/* complete success */
	return sock;
    }

    return -1;
}


int sock_spin() {

#if defined(ENABLE_EPOLL) || defined(ENABLE_POLL)
# ifdef ENABLE_EPOLL
    if (epollfd >= 0) return sock_spin_epoll();
# endif /* ENABLE_EPOLL */

# ifdef ENABLE_POLL
    return sock_spin_poll();
# endif /* ENABLE_POLL */

#elif defined(ENABLE_SELECT)
    return sock_spin_select();
#endif

    syslog(LOG_ERR,
	   "%s: no select(), no poll(), no epoll(), nothing(). "
	   "Hey hey, what can I do.", __FUNCTION__);

    return -1;
}

static int
sock_add_listeners_unix() {

    int rv, sock;
    struct sockaddr_un sa;


#ifdef NDEBUG
    if (sc_config->usock_path[0] != '/') {
	fprintf(stderr, "unix socket path needs to be absolute path. "
		"current path is '%s'\n", 
		sc_config->usock_path);
	return -1;
    }
#endif

    /* remove old */
    do {
	rv = unlink(sc_config->usock_path);
    } while(rv < 0 && errno == EINTR);

    if (rv < 0 && errno != ENOENT) {
	fprintf(stderr, "removing old unix socket '%s' failed: %s.\n",
		sc_config->usock_path, strerror(errno));
	return -1;
    }

    sa.sun_family = AF_UNIX;
    strncpy(sa.sun_path, sc_config->usock_path, sizeof(sa.sun_path));
    
    sock = sock_create(AF_UNIX, (struct sockaddr *)&sa, 
		       SUN_LEN(&sa), "<unix>");
    if (sock >= 0) {

	/* change sock owner */
	if (!geteuid() || !getuid()) {
	    do {
		rv = chown(sc_config->usock_path,
			   sc_config->usock_owner,
			   sc_config->usock_group);
	    } while(rv < 0 && errno == EINTR);
	    
	    if (rv < 0) {
		fprintf(stderr, "cannot change unix socket '%s' "
			"ownership to %u.%u: %s\n", 
			sc_config->usock_path,
			sc_config->usock_owner,
			sc_config->usock_group,
			strerror(errno));
		return -1;
	    }
	}

	do {
	    rv = chmod(sc_config->usock_path, 
		       sc_config->usock_mask);
	} while(rv < 0 && errno == EINTR);
	if (rv < 0) {
	    fprintf(stderr, "cannot change unix socket '%s' "
		    "mask to %o: %s\n",
		    sc_config->usock_path,
		    sc_config->usock_mask,
		    strerror(errno));
	    return -1;
	}

	/* else ok */
	sock_add(sock, (struct sockaddr *)&sa, sock_accept, NULL);
    } else {
	return -1;
    }

    return 0;
}

static int
sock_extract_port(char *host) {

    int port;
    char *cp;

    port = sc_config->listen_port;
    if (host && (cp = strchr(host, (int)'('))) {
	for(port = 0, cp++, port = 0; ; cp++) {
	    
	    if (*cp >= '0' && *cp <= '9') {
		port = port * 10;
		port += *cp - '0';
	    } else if (*cp == ')') {
		/* success */
		if (port < 1 || port > 65535) {
		    fprintf(stderr, 
			    "invalid port %i for hostspec '%s' in config\n",
			    port, host);
		    return -1;
		}
		
		/* null-terminate hostname */
		*strchr(host, (int)'(') = 0;
		break;

	    } else {
		fprintf(stderr,
			"invalid hostspec '%s' in config (%c)\n",
			host, *cp);
		return -1;
	    }
	}
    }

    return port;
}



static int
sock_fill_addr_ipv4(char *hostname, struct sockaddr_in *sa) {

#ifdef HAVE_GETADDRINFO
    int rv;
    struct addrinfo hint, *ainfo;
#else
    struct hostent *he;
#endif

    /* no host */
    if (!hostname || !*hostname) {
	sa->sin_addr.s_addr = INADDR_ANY;
	return 0;
    }

    /* dot notation? */
    sa->sin_addr.s_addr = inet_addr(hostname);
    if (sa->sin_addr.s_addr != INADDR_NONE) {
	return 0;
    }


#ifdef HAVE_GETADDRINFO

    memset(&hint, 0, sizeof(hint));
    hint.ai_family = AF_INET;
    hint.ai_socktype = 0;
    hint.ai_protocol = 0;

    if ((rv = getaddrinfo(hostname, NULL, &hint, &ainfo)) < 0) {

	if (rv == EAI_NONAME) {
	    fprintf(stderr, "host '%s' not found\n", hostname);
	} else {
	    fprintf(stderr, "getaddrinfo() for host '%s' failed: %s\n",
		    hostname, gai_strerror(rv));
	}
	return -1;

    } else {

	sa->sin_addr.s_addr = ((struct sockaddr_in *)
			       (ainfo->ai_addr))->sin_addr.s_addr;
	freeaddrinfo(ainfo);
	return 0;
    }

#else


    if (!(he = gethostbyname(hostname))) {

	fprintf(stderr, "host '%s' not found", hostname);
	return -1;

    } else if (he->h_addrtype != AF_INET ||
	       he->h_length != 4) {

	fprintf(stderr, "no IPV4 address for host '%s'", hostname);
	return -1;
    } else {

	memcpy(&(addr->s_addr), he->h_addr, sizeof(*addr));
	return 0;
    }
#endif

    return -1;
}


#ifdef ENABLE_IPV6
static int
sock_fill_addr_ipv6(char *hostname, struct sockaddr_in6 *sa) {

    int rv;
    struct addrinfo hint, *ainfo;

    if (!hostname || !*hostname) {
	sa->sin6_addr = in6addr_any;
	return 0;
    }

    /* numeric form? */
    if (!inet_pton(AF_INET6, hostname, sa->sin6_addr.s6_addr)) {
	return 0;
    }

    memset(&hint, 0, sizeof(hint));
    hint.ai_family = AF_INET6;
    hint.ai_socktype = 0;
    hint.ai_protocol = 0;

    if ((rv = getaddrinfo(hostname, NULL, &hint, &ainfo)) < 0) {

	if (rv == EAI_NONAME) {
	    fprintf(stderr, "host '%s' not found\n", hostname);
	} else {
	    fprintf(stderr, "getaddrinfo() for host '%s' failed: %s\n",
		    hostname, gai_strerror(rv));
	}

	return -1;

    } else {

	sa->sin6_addr = ((struct sockaddr_in6 *)
			 (ainfo->ai_addr))->sin6_addr;
	freeaddrinfo(ainfo);
	return 0;
    }

}
#endif

static int
sock_add_listeners_ipv4(struct sc_list_t *list) {

    int port, sock;
    char *host;
    struct sockaddr_in sa;

    while(list) {

	host = list->data;
	if ((port = sock_extract_port(host)) < 0) {
	    return -1;
	} else if (sock_fill_addr_ipv4(host, &sa) < 0) {
	    return -1;
	}

	sa.sin_family = AF_INET;
	sa.sin_port = htons(port);
	sock = sock_create(AF_INET, (struct sockaddr *)&sa, sizeof(sa),
			   host ? host : "<all>");
	if (sock >= 0) {
	    sock_add(sock, (struct sockaddr *)&sa, sock_accept, NULL);
	} else {
	    return -1;
	}

	list = list->next;
    }

    return 0;
}

#ifdef ENABLE_IPV6
static int
sock_add_listeners_ipv6(struct sc_list_t *list) {

    int port, sock;
    char *host;
    struct sockaddr_in6 sa;

    while(list) {

	host = list->data;
	if ((port = sock_extract_port(host)) < 0) {
	    return -1;
	} else if (sock_fill_addr_ipv6(host, &sa) < 0) {
	    return -1;
	}

	sa.sin6_family = AF_INET6;
	sa.sin6_flowinfo = 0;
	sa.sin6_port = htons(port);
	sock = sock_create(AF_INET6, (struct sockaddr *)&sa, sizeof(sa),
			   host ? host : "<default>");


	if (sock >= 0) {
	    sock_add(sock, (struct sockaddr *)&sa, sock_accept, NULL);
	} else if (errno == EAFNOSUPPORT) {

	    /* fail only if ipv6 interfaces were specified */
	    if (sc_config->listen_addrs_ipv6) {
		fprintf(stderr, 
			"IPv6 interfaces were specified but "
			"IPv6 is not supported on this machine\n");
		return -1;
	    } else {

		/* fail others also */
		syslog(LOG_NOTICE, "no IPv6 interfaces found");
		return 0;
	    }

	} else {
	    return -1;
	}

	list = list->next;
    }

    return 0;
}
#endif /* ENABLE_IPV6 */

int sock_init(enum sc_bootstages_t stage) {


    if (stage == SC_STAGE_START) {

#ifdef ENABLE_EPOLL
	epollfd = epoll_create(EPOLL_FDSIZE);
	if (epollfd < 0) {
	    syslog(LOG_WARNING, 
		   "not using epoll though enabled. "
		   "epoll_create() returned: %s", strerror(errno));
	} else {
	    syslog(LOG_NOTICE,
		   "using epoll() for io multiplexing");
	}
#endif

#ifdef ENABLE_POLL
	syslog(LOG_NOTICE,
	       "using poll() for io multiplexing");
#endif
	
#ifdef ENABLE_SELECT
	syslog(LOG_NOTICE,
	       "using plain old select() for io multiplexing");
	FD_ZERO(&peer_rdset);
	FD_ZERO(&peer_wrset);
#endif

	/* IOV */
#ifndef IOV_MAX
# ifdef _SC_IOV_MAX
	iov_max = sysconf(_SC_IOV_MAX);
# else
	iov_max = 8; /* sane default */
#endif /* _SC_IOV_MAC */
#else
	iov_max = IOV_MAX;
#endif /* IOV_MAX */
    }

    if (stage == SC_STAGE_START) {

	int at_least_one_socket_after_very_hard_trying = 0;

	/* unix listen sockets */
	if (!sc_config->disable_unix) {
	    if (sock_add_listeners_unix() < 0) return -1;
	    at_least_one_socket_after_very_hard_trying++;
	}

	/* ipv4 listen sockets */
	if (!sc_config->disable_ipv4) {
	    if (!sc_config->listen_addrs_ipv4) {

		struct sc_list_t temp = { NULL, NULL };
		if (sock_add_listeners_ipv4(&temp) < 0) return -1;
		
	    } else if (sock_add_listeners_ipv4
		       (sc_config->listen_addrs_ipv4) < 0) {
		return -1;
	    }
	    at_least_one_socket_after_very_hard_trying++;
	}

#ifdef ENABLE_IPV6
	/* ipv6 listen sockets */
	if (!sc_config->disable_ipv6) {
	    if (!sc_config->listen_addrs_ipv6) {
		
		struct sc_list_t temp = { NULL, NULL };
		if (sock_add_listeners_ipv6(&temp) < 0) return -1;
		
	    } else if (sock_add_listeners_ipv6
		       (sc_config->listen_addrs_ipv6) < 0) {
		return -1;
	    }
	    at_least_one_socket_after_very_hard_trying++;
	}
#endif

	if (!at_least_one_socket_after_very_hard_trying) {
	    fprintf(stderr, "all listening sockets disabled. "
		    "no unix, no ipv4, no ipv6. nothing. "
		    "exiting)\n");
	    return -1;
	}
    }



    return 0;
}

/*
  Local variables: ***
  mode: c ***
  c-basic-offset: 4 ***
  End: ***
*/

