/*
 * Session Caching Daemon
 *
 * Copyright (C) 2008 Juha Railoma <jrma@iki.fi>
 *
 *
 * This file is part of Session Caching Daemon.
 *
 * Session Caching Daemon is free software: you can redistribute
 * it and/or modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later version.
 *
 * Session Caching Daemon is distributed in the hope that it will
 * be useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Session Caching Daemon.  If not, see
 * <http://www.gnu.org/licenses/>.
 *
 *
 * For license terms, see the file COPYING in this directory.
 */

/*
 * TODO: dynamic-index (-1 == edellinen kytetty, huomioi 
 * session_index_release
 *
 * (...ja mithn viddua tuo edellinen tarkoitti, t: 4kk myhemmin...)
 *
 */


#include "scache.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <sys/types.h>
#include <obstack.h>

#define obstack_chunk_alloc    malloc
#define obstack_chunk_free     free

#define COLLECT_CHUNK_SIZE     (16 * 1024)

#define SC_EXPIRE_INTERVAL       600
#define SC_EXPIRE_MINTRIGGER     10
#define SC_EXPIRE_FORCED         10

struct sc_cnode_t {
    
    int keylen;
    int vlen;
    int flags;

    char *key;

    union {
	struct { /* SC_NODE_VALUE */
	    int data_len;
	    char *data;
	} data;
	struct sc_tree_t *tree; /* SC_NODE_SUBTREE */
    } u;

    struct sc_tree_t *parent;
    struct sc_cnode_t *prev;
    struct sc_cnode_t *next;
};

/* ring */
struct sc_relem_t {
    int len;
    void *data;
    struct sc_relem_t *next;
    struct sc_relem_t *prev;
};

struct sc_rnode_t {

    int size;
    int keylen;
    int flags;

    char *key;

    union {
	struct sc_relem_t *current;
	struct sc_tree_t *tree; /* SC_NODE_SUBTREE */
    } u;
};


/* main root */
#ifndef SEPARATE_SESSIONTREES
static struct sc_tree_t *session_tree;
#endif
static struct sc_tree_t *host_tree;
static struct sc_tree_t *auth_tree;

/* sessiontable */
static int session_idx_size;
static struct sc_session_t **session_idx;

static int expirable_count;

/* collect-obstack */
static struct obstack coll_obs;

#ifdef SESSION_NODES_ACCOUNTING
static inline void
nodes_track_add(struct sc_session_t *session, int64_t amount) {
    session->nodes_allocated += amount;
    session->partition->nodes_allocated += amount;
}

static inline void
nodes_track_sub(struct sc_session_t *session, int64_t amount) {
    session->nodes_allocated -= amount;
    session->partition->nodes_allocated -= amount;

    assert((session->nodes_allocated >= 0));
    assert((session->partition->nodes_allocated >= 0));
}
#else
#define nodes_track_add(x, y) ;
#define nodes_track_sub(x, y) ;
#endif


#ifdef SESSION_MEMORY_ACCOUNTING
static inline void
mem_track_add(struct sc_session_t *session, int64_t amount) {
    session->memory_allocated += amount;
    session->partition->memory_allocated += amount;
}

static inline void
mem_track_sub(struct sc_session_t *session, int64_t amount) {

    session->memory_allocated -= amount;
    session->partition->memory_allocated -= amount;
    assert((session->memory_allocated >= 0));
    assert((session->partition->memory_allocated >= 0));
}
#else
#define mem_track_add(x, y) ;
#define mem_track_sub(x, y) ;
#endif



/*************************************************
 *  NODE-CLEARING                                *
 ************************************************* */
static void
node_tree_release_r(char *key, 
		    struct sc_node_t *node,
		    struct sc_session_t *session) {

    /* release-callback-func for calling recursively
       from tree_destroy */
    if (node->flags == SC_NODE_SUBTREE) {
	tree_destroy(node->u.tree, (void *)node_tree_release_r, session);
    } 

    if (node->flags == SC_NODE_VALUE) {
	mem_track_sub(session, node->u.data.data_len);
	sc_unref(node->u.data.data);
    }

    mem_track_sub(session, node->keylen);
    nodes_track_sub(session, 1);
};


static void
node_clear(struct sc_session_t *session,
	   struct sc_node_t *node) {

    /* clear recursively contents, but don't free
       node itself */
    nodes_track_add(session, 1); /* fixup to avoid assert... */

    /* note, key remains untouched, only data-part and tree
       below gets cleared */
    node_tree_release_r(NULL, node, session);
    node->flags = SC_NODE_EMPTY;
}



/*************************************************
 *  NODE-KEYLEN                              *
 ************************************************* */
static struct sc_node_t *
node_alloc(struct sc_session_t *session, int type, int keylen) {

    struct sc_node_t *node = sc_alloc(sizeof(struct sc_node_t));
    if (type == SC_NODE_SUBTREE) {
	node->flags = SC_NODE_SUBTREE;
	node->u.tree = tree_create();
    } else {
	node->flags = SC_NODE_EMPTY;
    }

#ifdef SESSION_MEMORY_ACCOUNTING
    mem_track_add(session, keylen);
    node->keylen = keylen;
#endif

    nodes_track_add(session, 1);
    return node;

    keylen = keylen; /* make compiler happy */
}


/*************************************************
 *  NODE-MANAGEMENT                              *
 ************************************************* */
static void 
node_path_trim(char **path) {

    char *mark;

    while(**path && (**path == '/')) (*path)++;
    if ((mark = strrchr(*path, (int)'/')))
	while(!mark[1] && *mark == '/') *mark-- = 0;
}


/* node_path_build()
 *
 * ensure existence and build if necessary the parenting
 * directory-part of submitted path. 
 *
 *   "mkdir -p $(dirname path)"
 *
 * after execution path points to last component of path
 * (ie. $(basename path)) and returned node is node in
 * which that last component should be created.
 *
 */
static struct sc_node_t *
node_path_build(struct sc_session_t *session, char **path) {

    char *mark, *token;

    struct sc_node_t *node = NULL;
    struct sc_tree_t *ntree = NULL;
    struct sc_tree_t *tree = session->root_node->u.tree;

    assert(path);
    assert(session);
    assert((session->root_node->flags == SC_NODE_SUBTREE));

    node_path_trim(path);
    token = mark = *path;

    /* for every path-component but last */
    while(*mark && (mark = strchr(mark, (int)'/'))) {

	for(*path = mark; *++(*path) && **path == '/';); /* next */

	*mark = 0; 
	if (ntree) { /* shortcut, rest is missing anyway,
			no need to try tree_fetch. just build */

	    node = node_alloc(session, SC_NODE_SUBTREE, strlen(token));
	    tree_add_referable_d(ntree, token, node);
	    sc_unref(node);

	    tree = ntree = node->u.tree;

	} else if ((node = tree_fetch(tree, token))) {

	    if (node->flags == SC_NODE_SUBTREE) {

		/* recurse for next path component */
		tree = node->u.tree;
		
	    } else {
		
		/* replace content */
		node_clear(session, node);
		tree = ntree = tree_create(); /* future hits in 
						 if (ntree) */

#ifdef SESSION_MEMORY_ACCOUNTING
		node->keylen = strlen(token);
		mem_track_add(session, node->keylen);
#endif
		node->flags = SC_NODE_SUBTREE;
		node->u.tree = ntree;
	    }
	    
	} else {

	    /* end of tree structure, need to build rest */
	    node = node_alloc(session, SC_NODE_SUBTREE, strlen(token));
	    tree_add_referable_d(tree, token, node);
	    sc_unref(node);

	    tree = ntree = node->u.tree; /* future hits in
					    if (ntree) */
	}

	*mark = '/';
	token = mark = *path;
    }

    return node ? node : session->root_node;
}


static struct sc_node_t *
node_path_find(struct sc_session_t *session, char **path) {

    char *mark, *token;

    struct sc_node_t *node = session->root_node;
    struct sc_tree_t *tree = session->root_node->u.tree;

    assert(path);
    assert(session);
    assert((session->root_node->flags == SC_NODE_SUBTREE));

    node_path_trim(path);
    token = mark = *path;
    while(*mark && (mark = strchr(mark, (int)'/'))) {

	for(*path = mark; *++(*path) && **path == '/';); /* next */

	*mark = 0; node = tree_fetch(tree, token); *mark = '/';
	token = mark = *path;

	if (node && (node->flags == SC_NODE_SUBTREE)) {
	    tree = node->u.tree;
	} else {
	    break;
	}
    }

    if (node && !strchr(token, (int)'/')) return node; /* last */
    else return NULL;
}

/************************************************
 * CACHENODE
 ************************************************ */
static void
cnode_tree_release_r(char *key,
		     struct sc_cnode_t *cnode,
		     struct sc_cache_t *cache) {

    /* release-callback-func for calling recursively
       from tree_destroy */
    if (cnode->flags == SC_NODE_SUBTREE) {

	tree_destroy(cnode->u.tree, (void *)cnode_tree_release_r, cache);

    } else if (cnode->flags == SC_NODE_VALUE) {

	/* memory accounting */
	cache->cur_size -= cnode->vlen;
	assert((cache->cur_size >= 0));

	/* release data */
	sc_unref(cnode->u.data.data);

	/* remove from ring */
	assert((cache->ring->prev->flags == SC_NODE_VALUE));

	if (cache->ring == cnode) cache->ring = cnode->next;
	cnode->prev->next = cnode->next;
	cnode->next->prev = cnode->prev;
	if (cache->ring == cnode) cache->ring = NULL;

#ifndef NDEBUG
	cnode->next = cnode->prev = NULL;
#endif
    } else {

	assert(!(cnode->next || cnode->prev));

    }
}

static void
cnode_clear(struct sc_cache_t *cache,
	    struct sc_cnode_t *cnode) {

    cnode_tree_release_r(NULL, cnode, cache);
    cnode->flags = SC_NODE_EMPTY;
}

static struct sc_cnode_t *
cnode_alloc(struct sc_cache_t *cache, int type, int keylen) {

    struct sc_cnode_t *cnode = sc_alloc(sizeof(struct sc_cnode_t));

    memset(cnode, 0, sizeof(struct sc_cnode_t));
    if (type == SC_NODE_SUBTREE) {
	cnode->flags = SC_NODE_SUBTREE;
	cnode->u.tree = tree_create();
    } else {
	cnode->flags = SC_NODE_EMPTY;
    }

    cnode->keylen = keylen;
    //cache->cur_size += keylen;
    return cnode;
}

static struct sc_cnode_t *
cnode_path_build(struct sc_cache_t *cache, char **path) {

    char *mark, *token;

    struct sc_cnode_t *cnode = NULL;
    struct sc_tree_t *ntree = NULL;
    struct sc_tree_t *tree = cache->root_node->u.tree;

    assert(path);
    assert(cache);
    assert((cache->root_node->flags == SC_NODE_SUBTREE));

    node_path_trim(path);
    token = mark = *path;

    /* for every path-component but last */
    while(*mark && (mark = strchr(mark, (int)'/'))) {

	for(*path = mark; *++(*path) && **path == '/';); /* next */

	*mark = 0; 
	if (ntree) { /* shortcut, rest is missing anyway,
			no need to try tree_fetch. just build */

	    cnode = cnode_alloc(cache, SC_NODE_SUBTREE, strlen(token));
	    cnode->parent = ntree;
	    cnode->key = sc_strdup(token);

	    tree_add_referable_kd(ntree, cnode->key, cnode);
	    sc_unref(cnode->key);
	    sc_unref(cnode);

	    tree = ntree = cnode->u.tree;

	} else if ((cnode = tree_fetch(tree, token))) {

	    if (cnode->flags == SC_NODE_SUBTREE) {

		/* recurse for next path component */
		tree = cnode->u.tree;
		
	    } else {
		
		/* replace content */
		sc_ref(cnode->key); /* not to be freed on clear */
		cnode_clear(cache, cnode);

		tree = ntree = tree_create(); /* future hits in 
						 if (ntree) */
		cnode->flags = SC_NODE_SUBTREE;
		cnode->u.tree = ntree;
	    }
	    
	} else {

	    /* end of tree structure, need to build rest */
	    cnode = cnode_alloc(cache, SC_NODE_SUBTREE, strlen(token));
	    cnode->parent = tree;
	    cnode->key = sc_strdup(token);
	    tree_add_referable_kd(tree, cnode->key, cnode);
	    sc_unref(cnode->key);
	    sc_unref(cnode);

	    tree = ntree = cnode->u.tree; /* future hits in
					    if (ntree) */
	}

	*mark = '/';
	token = mark = *path;
    }

    return cnode ? cnode : cache->root_node;
}

static struct sc_cnode_t *
cnode_path_find(struct sc_cache_t *cache, char **path) {

    char *mark, *token;

    struct sc_cnode_t *cnode = cache->root_node;
    struct sc_tree_t *tree = cache->root_node->u.tree;

    assert(path);
    assert(cache);
    assert((cache->root_node->flags == SC_NODE_SUBTREE));

    node_path_trim(path);
    token = mark = *path;

    while(*mark && (mark = strchr(mark, (int)'/'))) {

	for(*path = mark; *++(*path) && **path == '/';); /* next */

	*mark = 0; cnode = tree_fetch(tree, token); *mark = '/';
	token = mark = *path;

	if (cnode && (cnode->flags == SC_NODE_SUBTREE)) {
	    tree = cnode->u.tree;
	} else {
	    break;
	}
    }

    assert((*token));

    if (cnode && !strchr(token, (int)'/')) return cnode; /* last */
    else return NULL;
}

static void 
cnode_unparent(struct sc_cache_t *cache, 
	       struct sc_cnode_t *cnode) {

    assert((cache));
    assert((cnode));
    assert((cnode->parent));

    cnode_clear(cache, cnode);
    if (tree_remove(cnode->parent, cnode->key) != cnode)
	abort();
}

/************************************************
 * RINGNODE
 ************************************************ */
static void
rnode_tree_release_r(char *key,
		     struct sc_rnode_t *rnode,
		     struct sc_ring_t *ring) {

    /* release-callback-func for calling recursively
       from tree_destroy */
    if (rnode->flags == SC_NODE_SUBTREE) {

	tree_destroy(rnode->u.tree, (void *)rnode_tree_release_r, ring);

    } else if (rnode->flags == SC_NODE_VALUE) {

	struct sc_relem_t *elem;

	/* release ring */
	if ((elem = rnode->u.current)) {
	    elem->prev->next = NULL;
	    do {

		struct sc_relem_t *pelem = elem;
		elem = elem->next;

		ring->cur_nodes--;
		ring->cur_memory -= pelem->len;
		assert((ring->cur_nodes >= 0));
		assert((ring->cur_memory >= 0));

		sc_unref(pelem->data);
		sc_unref(pelem);

		assert((rnode->size >= 0));

	    } while(elem);
	}

	assert((ring->cur_memory >= 0));
    }
}

static void
rnode_clear(struct sc_ring_t *ring,
	    struct sc_rnode_t *rnode) {

    rnode_tree_release_r(NULL, rnode, ring);
    rnode->flags = SC_NODE_EMPTY;
    rnode->size = 0;
    rnode->u.current = NULL;
}

static struct sc_rnode_t *
rnode_alloc(struct sc_ring_t *session, int type, int keylen) {

    struct sc_rnode_t *rnode = sc_alloc(sizeof(struct sc_rnode_t));
    if (type == SC_NODE_SUBTREE) {
	rnode->flags = SC_NODE_SUBTREE;
	rnode->u.tree = tree_create();
    } else {
	rnode->flags = SC_NODE_EMPTY;
	rnode->size = 0;
	rnode->u.current = NULL;
    }
    
    return rnode;

    keylen = keylen; /* make compiler happy */
}

static struct sc_rnode_t *
rnode_path_build(struct sc_ring_t *ring, char **path) {

    char *mark, *token;

    struct sc_rnode_t *rnode = NULL;
    struct sc_tree_t *ntree = NULL;
    struct sc_tree_t *tree = ring->root_node->u.tree;

    assert(path);
    assert(ring);
    assert((ring->root_node->flags == SC_NODE_SUBTREE));

    node_path_trim(path);
    token = mark = *path;

    /* for every path-component but last */
    while(*mark && (mark = strchr(mark, (int)'/'))) {

	for(*path = mark; *++(*path) && **path == '/';); /* next */

	*mark = 0; 
	if (ntree) { /* shortcut, rest is missing anyway,
			no need to try tree_fetch. just build */

	    rnode = rnode_alloc(ring, SC_NODE_SUBTREE, strlen(token));
	    tree_add_referable_d(ntree, token, rnode);
	    sc_unref(rnode);

	    tree = ntree = rnode->u.tree;

	} else if ((rnode = tree_fetch(tree, token))) {

	    if (rnode->flags == SC_NODE_SUBTREE) {

		/* recurse for next path component */
		tree = rnode->u.tree;
		
	    } else {
		
		/* replace content */
		rnode_clear(ring, rnode);
		tree = ntree = tree_create(); /* future hits in 
						 if (ntree) */
		rnode->flags = SC_NODE_SUBTREE;
		rnode->u.tree = ntree;
	    }
	    
	} else {

	    /* end of tree structure, need to build rest */
	    rnode = rnode_alloc(ring, SC_NODE_SUBTREE, strlen(token));
	    tree_add_referable_d(tree, token, rnode);
	    sc_unref(rnode);

	    tree = ntree = rnode->u.tree; /* future hits in
					    if (ntree) */
	}

	*mark = '/';
	token = mark = *path;
    }

    return rnode ? rnode : ring->root_node;
}

static struct sc_rnode_t *
rnode_path_find(struct sc_ring_t *ring, char **path) {

    char *mark, *token;

    struct sc_rnode_t *rnode = ring->root_node;
    struct sc_tree_t *tree = ring->root_node->u.tree;

    assert(path);
    assert(ring);
    assert((ring->root_node->flags == SC_NODE_SUBTREE));

    node_path_trim(path);
    token = mark = *path;
    while(*mark && (mark = strchr(mark, (int)'/'))) {

	for(*path = mark; *++(*path) && **path == '/';); /* next */

	*mark = 0; rnode = tree_fetch(tree, token); *mark = '/';
	token = mark = *path;

	if (rnode && (rnode->flags == SC_NODE_SUBTREE)) {
	    tree = rnode->u.tree;
	} else {
	    break;
	}
    }

    if (rnode && !strchr(token, (int)'/')) return rnode; /* last */
    else return NULL;
}


/* ======================================== */

static inline struct sc_partition_t *
partition_lookup(const char *auth, const char *host) {

    struct sc_partition_t *partition;

    /* auth first, if allowed and resolvable */
    if (auth && *auth && sc_config->enable_auth_partitions) {

	partition = tree_fetch(auth_tree, auth);
	if (!partition && (partition = config_get_authinfo(auth))) {
	    tree_set_referable_d(auth_tree, (char *)auth, partition);
	}

	if (partition) return partition;
    }

    /* host next if allowed and resolvable */
    if (host && *host && sc_config->enable_host_partitions) {

	partition = tree_fetch(host_tree, host);
	if (!partition && (partition = config_get_hostinfo(host))) {
	    tree_set_referable_d(host_tree, (char *)host, partition);
	}

	if (partition) return partition;
    }

    /* fallback and cache them permanently */
    partition = config_get_default_partition();
    if (auth && sc_config->enable_auth_partitions) {
	tree_set_referable_d(auth_tree, (char *)auth, partition);
    }

    if (host && sc_config->enable_host_partitions) {
	tree_set_referable_d(host_tree, (char *)host, partition);
    }
    return partition;
}





/*************************************************
 *  SESSIONS                                     *
 ************************************************* */
static void
sessiontable_expand() {

    int new_size = session_idx_size + 128;
    
    session_idx = realloc(session_idx, (new_size * sizeof(session_idx[0])));
    while(session_idx_size < new_size) {
	session_idx[session_idx_size++] = NULL;
    }
};

static int
sessiontable_alloc(struct sc_session_t *session) {

    int i;

    assert((session));
    sc_ref(session);

    for(i = 0; i < session_idx_size; i++) {
	if (session_idx[i] == NULL) {
	    session_idx[i] = session;

	    if (sc_stats->sessiontable_max < i) 
		sc_stats->sessiontable_max = i;
	    return i;
	}
    }
    
    /* fallthru */
    sessiontable_expand();
    session_idx[i] = session;
    sc_stats->sessiontable_max = i;
    return i;
}

static void
sessiontable_free(int idx) {
    
    assert((session_idx[idx]));
    assert((session_idx[idx]->idx == idx));
    
    sc_unref(session_idx[idx]);
    session_idx[idx] = NULL;
};


static int
partition_verify_limits(struct sc_partition_t *partition) {

    /* session verify, limits, expiration */
#ifdef SESSION_MEMORY_ACCOUNTING
    if (partition->memory_allocated > partition->memory_limit) {

	if (sc_config->loglevel >= SC_LOG_NOTICE) {
	    syslog(LOG_NOTICE, 
		   "memory limits reached on partition '%s'. "
		   "%lu allocated when limit is %lu\n",
		   partition->id,
		   (unsigned long)partition->memory_allocated,
		   (unsigned long)partition->memory_limit);

	}

	errno = ENOSPC;
	return -1;
    }
#endif

#ifdef SESSION_NODES_ACCOUNTING
    if (partition->nodes_allocated > partition->nodes_limit) {

	if (sc_config->loglevel >= SC_LOG_NOTICE) {
	    syslog(LOG_NOTICE, 
		   "node limits reached on partition '%s'. "
		   "%lu allocated when limit is %lu\n",
		   partition->id,
		   (unsigned long)partition->nodes_allocated,
		   (unsigned long)partition->nodes_limit);

	}

	errno = ENOSPC;
	return -1;
    }
#endif

    return 0;
}



/* clear session contents, but leave in tree.
   may be reused, but then empty */
void
session_clear(struct sc_session_t *session) {

    assert((session));

    /* session-tree */
    if (session->root_node) {
	node_clear(session, session->root_node);
	nodes_track_sub(session, 1);
	sc_unref(session->root_node);
    }
    session->root_node = NULL;

    if (session->secret) sc_unref(session->secret);
    session->secret = NULL;

    sc_stats->session_clears++;

#ifndef NDEBUG
    /* not needed any more */
    session->partition = NULL;
#endif
}

/* destroy session, remove from index and tree */
void
session_destroy(struct sc_session_t *session) {

    assert((session));
    
    /* delete from tree */
    if (!tree_remove(
#ifdef SEPARATE_SESSIONTREES
		     session->partition->session_tree, 
#else
		     session_tree,
#endif
		     session->tag)) {

	if (sc_config->loglevel >= SC_LOG_FATAL) {
	    syslog(LOG_ERR, 
		   "session '%s' was not in tree. should not happen",
		   session->tag);
	}
    }

    /* clear contents */
    session_clear(session);

    /* clear */
    sessiontable_free(session->idx);
    sc_unref(session->tag);
    sc_unref(session);

    sc_stats->session_destroys++;

#ifndef NDEBUG
    memset(session, 0, sizeof(*session));
#endif
}


static inline int
session_verify_limits(struct sc_session_t *session) {

    /* verify limits */
    if (partition_verify_limits(session->partition) < 0) {
	
	/* partition full */
	if (!(session->flags & SCACHE_OPT_KEEPONELIMIT)) {

	    syslog(LOG_NOTICE,
		   "partition '%s' full, clearing session '%s'",
		   session->partition->id,
		   session->tag);

	    session->expires = sc_lazytime - 1;
	    session_clear(session);
	}

	return -1;
    }

    return 0;
}

struct sc_session_t *
session_create(const char *tag,
	       const char *secret,
	       const char *auth,
	       const char *host,
	       int flags,
	       int expires) {

    struct sc_partition_t *partition;
    struct sc_session_t *session;

    /* verify arguments */
    if (!tag || !*tag) { /* tag must exist */
	errno = EINVAL;
	return NULL;
    }

    /* empty secret is null secret */
    if (!secret || !*secret) secret = NULL;

    /* collect host and auth info */
    partition = partition_lookup(auth, host);

    /* not over limit */
    if (partition_verify_limits(partition) < 0) {

	errno = ENOSPC;
	return NULL;
    }

#ifdef SEPARATE_SESSIONTREES
    session = tree_fetch(partition->session_tree, tag);
#else
    session = tree_fetch(session_tree, tag);
#endif
    if (session) { /* already exists, is same, reset */

	if (session->secret && 
	    (!secret || strcmp(secret, session->secret))) {

	    /* secret mismatch */
	    if (sc_config->loglevel >= SC_LOG_ERROR) {

		syslog(LOG_NOTICE,
		       "session already exists, but secret differs "
		       "on partition '%s'",
		       partition->id);

		errno = EEXIST;
		return NULL;
	    }
	}

	/* secret matches or null, so reset necessary parts */
	session_clear(session);
	sc_stats->session_reuses++;

    } else {

	session = sc_alloc(sizeof(struct sc_session_t));
	memset(session, 0, sizeof(struct sc_session_t));

	/* fill rest */
	session->tag = sc_strdup(tag);
	session->flags = 0;
	session->idx = sessiontable_alloc(session);

	/* add to tree */
#ifdef SEPARATE_SESSIONTREES
	tree_set_referable_kd(partition->session_tree, 
			      session->tag, session);
#else
	tree_set_referable_kd(session_tree, 
			      session->tag, session);
#endif	
    }

    
    session->partition = partition; /* do before node_alloc 
				       so that nodes accounting gets
				       right */
    session->expires = session->maxidle = expires;
    session->expires += sc_lazytime;
    session->root_node = node_alloc(session, SC_NODE_SUBTREE, 0);
    if (secret) session->secret = sc_strdup(secret);

    sc_stats->session_creates++;
    return session;
}

int 
session_rename(struct sc_session_t *session, const char *tag) {

    void *node = NULL;
    char *newtag = sc_strdup(tag);

    assert((session));

    /* insert with new name */
#ifdef SEPARATE_SESSIONTREES
    node = tree_add_referable_kd(session->partition->session_tree, 
				 newtag, session);
#else
    node = tree_add_referable_kd(session_tree, newtag, session);
#endif

    if (!node) {

	/* insert with identifier */
#ifdef SEPARATE_SESSIONTREES
	node = tree_remove(session->partition->session_tree, 
			   session->tag);
#else
	node = tree_remove(session_tree, session->tag);
#endif

	sc_unref(session->tag);
	session->tag = newtag;
	return 0;

    } else {
	errno = EEXIST;
	return -1;
    }
}

int
session_drop(struct sc_session_t *session) {

    assert((session));

    /* destroy immediately, so that can be reassigned */
    session_destroy(session);
    return 0;
}


void
session_expire() {

    int i, destroyed, cleared;
    int expire_delay;

    static int forced_run = 0;
    static time_t next_run = 1; /* might be unsigned */
    struct sc_session_t *session;

    if ((++forced_run % SC_EXPIRE_FORCED) &&
	(expirable_count < SC_EXPIRE_MINTRIGGER) &&
	(sc_lazytime < next_run)) return;


    next_run = sc_lazytime + SC_EXPIRE_INTERVAL;
    expire_delay = sc_config->expire_delay;

    for(i = destroyed = cleared = 0; i < session_idx_size; i++) {

	if (!(session = session_idx[i])) continue;


	if (session->expires < (sc_lazytime - expire_delay)) {

	    session_destroy(session);
	    destroyed++;

	} else if (session->expires < sc_lazytime &&
		   session->root_node) {

	    /* keep in tree */
	    session_clear(session);
	    cleared++;
	}
    }

    expirable_count = 0;
    if (sc_config->loglevel >= SC_LOG_DEBUG) {
	syslog(LOG_NOTICE, "session_expire(): %u cleared %u destroyed",
	       cleared, destroyed);
    }
}

static struct sc_session_t *
session_create_dummy() {

    struct sc_session_t *session;

    session = sc_alloc(sizeof(struct sc_session_t));
    memset(session, 0, sizeof(struct sc_session_t));

    session->tag = sc_strdup("");
    session->idx = -1;
    session->partition = config_get_unlimited_partition();
    session->root_node = node_alloc(session, SC_NODE_SUBTREE, 0);

    return session;
}


struct sc_session_t *
session_host_lval(struct sc_session_t *session) {

    assert((session));
    if (!session->partition->lvsession && 
	session->partition->allow_lvals) {

	/* dummy-session */
	session->partition->lvsession = session_create_dummy();
    }

    return session->partition->lvsession;
}

struct sc_session_t *
session_host_shared(struct sc_session_t *session) {

    assert((session));
    if (!session->partition->shsession && 
	session->partition->allow_shared) {

	/* dummy-session */
	session->partition->shsession = session_create_dummy();
    }

    if (partition_verify_limits(session->partition) < 0) return NULL;
    return session->partition->shsession;
}

struct sc_session_t *
session_host_session(const char *auth, const char *host) {

    struct sc_partition_t *partition = partition_lookup(auth, host);
    if (!partition->snsession && partition->allow_session) {

	/* dummy-session */
	partition->snsession = session_create_dummy();
    }

    if (partition_verify_limits(partition) < 0) return NULL;
    return partition->snsession;
}


struct sc_session_t *
session_open(const char *tag, 
	     const char *secret, 
	     const char *auth,
	     const char *host,
	     int idx) {

    struct sc_session_t *session = NULL;
#ifdef SEPARATE_SESSIONTREES
    struct sc_partition_t *partition = NULL;
#endif

    assert((tag));

    if (!tag || !*tag) return NULL;
    if (!secret || !*secret) secret = NULL;

#ifdef SEPARATE_SESSIONTREES
    partition = partition_lookup(auth, host);
#endif

    /* try shortcutting on supplied index-slot */
    if (idx >= 0 && idx < session_idx_size && session_idx[idx]) {
	if (!strcmp(tag, session_idx[idx]->tag)) {

#ifdef SEPARATE_SESSIONTREES
	    if (session_idx[idx]->partition == partition) {
		/* should also match partition */
		sc_stats->session_hit_by_idx++;
		session = session_idx[idx];
	    }
#else
	    sc_stats->session_hit_by_idx++;
	    session = session_idx[idx];
#endif
	} else {
	    sc_stats->session_miss_by_idx++;
	}
    }


    /* else fallback to search */
    if (!session) {
#ifdef SEPARATE_SESSIONTREES
	session = tree_fetch(partition->session_tree, tag);
#else
	session = tree_fetch(session_tree, tag);
#endif
	if (!session) {
	    sc_stats->session_miss_by_search++;
	    return NULL;
	}
    }

    /* if secret exists, verify */
    if (session->secret) {
	if (!secret || strcmp(secret, session->secret)) {

	    /* secret mismatch */
	    sc_stats->session_secret_mismatches++;
	    return NULL;
	}
    }

    /* verify expiration... */
    if (session->expires < sc_lazytime) {
	expirable_count++;
	return NULL;
    }
	
    /* update atime (DEP:SESSION_EXPIRATION */
    session->expires = sc_lazytime + session->maxidle;
    return session;
}




/* always set key and build parenting tree */
int
session_set(struct sc_session_t *session,
	    char *key, char *data, int length) {

    struct sc_node_t *item, *node;

    assert((session));
    assert((key));
    assert((data));

    /* limits */
    if (session_verify_limits(session) < 0) return -1;

    node = node_path_build(session, &key);
    assert((node));

    if ((item = tree_fetch(node->u.tree, key))) {

	node_clear(session, item);

#ifdef SESSION_MEMORY_ACCOUNTING
	item->keylen = strlen(key);
	mem_track_add(session, item->keylen);
#endif

    } else {

	item = node_alloc(session, SC_NODE_VALUE, strlen(key));
	tree_set_referable_d(node->u.tree, key, item);
	sc_unref(item);
    }

    item->flags = SC_NODE_VALUE;
    item->u.data.data_len = length;
    item->u.data.data = data;
    mem_track_add(session, length);
    return 0;
}


/* add key if key doesn't exist */
int
session_add(struct sc_session_t *session,
	    char *key, char *data, int length) {

    struct sc_node_t *item, *node;

    assert((session));
    assert((key));
    assert((data));

    /* limits */
    if (session_verify_limits(session) < 0) return -1;

    if ((node = node_path_find(session, &key)) &&
	(item = tree_fetch(node->u.tree, key))) {

	if (item->flags != SC_NODE_EMPTY) {
	    errno = EEXIST;
	    return -1; /* exist */
	}
    }

    return session_set(session, key, data, length);
}

/* replace key if key exists */
int
session_replace(struct sc_session_t *session,
		char *path, char *data, int length) {


    struct sc_node_t *item, *node;

    assert((session));
    assert((path));
    
    /* limits */
    if (session_verify_limits(session) < 0) return -1;

    node = node_path_find(session, &path);
    if (node && 
	(node->flags == SC_NODE_SUBTREE) &&
	(item = tree_fetch(node->u.tree, path))) {
	
	/* both missing nodes and empty nodes must be considered
	   missing and so unreplaceable */
	if (item->flags != SC_NODE_EMPTY) {

	    node_clear(session, item);
	    mem_track_add(session, item->keylen);

	    if (data && length) {
		item->flags = SC_NODE_VALUE;
		item->u.data.data_len = length;
		item->u.data.data = data;
		
		mem_track_add(session, length);

	    } else {
		/* node is left empty, and should be considered 
		   equal to missing path we just mark it empty 
		   instead of removing from (and balancing) tree */
	    }
	}

	return 0;
    }

    errno = ENOENT;
    return -1;
}




int
session_unset(struct sc_session_t *session, char *path) {
    return session_replace(session, path, NULL, 0);
}


int
session_get(struct sc_session_t *session, char *path,
	    char **data, int *dlen) {

    struct sc_node_t *item, *node;

    assert((session));
    assert((path));
    assert((data));
    assert((dlen));

    node = node_path_find(session, &path);
    if (node && 
	(node->flags == SC_NODE_SUBTREE) &&
	(item = tree_fetch(node->u.tree, path))) {

	if (item->flags == SC_NODE_VALUE) {

	    *data = item->u.data.data;
	    *dlen = item->u.data.data_len;
	    return 0;
	}
    };

    /* fail */
    *data = NULL;
    *dlen = 0;
    errno = ENOENT;
    return -1;
}


enum scache_nodetypes_t
session_stat(struct sc_session_t *session, char *path,
	     char **data, int *datalen) {

    struct sc_node_t *item, *node;

    assert((session));
    assert((path));

    *data = NULL;
    *datalen = 0;

    node = node_path_find(session, &path);
    if (node && 
	(node->flags == SC_NODE_SUBTREE) &&
	(item = tree_fetch(node->u.tree, path))) {

	if (item->flags == SC_NODE_VALUE) {
	    *data = item->u.data.data;
	    *datalen = item->u.data.data_len;
	    return SCNODE_VALUE;
	} else if (item->flags == SC_NODE_SUBTREE) {
	    return SCNODE_BRANCH;
	} else {
	    return SCNODE_EMPTY; /* note below... */
	}
    };

    /* note: SCACHE_TYPE_EMPTY and SCACHE_TYPE_NONEXISTENT should
       be considered as a missing node. different types are
       just for debug */
    return SCNODE_NONEXISTENT;
}

/* *****************************
 * COLLECT
 * ***************************** */
static void
tree_collect_r(const char *key, void *value, void *counter) {

    int len;
    char *buffer;
    enum scache_nodetypes_t type;
    struct sc_node_t *node = (struct sc_node_t *)value;

    /* nodetype */
    switch(node->flags) {
    case SC_NODE_VALUE:
	type = SCNODE_VALUE;
	break;
    case SC_NODE_SUBTREE:
	type = SCNODE_BRANCH;
	break;
    default:
	return; /* skip over deleted */
    }

    if (!(node->flags == SC_NODE_VALUE ||
	  node->flags == SC_NODE_SUBTREE)) return;

    /* store name as serialized form */
#ifdef SESSION_MEMORY_ACCOUNTING
    buffer = alloca(node->keylen + 128);
    len = sprintf(buffer, "s:%u:\"%s\";i:%u;",
		  node->keylen, key, type);
#else
    buffer = alloca(strlen(key) + 128);
    len = sprintf(buffer, "s:%u:\"%s\";i:%u;",
		  strlen(key), key, type);
#endif
    obstack_grow(&coll_obs, buffer, len);
    (*(int *)counter)++;
}

char *
session_collect(struct sc_session_t *session, char *path) {

    int count = 0;
    char *header, buffer[16];
    struct sc_node_t *node;

    assert((session));
    assert((path));

    /* get parent tree */
    node = node_path_find(session, &path);
    if (!(node &&
	  (node->flags == SC_NODE_SUBTREE))) {
	
	/* not a subtree */
	errno = ENOENT;
	return NULL;
    }

    /* lookup self */
    if ((node != session->root_node) || *path) {
	node = tree_fetch(node->u.tree, path);
	if (!(node &&
	      (node->flags == SC_NODE_SUBTREE))) {
	    
	    errno = ENOENT;
	    return NULL;
	}
    } else {
	/* listed root (/) */
    }

    /* collect */
    obstack_free(&coll_obs, NULL);
    obstack_init(&coll_obs);
    obstack_grow(&coll_obs, buffer, sizeof(buffer));
    tree_enum_asc(node->u.tree, (void *)tree_collect_r, (void *)&count);

    /* finalize */
    obstack_grow0(&coll_obs, "}", 1);
    header = obstack_finish(&coll_obs);
    count = sprintf(buffer, "sa:%u:{", count); /* 's' = serialized */
    header += (sizeof(buffer) - count);
    memcpy(header, buffer, count);

    return header;
}


/* *****************************
 * CACHE 
 * ***************************** */

static struct sc_cache_t *
cache_lookup(struct sc_session_t *session) {

    struct sc_partition_t *partition = session->partition;

    assert((session->partition));

    if (!partition->cache) {
	if (!partition->allow_cache) {
	    errno = EACCES;
	    return NULL;
	}

	partition->cache = sc_alloc(sizeof(struct sc_cache_t));
	memset(partition->cache, 0, sizeof(struct sc_cache_t));
	partition->cache->root_node = 
	    cnode_alloc(partition->cache, SC_NODE_SUBTREE, 0);
	partition->cache->partition = partition;
    }

    return partition->cache;
}

int 
cache_clear(struct sc_session_t *session,
	    char *key) {

    struct sc_cache_t *cache;
    struct sc_cnode_t *item, *cnode;

    assert((session));
    assert((key));

    if (!(cache = cache_lookup(session))) return -1;

    node_path_trim(&key);
    if (!*key) {

	/* root-node-clear is stupid exception to all,
	   we clear and reinitialize it */
	cache->partition->cache_dels++;
	sc_stats->all_cache_dels++;

	cnode_clear(cache, cache->root_node);
	cache->root_node = cnode_alloc(cache, SC_NODE_SUBTREE, 0);

	assert(!cache->ring);

    } else if (!(cnode = cnode_path_find(cache, &key))) {

	/* stats */
	cache->partition->cache_misses++;
	sc_stats->all_cache_misses++;

    } else if (cnode && 
	       (cnode->flags == SC_NODE_SUBTREE) &&
	       (item = tree_fetch(cnode->u.tree, key))) {

	cache->partition->cache_dels++;
	sc_stats->all_cache_dels++;

	cnode_unparent(cache, item);
    }

    /* always success despite anything expired or not */
    return 0;
}

static void
cache_expire(struct sc_cache_t *cache) {

    while(cache->cur_size > cache->partition->cache_max) {

	cache->partition->cache_expires++;
	sc_stats->all_cache_expires++;

	assert((cache->ring->prev->flags == SC_NODE_VALUE));
	cnode_unparent(cache, cache->ring->prev);
    }
}

int 
cache_put(struct sc_session_t *session,
	  char *key, char *data, int length) {

    struct sc_cache_t *cache;
    struct sc_cnode_t *item, *cnode;

    assert((session));
    assert((key));
    assert((data));

    /* enough space */
    if (!(cache = cache_lookup(session))) return -1;
    if (cache->cur_size > cache->partition->cache_max) cache_expire(cache);
    if (cache->cur_size > cache->partition->cache_max) {
	errno = ENOSPC;
	return -1;
    }

    /* statistics */
    cache->partition->cache_adds++;
    sc_stats->all_cache_adds++;

    cnode = cnode_path_build(cache, &key);
    assert(cnode);

    if ((item = tree_fetch(cnode->u.tree, key))) {

	/* item remains untouched + compensate size reductions */
	//cache->cur_size += item->keylen;
	cnode_clear(cache, item);

    } else {

	item = cnode_alloc(cache, SC_NODE_VALUE, strlen(key));
	item->parent = cnode->u.tree;
	item->key = sc_strdup(key);
	tree_set_referable_kd(cnode->u.tree, item->key, item);

	sc_unref(item->key);
	sc_unref(item);
    }

    item->flags = SC_NODE_VALUE;
    item->u.data.data_len = length;
    item->u.data.data = data;
    item->vlen = length;
    cache->cur_size += length;

    /* to top of ring */
    if (!cache->ring) {

	cache->ring = item->next = item->prev = item;

    } else {

	item->next = cache->ring;
	item->prev = cache->ring->prev;
	item->next->prev = item->prev->next = item;
	cache->ring = item;
    }

    return 0;
}


int 
cache_get(struct sc_session_t *session,
	  char *key, char **data, int *dlen) {

    struct sc_cache_t *cache;
    struct sc_cnode_t *item, *cnode;

    assert((session));
    assert((key));
    assert((data));
    assert((dlen));

    /* preset */
    *data = NULL;
    *dlen = 0;

    if (!(cache = cache_lookup(session))) {

	return -1;

    } else if (!(cnode = cnode_path_find(cache, &key))) {

	/* stats */
	cache->partition->cache_misses++;
	sc_stats->all_cache_misses++;

	errno = ENOENT;
	return -1;

    } else if (cnode && 
	       (cnode->flags == SC_NODE_SUBTREE) &&
	       (item = tree_fetch(cnode->u.tree, key))) {

	assert((item->next && item->prev));

	/* stats */
	cache->partition->cache_hits++;
	sc_stats->all_cache_hits++;

	if (item->flags == SC_NODE_VALUE) {

	    /* move up in ring */
	    if (cache->ring != item) {

		/* loosen */
		item->prev->next = item->next;
		item->next->prev = item->prev;

#if 0
		/* bind to prev */
		item->next = item->prev;
		item->prev = item->prev->prev;
		item->prev->next = item->next->prev = item;
		if (cache->ring->prev == item) cache->ring = item;
#else
		/* bind to top */
		item->next = cache->ring;
		item->prev = cache->ring->prev;
		item->next->prev = item->prev->next = item;
		cache->ring = item;
#endif
	    }

	    *data = item->u.data.data;
	    *dlen = item->u.data.data_len;
	    return 0;
	}
    };

    /* fail */
    errno = ENOENT;
    return -1;
}

/*
int 
cache_get(const char *auth,
	  const char *host, 
	  char *key, char **data, int *datalen) {

    struct sc_cache_t *cache;
    struct sc_cachenode_t *cnode;

    assert((key));
    assert((data));
    assert((datalen));

    if (!(cache = cache_lookup(auth, host))) {
	return -1;
    } else if (!(cnode = tree_fetch(cache->tree, key))) {

	cache->partition->cache_misses++;
	sc_stats->all_cache_misses++;

	errno = ENOENT;
	return -1;
    }

    assert((cnode->next && cnode->prev));
    cache->partition->cache_hits++;
    sc_stats->all_cache_hits++;

    if (cache->ring != cnode) {

	cnode->prev->next = cnode->next;
	cnode->next->prev = cnode->prev;

#if 0
	cnode->next = cnode->prev;
	cnode->prev = cnode->prev->prev;
	cnode->prev->next = cnode->next->prev = cnode;
	if (cache->ring->prev == cnode) cache->ring = cnode;
#else
	cnode->next = cache->ring;
	cnode->prev = cache->ring->prev;
	cnode->next->prev = cnode->prev->next = cnode;
	cache->ring = cnode;
#endif
    }

    *data = cnode->value;
    *datalen = cnode->vlen;
    return 0;
}
*/

/* ====================================================
   RING 
   ==================================================== */
static struct sc_ring_t *
ring_lookup(struct sc_session_t *session) {

    struct sc_partition_t *partition = 
	    session->partition;

    assert((session->partition));

    if (!partition->ring) {
	if (!partition->allow_ring) {
	    errno = EACCES;
	    return NULL;
	}

	partition->ring = sc_alloc(sizeof(struct sc_ring_t));
	memset(partition->ring, 0, sizeof(struct sc_ring_t));
	partition->ring->partition = partition;
	partition->ring->root_node =
	    rnode_alloc(partition->ring, SC_NODE_SUBTREE, 0);
    }

    return partition->ring;
}

struct sc_rnode_t *
ring_getpath(struct sc_session_t *session, char *key) {

    struct sc_ring_t *ring;
    struct sc_rnode_t *item;

    if (!(ring = ring_lookup(session))) {
	/* fallthru, keep errno */
    } else if ((item = rnode_path_find(ring, &key)) &&
	       (item->flags == SC_NODE_SUBTREE) &&
	       (item = tree_fetch(item->u.tree, key))) {
    
	if ((item->flags == SC_NODE_VALUE)) return item;
	else errno = ENOENT;
    }

    /* fail */
    return NULL;
}

int
ring_add(struct sc_session_t *session,
	 char *key, char *data, int datalen, int as_first) {

    struct sc_ring_t *ring;
    struct sc_relem_t *elem;
    struct sc_rnode_t *item, *rnode;

    assert((session));
    assert((key));
    assert((data));

    if (!(ring = ring_lookup(session))) return -1;
    if ((ring->cur_nodes >= ring->partition->ring_nodes_limit) ||
	(ring->cur_memory >= ring->partition->ring_memory_limit)) {
	errno = ENOSPC;
	return -1;
    }

    rnode = rnode_path_build(ring, &key);
    assert((rnode));

    if ((item = tree_fetch(rnode->u.tree, key))) {

	/* clear if branch */
	if (item->flags == SC_NODE_SUBTREE)
	    rnode_clear(ring, item);

    } else {

	/* newnode */
	item = rnode_alloc(ring, SC_NODE_VALUE, strlen(key));
	item->key = sc_strdup(key);
	tree_set_referable_kd(rnode->u.tree, item->key, item);

	sc_unref(item->key);
	sc_unref(item);
    }

    /* relem */
    elem = sc_alloc(sizeof(struct sc_relem_t));
    elem->len = datalen;
    elem->data = data;

    /* increment allocated space */
    ring->cur_nodes++;
    ring->cur_memory += datalen;

    /* to ring */
    item->flags = SC_NODE_VALUE;
    item->size++;

    if (!item->u.current) {
	item->u.current = elem->next = elem->prev = elem;
    } else {

	elem->next = item->u.current;
	elem->prev = elem->next->prev;
	elem->prev->next = elem->next->prev = elem;
	if (as_first) item->u.current = elem;
    }

    return 0;
}

/* get from ring */
int
ring_fetch(struct sc_session_t *session,
	   char *key, char **dataptr, int *datalen, int op) {

    struct sc_ring_t *ring;
    struct sc_rnode_t *item;
    struct sc_relem_t *elem;

    assert((session));
    assert((key));
    assert((dataptr));
    assert((datalen));

    /* preset */
    *dataptr = NULL;
    *datalen = 0;

    if (!(ring = ring_lookup(session))) {
	/* fallthru, keep errno */
    } else if ((item = rnode_path_find(ring, &key)) &&
	       (item->flags == SC_NODE_SUBTREE) &&
	       (item = tree_fetch(item->u.tree, key)) &&
	       (item->u.current)) {
	    
	elem = item->u.current;
	assert((elem->next && elem->prev));
	
	if (op) { /* remove first or remove last */
	    
	    /* unlink from end */
	    item->size--;
	    ring->cur_nodes--;

	    assert((item->size >= 0));
	    assert((ring->cur_nodes >= 0));

	    if (op < 0) elem = elem->prev;
	    else if (op > 0) item->u.current = elem->next;

	    if (!item->size) { /* last-one */
		item->u.current = NULL;
	    } else {
		elem->prev->next = elem->next;
		elem->next->prev = elem->prev;
	    }

	    assert(!(elem->next = NULL)); /* optimizes out */
	    assert(!(elem->prev = NULL));

	    /* decount allocated space */
	    ring->cur_memory -= elem->len;
	    assert((ring->cur_memory >= 0));

	    *dataptr = elem->data;
	    *datalen = elem->len;
	    sc_unref(elem);

	    return 0;

	} else {

	    /* not to remove from ring */
	    *dataptr = elem->data;
	    *datalen = elem->len;
	    sc_ref(*dataptr); /* need to compensate for unref on
				 proto.c:handle_rnfetch_common() */

	    return 0;
	}
    }

    /* fail */
    errno = ENOENT;
    return -1;
}

/* ring, misc operations */
int
ring_purge(struct sc_session_t *session, char *key) {

    struct sc_ring_t *ring;
    struct sc_rnode_t *item, *rnode;

    assert((session));
    assert((key));

    if (!(ring = ring_lookup(session))) {
	return -1;
    } else if ((rnode = rnode_path_find(ring, &key)) &&
	       (rnode->flags == SC_NODE_SUBTREE) &&
	       (item = tree_fetch(rnode->u.tree, key))) {

	assert((item->flags != SC_NODE_SUBTREE));
	if (item->flags != SC_NODE_EMPTY) {
	    rnode_clear(ring, item);
	    return 0;
	}
    }

    errno = ENOENT;
    return -1;
}

int 
ring_size(struct sc_session_t *session, char *key) {

    struct sc_rnode_t *rnode;

    assert((session));
    assert((key));

    if (!(rnode = ring_getpath(session, key))) return -1;
    return rnode->size;
}

int
ring_rotate(struct sc_session_t *session, char *key, int count) {

    int i;
    struct sc_rnode_t *rnode;

    assert((session));
    assert((key));
    if (!(rnode = ring_getpath(session, key))) return -1;

    /* direction */
    if (count > 0) {
	count = count % rnode->size;
	for(i = 0; i < count; i++)
	    rnode->u.current = rnode->u.current->next;
    } else if (count < 0) {
	count = (-count) % rnode->size;
	for(i = 0; i < count; i++)
	    rnode->u.current = rnode->u.current->prev;
    }

    return 0;
}

/* =======================================================
   INIT */
int
store_init(enum sc_bootstages_t stage) {

    if (stage == SC_STAGE_PREPARE) {

	/* collect-obstack */
	memset(&coll_obs, 0, sizeof(struct obstack));
	obstack_init(&coll_obs);
	obstack_chunk_size(&coll_obs) = COLLECT_CHUNK_SIZE;

    } else if (stage == SC_STAGE_START) {
#ifndef SEPARATE_SESSIONTREES
	session_tree = tree_create();
#endif
	host_tree = tree_create();
	auth_tree = tree_create();
    }

    return 0;
}


/*
  Local variables: ***
  mode: c ***
  c-basic-offset: 4 ***
  End: ***
*/


